///////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2011 Sigma Designs, Inc.  All Rights Reserved.
//
// Ccomponent.c
//
// Description:
//
//  component implementation
//
// History:
//	Jan 14, 2010	(dt) created
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include "qlautoconf.h"

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <linux/mm.h>
#include <asm/uaccess.h>

#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/videodev2.h>

#include "qpdbgmgr.h"
#include "cobjectmgr.h"
#include "cdevice.h"
#include "ccomponent.h"
#include "qplinuxkermap.h"

#ifdef SUPPORT_PL3XX_OSD
#include "qlosd.h"
#endif
#include "qlvideoprop.h"
#include "plv4ldevversion.h"

#include "c353.h"
#include "cadv7604.h"
#include "cgv7601.h"
#include "cqlcodeclib.h"
#include "pl330.h"
#include "pl330-dbg.h"
#include "c353-ctrl.h"
#include "frontend.h"

#define MPEG_FRAME_SIZE 64 * 1024


extern struct vm_operations_struct plv4l_mmap_ops;

/////////////////////////////////////////////////////////////////////////////
//
//

static EQP_Errors CComponent_Start_Index(CComponent *This, COMPONENT_PORT_TYPE type);
static EQP_Errors CComponent_Start(CComponent *This);
static EQP_Errors CComponent_Stop_Index(CComponent *This, COMPONENT_PORT_TYPE type);
static void CComponent_Event_CB(void *cxt,IFrontend_event_e event);
static void pl330_default_config(struct CComponent *component);


/* check stream is stoped, called by dqbuf */
void ccomponent_wait_qbuf(struct work_struct *work)
{
    CDevice *dev = container_of(work, struct CDevice, wq_wait_qbuf);
    long timeout;

    /* wait for next qbuf, if timeout set stream_started to zero */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)
    timeout = interruptible_sleep_on_timeout(&dev->wq_qbuf, HZ);
    if (!timeout)
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s wait for qbuf timeout\n", __func__);
        atomic_set(&dev->stream_started, 0);
    }
#else
    timeout = wait_event_interruptible_timeout(dev->wq_qbuf, dev->qbuf_event, HZ);
    if (timeout < 1)
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s wait for qbuf timeout\n", __func__);
        atomic_set(&dev->stream_started, 0);
    }
    dev->qbuf_event = false;
#endif        
}

/* check stream is stoped, called by qbuf */
void ccomponent_wait_dqbuf(struct work_struct *work)
{
    CDevice *dev = container_of(work, struct CDevice, wq_wait_dqbuf);
    long timeout;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)
    /* wait for next dqbuf, if timeout set stream_started to zero */
    timeout = interruptible_sleep_on_timeout(&dev->wq_dqbuf, HZ);
    if (!timeout)
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s wait for dqbuf timeout\n", __func__);
        atomic_set(&dev->stream_started, 0);
    }
#else
    timeout = wait_event_interruptible_timeout(dev->wq_dqbuf, dev->dqbuf_event, HZ);
    if (timeout < 1)
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s wait for dqbuf timeout\n", __func__);
        atomic_set(&dev->stream_started, 0);
    }
    dev->dqbuf_event = false;
#endif        
}

static EQP_Errors CComponent_PortOpen(CComponent *This,
                                      PCOMPONENT_PORT_OPEN pOpenParam
                                      )
{
    EQP_Errors qpStatus;
    CPort *pPort;
    PORT_DIRECTION dir;
    TQP_UINT32 dwOpenType;
    PCOMPONENT_PORT_OPEN_FORMAT pOpenFormat = QP_NULL;
    TQP_UINT32 frame_nbs;
    TQP_UINT32 frame_size;
    TQP_BOOL bBufferPartialFill;
    TQP_BOOL bBufferFrameAligned;
    TQP_BOOL bSupportUserPtr;

    TQP_BOOL bForceFullBuffer = QP_FALSE;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s [\n",
                 __FUNCTION__
                 ));

    if (pOpenParam->type & COMPONENT_PORT_TYPE_FORCE_FULL_BUFFER)
    {
        bForceFullBuffer = QP_TRUE;
        pOpenParam->type &= ~COMPONENT_PORT_TYPE_FORCE_FULL_BUFFER;
    }

    // validate parameters
    //
    if (!pOpenParam ||
            (pOpenParam->type >= MAX_COMPONENT_PORT) ||
            (pOpenParam->type < 0)
            )
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                     "%s pOpenParam(0x%x) type(%d)\n",
                     __FUNCTION__,
                     pOpenParam,
                     pOpenParam ? pOpenParam->type : -1
                     ));
        return (QPERR_PARMS);
    }

    ENTER_CRITICAL(&This->m_Object)

    pPort = &This->m_Ports[pOpenParam->type];
    pl330_trace(PL330_TRACE_PORT, "%s port[%d] open\n", __func__, pOpenParam->type);
    //	port_type = pOpenParam->type;

    if (pPort->m_valid)
    {
        pl330_trace(PL330_TRACE_PORT, "%s port[%d] opened already\n",
                    __func__, pOpenParam->type);
        LEAVE_CRITICAL(&This->m_Object)
        return (QPERR_INVALID);
    }

    // only allow user pointer for PCIe bus
    //
    bSupportUserPtr = (PLV4LDEV_TYPE_QL300_PCIE_EVK == This->m_pDevice->m_type);

    // setup port specific parameters
    //
    switch (pOpenParam->type)
    {
    case COMPONENT_PORT_YUV_IN:
        dir = PORT_DIR_WRITE;
        dwOpenType = QPMPGCODEC_ENC_YUV_IN;
        pOpenFormat = &pOpenParam->format;
        frame_nbs = 4;
        frame_size = (pOpenParam->format.yuv.nWidth *
                pOpenParam->format.yuv.nHeight *
                pOpenParam->format.yuv.nBitCount) / 8
                ;
        frame_size = (frame_size + PAGE_SIZE - 1) & PAGE_MASK;
        bBufferPartialFill = QP_FALSE;
        bBufferFrameAligned = QP_TRUE;
        break;

    case COMPONENT_PORT_COMP_IN:
        dir = PORT_DIR_WRITE;
        dwOpenType = QPMPGCODEC_DEC_MPEG_IN;
        pOpenFormat = &pOpenParam->format;
#ifdef CONFIG_HOST_PLATFORM_ASPEN
        frame_nbs = 2;
#else
        frame_nbs = QPCODECLIB_MAX_DATA_ENTRIES / 16;
#endif
        frame_size = 1024 * 1024;
        bBufferFrameAligned = QP_FALSE;
        bBufferPartialFill = QP_TRUE;
        // disable user pointer support for this port
        // because it may need to access user data
        // from the driver
        //
        bSupportUserPtr = QP_FALSE;
        break;

    case COMPONENT_PORT_COMP_AUD_IN:
        dir = PORT_DIR_WRITE;
        dwOpenType = QPMPGCODEC_DEC_AES_IN;
        frame_nbs = QPCODECLIB_MAX_DATA_ENTRIES / 4;
        frame_size = 8 * 1024;
        if (frame_size < PAGE_SIZE)
            frame_size = PAGE_SIZE;
        bBufferFrameAligned = QP_FALSE;
        bBufferPartialFill = QP_TRUE;
        // disable user pointer support for this port
        // because it may need to access user data
        // from the driver
        //
        bSupportUserPtr = QP_FALSE;
        break;

    case COMPONENT_PORT_COMP_OUT:
        pl330_trace(PL330_TRACE_PORT, "%s COMPONENT_PORT_COMP_OUT\n", __func__);
        dir = PORT_DIR_READ;
        dwOpenType = QPMPGCODEC_ENC_MPEG_OUT;
        pOpenFormat = &pOpenParam->format;
        This->m_pDevice->v4l2_port_type = COMPONENT_PORT_COMP_OUT;
        if (QP_COMP_VID_DATA_TYPE_MJPEG == pOpenParam->format.vid.nDataType)
        {
#if defined(CONFIG_HOST_PLATFORM_PL330SOC)
            frame_nbs = 1;
#else
            frame_nbs = 8;
#endif
            frame_size = 1024 * 1024;
            bBufferFrameAligned = QP_TRUE;
        }
        else
        {
            frame_nbs = QPCODECLIB_MAX_DATA_ENTRIES / 16;
#ifdef CONFIG_HOST_PLATFORM_ASPEN
            frame_size = 256 * 1024;
#else
            frame_size = 64 * 1024;
#endif
            bBufferFrameAligned = QP_FALSE;
        }

        if (bForceFullBuffer)
        {
            bBufferPartialFill = QP_FALSE;
        }
        else
        {
            bBufferPartialFill = QP_TRUE;
        }
        break;

    case COMPONENT_PORT_COMP_AUD_OUT:
        dir = PORT_DIR_READ;
        dwOpenType = QPMPGCODEC_ENC_AES_OUT;
        frame_nbs = QPCODECLIB_MAX_DATA_ENTRIES / 4;
        frame_size = 8 * 1024;
        if (frame_size < PAGE_SIZE)
            frame_size = PAGE_SIZE;
        bBufferFrameAligned = QP_TRUE;
        bBufferPartialFill = QP_TRUE;
        break;

    case COMPONENT_PORT_INDEX_OUT:
        dir = PORT_DIR_READ;
        dwOpenType = QPMPGCODEC_ENC_INDEX_OUT;
        frame_nbs = 128;
        frame_size = 30 * 20;
        bBufferFrameAligned = QP_FALSE;
        bBufferPartialFill = QP_TRUE;
        break;

    case COMPONENT_PORT_YUV_OUT:
        dir = PORT_DIR_READ;
        dwOpenType = (COMPONENT_TYPE_ENCODE == This->m_type) ? QPMPGCODEC_ENC_YUV_OUT : QPMPGCODEC_DEC_YUV_OUT;
        pOpenFormat = &pOpenParam->format;
        frame_nbs = 4;
        frame_size = (pOpenParam->format.yuv.nWidth *
                pOpenParam->format.yuv.nHeight *
                pOpenParam->format.yuv.nBitCount) / 8
                ;
        frame_size = (frame_size + PAGE_SIZE - 1) & PAGE_MASK;
        bBufferPartialFill = QP_FALSE;
        bBufferFrameAligned = QP_TRUE;
        This->m_pDevice->v4l2_port_type = COMPONENT_PORT_YUV_OUT;
        pl330_trace(PL330_TRACE_PORT, "%s %dx%d bit count=%d frame size=%d\n",
                    __func__, pOpenParam->format.yuv.nWidth,
                    pOpenParam->format.yuv.nHeight,
                    pOpenParam->format.yuv.nBitCount,
                    frame_size);
        break;

    case COMPONENT_PORT_PCM_OUT:
    {
        pl330_trace(PL330_TRACE_PORT, "%s COMPONENT_PORT_PCM_OUT\n", __func__);
        dir = PORT_DIR_READ;
        dwOpenType = (COMPONENT_TYPE_ENCODE == This->m_type) ||
                (COMPONENT_TYPE_PCM == This->m_type) ?
                QPMPGCODEC_ENC_PCM_OUT : QPMPGCODEC_DEC_PCM_OUT;
        pOpenFormat = &pOpenParam->format;
        frame_nbs = 64;
        This->m_pDevice->audio_port_type = COMPONENT_PORT_PCM_OUT;

        if (COMPONENT_TYPE_ENCODE == This->m_type ||
                COMPONENT_TYPE_PCM == This->m_type)
        {
            IMpegCodec *pMpegCodec = CDevice_getMpegCodec(This->m_pDevice);
            QPCODEC_AUDIO_CONTROL_PARAM audCntl;
            QPCODEC_AUDIO_CONTROL_EX audCntlEx;
            TQP_UINT32 dwSize;

            pMpegCodec->Get(pMpegCodec,
                            (TQP_REFGUID) & PROPSETID_QPCODEC_CONTROL,
                            This->m_hTask,
                            QPCODEC_PROP_AUDIO_CONTROL_PARAM,
                            QP_NULL,
                            &audCntl,
                            &dwSize
                            );

            pMpegCodec->Get(pMpegCodec,
                            (TQP_REFGUID) & PROPSETID_QPCODEC_CONTROL,
                            This->m_hTask,
                            QPCODEC_PROP_AUDIO_CONTROL_EX,
                            QP_NULL,
                            &audCntlEx,
                            &dwSize
                            );

            frame_size = audCntlEx.sample_cnt * audCntl.ch_num * 2;
            frame_size = (frame_size + PAGE_SIZE - 1) & PAGE_MASK;
            bBufferPartialFill = QP_TRUE;
            pl330_trace(PL330_TRACE_PORT, "%s frame size=%d\n",
                        __func__, frame_size);

        }
        else
        {
            frame_size = PAGE_SIZE;
            bBufferPartialFill = QP_FALSE;
        }
        bBufferFrameAligned = QP_FALSE;
        break;
    }

    case COMPONENT_PORT_ACTIVITY_OUT:
        dir = PORT_DIR_READ;
        dwOpenType = QPMPGCODEC_ENC_ACTIVITY_OUT;
        pOpenFormat = &pOpenParam->format;
        frame_nbs = 4;
        frame_size = ((pOpenParam->format.activity.nWidth / 16) *
                (pOpenParam->format.activity.nHeight / 16) *
                pOpenParam->format.activity.nBitCount) / 8
                ;
        frame_size = (frame_size + PAGE_SIZE - 1) & PAGE_MASK;

        bBufferFrameAligned = QP_FALSE;
        bBufferPartialFill = QP_TRUE;
        break;

    case COMPONENT_PORT_PCM_IN:
        dir = PORT_DIR_WRITE;
        dwOpenType = QPMPGCODEC_ENC_PCM_IN;
#if defined(CONFIG_HOST_PLATFORM_SMP8644_REF) || defined(CONFIG_HOST_PLATFORM_SMP8654_REF)
        frame_nbs = QPCODECLIB_MAX_DATA_ENTRIES / 8;
        frame_size = 16 * 1024;
#else
        frame_nbs = QPCODECLIB_MAX_DATA_ENTRIES / 4;
        frame_size = 4 * 1024;
#endif
        if (frame_size < PAGE_SIZE)
            frame_size = PAGE_SIZE;
        bBufferFrameAligned = QP_FALSE;
        bBufferPartialFill = QP_FALSE;
        break;

    case COMPONENT_PORT_VIRTUAL_IN:
    case COMPONENT_PORT_VIRTUAL_OUT:
        dir = PORT_DIR_NONE;
        dwOpenType = QPMPGCODEC_CODEC_VIRTUAL;
        frame_nbs = 0;
        frame_size = 0;
        bBufferFrameAligned = QP_FALSE;
        bBufferPartialFill = QP_TRUE;
        break;

    default:
        LEAVE_CRITICAL(&This->m_Object)
        return (QPERR_PARMS);
    }

    // instantiate the port
    //
    if (!CPort_Constructor(pPort,
                           This,
                           OBJECT_CRITICAL_HEAVY,
                           pOpenParam->type,
                           dir,
                           dwOpenType,
                           pOpenFormat,
                           pOpenParam->memory,
                           (V4L2_MEMORY_MMAP == pOpenParam->memory) ? frame_nbs : QPCODECLIB_MAX_DATA_ENTRIES,
                           frame_size,
                           bBufferPartialFill,
                           bBufferFrameAligned,
                           bSupportUserPtr
                           ))
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                     "%s CPort_Constructor() failed\n",
                     __FUNCTION__
                     ));
        LEAVE_CRITICAL(&This->m_Object)
        return (QPERR_FAIL);
    }

    qpStatus = CPort_Create(pPort);
    if (!QP_SUCCESS(qpStatus))
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                     "%s CPort_Create() failed status(%d)\n",
                     __FUNCTION__,
                     qpStatus
                     ));
        CPort_Destructor(pPort);
    }

    LEAVE_CRITICAL(&This->m_Object)

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s status(%d) ]\n",
                 __FUNCTION__,
                 qpStatus
                 ));

    return (qpStatus);
}

static EQP_Errors CComponent_PortClose(CComponent *This,
                                       COMPONENT_PORT_TYPE type
                                       )
{
    CPort *pPort;

    pl330_trace(PL330_TRACE_PORT, "%s port[%d] close\n",
                __func__, type);
    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s\n",
                 __FUNCTION__
                 ));

    if ((type >= MAX_COMPONENT_PORT) ||
            (type < 0)
            )
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                     "%s invalid type(%d)\n",
                     __FUNCTION__,
                     type
                     ));
        return (QPERR_PARMS);
    }

    ENTER_CRITICAL(&This->m_Object)

    pPort = &This->m_Ports[type];

    if (pPort->m_valid)
    {
        CPort_Close(pPort);
        CPort_Destructor(pPort);
    }

    LEAVE_CRITICAL(&This->m_Object)
    return (QPERR_SUCCESS);
}

int ccomponent_port_open(CComponent *This, PCOMPONENT_PORT_OPEN pOpenParam)
{
    return CComponent_PortOpen(This, pOpenParam);
}

int ccomponent_port_close(CComponent *This, COMPONENT_PORT_TYPE type)
{
    return CComponent_PortClose(This, type);
}

static const struct v4l2_queryctrl video_ctrls[] = {
    {
        .id = V4L2_CID_MPEG_VIDEO_BITRATE,
        .name = "Video Bitrate",
        .minimum = 50000,
        .maximum = 50000000,
        .step = 50000,
        .default_value = 16000000,
        .type = V4L2_CTRL_TYPE_INTEGER,
    }
};

#define V4L2_CID_PRIVATE_LASTP1  (V4L2_CID_PRIVATE_BASE + 0)

static const unsigned int CTRLS = ARRAY_SIZE(video_ctrls);

static const struct v4l2_queryctrl *ctrl_by_id(IFrontend *frontend,unsigned int id)
{
    unsigned int i;

    for (i = 0; i < CTRLS; i++)
        if (video_ctrls[i].id == id)
            return video_ctrls + i;
    for (i = 0; i < frontend->v4l2_queryctrls_num; i++)
        if (frontend->v4l2_queryctrls[i].id == id)
            return &frontend->v4l2_queryctrls[i];
#if 0
    switch (id)
    {
    case V4L2_CID_MPEG_VIDEO_BITRATE:
        return v4l2_ctrl_query_fill(ctrl,
                                    1000000, 50000000, 1000000, 16000000);
    }
#endif
    return NULL;
}

static int ccomponent_queryctrl(CComponent *This, void *arg)
{
    struct v4l2_queryctrl *ctrl = (struct v4l2_queryctrl *) arg;
    const struct v4l2_queryctrl *queryctrl;
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    //IVideoDecoder *video_dec_if = CDevice_getVideoDecoder(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend; 
#if 0
    if ((ctrl->id < V4L2_CID_BASE ||
            ctrl->id >= V4L2_CID_LASTP1) &&
            (ctrl->id < V4L2_CID_PRIVATE_BASE ||
            ctrl->id >= V4L2_CID_PRIVATE_LASTP1))
        return -EINVAL;
#endif
    queryctrl = ctrl_by_id(frontend,ctrl->id);

    if (queryctrl == NULL)
        return -EINVAL;

    *ctrl = *queryctrl;

    return 0;
}

static int ccomponent_g_ctrl(CComponent *This, void *arg)
{
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    //	IVideoDecoder *video_dec_if = CDevice_getVideoDecoder(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    CQLCodec *codec = codeclib->m_pMpgCodec;
    ENCODER_SETTING *encoder_setting = &codec->encoder_setting;
    
    IFrontend *frontend = &codeclib->m_pVidDecoder->m_frontend;
    int ret;

    struct v4l2_control *ctrl = (struct v4l2_control *) arg;
    const struct v4l2_queryctrl *supported_ctrl;
    supported_ctrl = ctrl_by_id(frontend,ctrl->id);

    //u32 value = 0;

    if (supported_ctrl == NULL)
        return -EINVAL;
    switch(ctrl->id)
    {
        case V4L2_CID_MPEG_VIDEO_BITRATE:
		 ctrl->value = encoder_setting->bit_rate * 1000;
        return 0;
    }
    
    if((ret=frontend_g_ctrl(frontend,ctrl))!=-1)
    {
        return ret;
    }
    else
    {
        pl330_trace(PL330_TRACE_VIDIOC, "g_ctrl not implement\n");
        return -EIO;
    }

    return 0;
}

static int check_range(const struct v4l2_queryctrl *queryctrl, struct v4l2_control *ctrl)
{
    if (ctrl->value > queryctrl->maximum ||
            ctrl->value < queryctrl->minimum)
        return -ERANGE;
    return 0;
}

static int ccomponent_s_ctrl(CComponent *This, void *arg)
{
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    //IVideoDecoder *video_dec_if = CDevice_getVideoDecoder(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    CQLCodec *codec = codeclib->m_pMpgCodec;
    ENCODER_SETTING *encoder_setting = &codec->encoder_setting;
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend; 
    int err;
    int ret;
    struct v4l2_control *ctrl = (struct v4l2_control *) arg;
    const struct v4l2_queryctrl *supported_ctrl;
    supported_ctrl = ctrl_by_id(frontend,ctrl->id);

    if (supported_ctrl == NULL)
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s unknown control %d\n",
                    __func__, ctrl->id);
        return -EINVAL;
    }

    err = check_range(supported_ctrl, ctrl);
    if (err)
        return err;
    switch(ctrl->id)
    {
        case V4L2_CID_MPEG_VIDEO_BITRATE:
		encoder_setting->bit_rate = ctrl->value / 1000;
		pl330_set_config(This , "RateControl-AvgBitRate", &encoder_setting->bit_rate, true);
		pl330_set_config(This , "VbrBitRate-MinBitRate", &encoder_setting->bit_rate, true);
		pl330_set_config(This , "VbrBitRate-MaxBitRate", &encoder_setting->bit_rate, true);

        return 0;
    }
    if((ret=frontend_s_ctrl(frontend,ctrl))!=-1)
    {
        return ret;
    }
    else
    {
        pl330_trace(PL330_TRACE_VIDIOC, "s_ctrl not implement\n");
        return -EIO;
    }

    return 0;
}

static int ccomponent_enum_input(CComponent *This, void *arg)
{
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    //CQLCodec 		*pQLCodec=codeclib->m_pMpgCodec;
    //CADV7604 *adv7604 = (CADV7604 *) codeclib->m_pVidDecoder;
    //	struct adv7604_state *state = &adv7604->state;

    struct v4l2_input *input = (struct v4l2_input *) arg;
    unsigned int n = input->index;
    const frontend_input_info_t *input_info=NULL;

//    static const char *input_name[] = {
//        [C353_HDMI] = "HDMI",
//        [C353_VGA] = "VGA",
//    };
//
//    if (n >= C353_INPUT_NUM)
//    {
//        return -EINVAL;
//    }
    input_info=frontend_g_input_info(frontend,n);
    if(input_info==NULL)
        return -EINVAL;
    
    strcpy(input->name, input_info->name);
    input->status=0;
    switch(input_info->status)
    {
    case FRONTEND_NO_SIGNAL:
        input->status=V4L2_IN_ST_NO_SIGNAL;
        break;
    case FRONTEND_NO_SYNC:
        input->status=V4L2_IN_ST_NO_SYNC;
        break;
    case FRONTEND_LOCK:
        input->status=0;
        break;
    default:    
        input->status=V4L2_IN_ST_NO_SIGNAL;
        break;
    }
    
       
    

    

    return 0;
}

static int ccomponent_g_input(CComponent *This, void *arg)
{
    unsigned int *input = (unsigned int *) arg;

    // get input
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    

    *input=frontend_g_input(frontend);
    

    return 0;
}

static int ccomponent_s_input(CComponent *This, void *arg)
{
    unsigned int *input = (unsigned int *) arg;

    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    int ret;
    const frontend_input_info_t *input_info=NULL;

    if (atomic_read(&This->m_pDevice->stream_started))
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s stream started\n",
                    __func__);
        //return -EBUSY; //external ioctl access
    }

    input_info=frontend_g_input_info(frontend,*input);
    if (input_info==NULL)
        return -EINVAL;

    

    if((ret=frontend_s_input(frontend,*input)))
    {
        return ret;
    }
    

    return 0;
}

static struct v4l2_capability pl330_capability = {
    .driver = MODULE_NAME,
    .card = "AVerMedia C353",
    .bus_info = "pcie",
    .version = KERNEL_VERSION(1, 4, 8120),
    .capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_STREAMING | V4L2_CAP_DEVICE_CAPS,
    .device_caps = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_STREAMING
};

static int ccomponent_querycap(CComponent *This, void *arg)
{
    struct v4l2_capability *video_cap = (struct v4l2_capability *) arg;
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    PL330_boards_info *board_info = &codeclib->m_InitData.board_info;

    memcpy(video_cap, &pl330_capability, sizeof (struct v4l2_capability));
    if(board_info && board_info->boardname)
        snprintf(video_cap->card,sizeof(video_cap->card),"AVerMedia %s",board_info->boardname);
        //ioctl get board info
    return 0;
}

static struct pl330_format formats[] = {
    {
        .name = "H264 Encoded Stream",
        .pixelformat = V4L2_PIX_FMT_MPEG,
        .depth = 16,
        .flags = V4L2_FMT_FLAG_COMPRESSED,
    },
    {
        .name = "YVU 4:2:0 12bpp",
        .pixelformat = V4L2_PIX_FMT_YVU420,
        .depth = 8,
        .flags = 0,
    },
    {
        .name = "YUV 4:2:0 12bpp",
        .pixelformat = V4L2_PIX_FMT_YUV420,
        .depth = 8,
        .flags = 0,
    },
    {
        .name = "NV12 4:2:0 12bpp",
        .pixelformat = V4L2_PIX_FMT_NV12,
        .depth = 8,
        .flags = 0,
    }
};

const static unsigned int frame_interval_50[] ={
    FRAME_INTERVAL_50,
};

const static unsigned int frame_interval_60[] ={
    FRAME_INTERVAL_60,
};

const static unsigned int frame_interval_50_60[] ={
    FRAME_INTERVAL_50,
    FRAME_INTERVAL_60,
};

const static unsigned int frame_interval_60_75[] ={
    FRAME_INTERVAL_60,
    FRAME_INTERVAL_75,
};

const static unsigned int frame_interval_60_85[] ={
    FRAME_INTERVAL_60,
    FRAME_INTERVAL_72,
    FRAME_INTERVAL_75,
    FRAME_INTERVAL_85,
};

const static unsigned int frame_interval_60_70_85[] ={
    FRAME_INTERVAL_60,
    FRAME_INTERVAL_70,
    FRAME_INTERVAL_75,
    FRAME_INTERVAL_85,
};

const static unsigned int frame_interval_1920x1080[] ={
    FRAME_INTERVAL_15,
    FRAME_INTERVAL_24,
    FRAME_INTERVAL_25,
    FRAME_INTERVAL_30,
    FRAME_INTERVAL_50,
    FRAME_INTERVAL_60
};

const static unsigned int frame_interval_all[] ={
    FRAME_INTERVAL_15,
    FRAME_INTERVAL_24,
    FRAME_INTERVAL_25,
    FRAME_INTERVAL_30,
    FRAME_INTERVAL_50,
    FRAME_INTERVAL_60,
    FRAME_INTERVAL_75
};


const static struct pl330_frame frame_formats[] ={
    {
        .width = 640, .height = 480,
        .frame_type = ARRAY_SIZE(frame_interval_60),
        .frameinterval = frame_interval_60,
        .analog_frame_type = ARRAY_SIZE(frame_interval_60_85),
        .analog_frameinterval = frame_interval_60_85
    },
    {
        .width = 720, .height = 480,
        .frame_type = ARRAY_SIZE(frame_interval_60),
        .frameinterval = frame_interval_60,
        .analog_frame_type = 0,
        .analog_frameinterval = NULL
    },
    {
        .width = 720, .height = 576,
        .frame_type = ARRAY_SIZE(frame_interval_50),
        .frameinterval = frame_interval_50,
        .analog_frame_type = 0,
        .analog_frameinterval = NULL
    },
    {
        .width = 800, .height = 600,
        .frame_type = ARRAY_SIZE(frame_interval_60),
        .frameinterval = frame_interval_60,
        .analog_frame_type = ARRAY_SIZE(frame_interval_60_85),
        .analog_frameinterval = frame_interval_60_85
    },
    {
        .width = 1024, .height = 768,
        .frame_type = ARRAY_SIZE(frame_interval_60),
        .frameinterval = frame_interval_60,
        .analog_frame_type = ARRAY_SIZE(frame_interval_60_85),
        .analog_frameinterval = frame_interval_60_85

    },
    {
        .width = 1280, .height = 720,
        .frame_type = ARRAY_SIZE(frame_interval_50_60),
        .frameinterval = frame_interval_50_60,
        .analog_frame_type = ARRAY_SIZE(frame_interval_50_60),
        .analog_frameinterval = frame_interval_50_60
    },
    {
        .width = 1280, .height = 800,
        .frame_type = ARRAY_SIZE(frame_interval_60),
        .frameinterval = frame_interval_60,
        .analog_frame_type = ARRAY_SIZE(frame_interval_60),
        .analog_frameinterval = frame_interval_60
    },
    {
        .width = 1280, .height = 1024,
        .frame_type = ARRAY_SIZE(frame_interval_60),
        .frameinterval = frame_interval_60,
        .analog_frame_type = ARRAY_SIZE(frame_interval_60),
        .analog_frameinterval = frame_interval_60
    },
    {
        .width = 1360, .height = 768,
        .frame_type = ARRAY_SIZE(frame_interval_60),
        .frameinterval = frame_interval_60,
        .analog_frame_type = 0,
        .analog_frameinterval = NULL,
    },
    {
        .width = 1440, .height = 900,
        .frame_type = ARRAY_SIZE(frame_interval_60),
        .frameinterval = frame_interval_60,
        .analog_frame_type = ARRAY_SIZE(frame_interval_60),
        .analog_frameinterval = frame_interval_60
    },
#if 0
    {
        .width = 1400, .height = 1050,
        .frame_type = V4L2_FRMIVAL_TYPE_DISCRETE
    },
#endif
    {
        .width = 1920, .height = 1080,
        .frame_type = ARRAY_SIZE(frame_interval_1920x1080),
        .frameinterval = frame_interval_1920x1080,
        .analog_frame_type = ARRAY_SIZE(frame_interval_1920x1080),
        .analog_frameinterval = frame_interval_1920x1080
    },
    {
        .width = 1680, .height = 1050,
        .frame_type = ARRAY_SIZE(frame_interval_1920x1080),
        .frameinterval = frame_interval_1920x1080,
        .analog_frame_type = ARRAY_SIZE(frame_interval_1920x1080),
        .analog_frameinterval = frame_interval_1920x1080
    }
};

static int ccomponent_enum_fmt(CComponent *This, void *arg)
{
    struct v4l2_fmtdesc *fmtd = (struct v4l2_fmtdesc *) arg;

    //	ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    //CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);

    if (fmtd->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
        return -EINVAL;

    if (ARRAY_SIZE(formats) <= fmtd->index)
        return -EINVAL;

    fmtd->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    strlcpy(fmtd->description, formats[fmtd->index].name, sizeof (fmtd->description));
    fmtd->pixelformat = formats[fmtd->index].pixelformat;
    fmtd->flags = formats[fmtd->index].flags;

    return 0;
}

static int ccomponent_g_fmt(CComponent *This, void *arg)
{
    struct v4l2_format *f = (struct v4l2_format *) arg;
    //	struct v4l2_pix_format *pix = &f->fmt.pix;

    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    
    const frontend_frameinfo_t *frameinfo;
    u32 pixfmt;
    int out_width,out_height;
    


    if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
        return -EINVAL;

    
    frameinfo=frontend_g_frameinfo(frontend);
    
    
    if(frameinfo==NULL)
    {
       out_width=0;
       out_height=0;
    }else
    {
        if (This->out_width == 0 || This->out_height == 0)
        {
            out_width=frameinfo->width;
            out_height=frameinfo->height;
        }else
        {
            out_width=This->out_width;
            out_height=This->out_height;
        }
    }
        
    
    
    f->fmt.pix.width = out_width;
    f->fmt.pix.height = out_height;

    /* interlaced or progressive */
    
    f->fmt.pix.field = V4L2_FIELD_NONE; 
    if(frameinfo)
    {
        if(frameinfo->is_interlace)
            f->fmt.pix.field = V4L2_FIELD_INTERLACED;
        
    }
    pixfmt=frontend_g_pixfmt(frontend);
    
    switch (pixfmt)
    {
    case V4L2_PIX_FMT_MPEG:
        f->fmt.pix.pixelformat = V4L2_PIX_FMT_MPEG;
        f->fmt.pix.sizeimage = MPEG_FRAME_SIZE;
        f->fmt.pix.bytesperline = 0;
        break;
    case V4L2_PIX_FMT_NV12:
        f->fmt.pix.pixelformat = V4L2_PIX_FMT_NV12;
        f->fmt.pix.bytesperline = formats[1].depth * out_width / 8;
        f->fmt.pix.sizeimage = (f->fmt.pix.bytesperline * f->fmt.pix.height + PAGE_SIZE - 1) & PAGE_MASK;
        break;
    case V4L2_PIX_FMT_YUV420:
        f->fmt.pix.pixelformat = V4L2_PIX_FMT_YUV420;
        f->fmt.pix.bytesperline = formats[1].depth * out_width / 8;
        f->fmt.pix.sizeimage = (f->fmt.pix.bytesperline * f->fmt.pix.height + PAGE_SIZE - 1) & PAGE_MASK;
        break;
    default:
    case V4L2_PIX_FMT_YVU420:
        f->fmt.pix.pixelformat = V4L2_PIX_FMT_YVU420;
        f->fmt.pix.bytesperline = formats[1].depth * out_width / 8;
        f->fmt.pix.sizeimage = (f->fmt.pix.bytesperline * f->fmt.pix.height + PAGE_SIZE - 1) & PAGE_MASK;
        break;
    }
    f->fmt.pix.colorspace = V4L2_COLORSPACE_SMPTE170M;
    f->fmt.pix.priv = 0;

    pl330_trace(PL330_TRACE_VIDIOC, "%s port_type=%d sizeimage=%d\n",
                __func__, This->m_pDevice->v4l2_port_type, f->fmt.pix.sizeimage);
    /*
            if(state->timings.bt.reserved[0] & V4L2_DV_BT_STD_CEA861) {
                    f->fmt.pix.colorspace = (state->timings.bt.height <= 576) ?
                            V4L2_COLORSPACE_SMPTE170M : V4L2_COLORSPACE_REC709;
            }
     */

    return 0;
}

static int ccomponent_try_fmt(CComponent *This, void *arg)
{
    struct v4l2_format *f = (struct v4l2_format *) arg;
    //	struct v4l2_pix_format *pix = &f->fmt.pix;

    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    const frontend_frameinfo_t *frameinfo;
    int i;
    __u16 rw, rh;
    __u8 *fcc;
    struct pl330_format *format=NULL;
    int width_bound=1920,height_bound=1080;

    fcc = (__u8 *) & f->fmt.pix.pixelformat;
    pl330_trace(PL330_TRACE_VIDIOC, "%s %ux%u 0x%08x (%c%c%c%c)\n", __func__,
                f->fmt.pix.width,
                f->fmt.pix.height,
                f->fmt.pix.pixelformat, fcc[0], fcc[1], fcc[2], fcc[3]);

    for (i = 0; i < ARRAY_SIZE(formats); i++)
    {
        if (f->fmt.pix.pixelformat == formats[i].pixelformat)
        {
            format = &formats[i];
            break;
        }
    }


    if (format==NULL)
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s Unsupported format 0x%08x.\n",
                    __func__,
                    f->fmt.pix.pixelformat);
        return -EINVAL;
    }
    frameinfo=frontend_g_frameinfo(frontend);
   
    rw = f->fmt.pix.width;
    rh = f->fmt.pix.height;

    
    if(frameinfo)
    {
        width_bound=frameinfo->width;
        height_bound=frameinfo->height;
    }
    if(rw==0 || rh==0)
    {
        f->fmt.pix.width=width_bound;
        f->fmt.pix.height=height_bound;
    }   
    if(rw > width_bound)
        f->fmt.pix.width=width_bound;
    if(rh > height_bound)
        f->fmt.pix.height=height_bound;


    pl330_trace(PL330_TRACE_VIDIOC, "%s %ux%u\n", __func__,
                f->fmt.pix.width,
                f->fmt.pix.height);

    /* interlaced or progressive */
    f->fmt.pix.field = V4L2_FIELD_NONE;
    if(frameinfo)
        if(frameinfo->is_interlace)
            f->fmt.pix.field =  V4L2_FIELD_INTERLACED ;
    f->fmt.pix.pixelformat = format->pixelformat;
    if(frontend_g_pixfmt(frontend)==0)
        frontend_s_pixfmt(frontend,f->fmt.pix.pixelformat);

    switch (format->pixelformat)
    {
    case V4L2_PIX_FMT_MPEG:
        f->fmt.pix.sizeimage = MPEG_FRAME_SIZE;
        f->fmt.pix.bytesperline = 0;
        break;
    case V4L2_PIX_FMT_YUV420:
        f->fmt.pix.bytesperline = formats[1].depth * f->fmt.pix.width / 8;
        f->fmt.pix.sizeimage = (f->fmt.pix.bytesperline * f->fmt.pix.height + PAGE_SIZE - 1) & PAGE_MASK;
        break;
    case V4L2_PIX_FMT_NV12:
        f->fmt.pix.bytesperline = formats[1].depth * f->fmt.pix.width / 8;
        f->fmt.pix.sizeimage = (f->fmt.pix.bytesperline * f->fmt.pix.height + PAGE_SIZE - 1) & PAGE_MASK;
        break;
    case V4L2_PIX_FMT_YVU420:
        f->fmt.pix.bytesperline = formats[1].depth * f->fmt.pix.width / 8;
        f->fmt.pix.sizeimage = (f->fmt.pix.bytesperline * f->fmt.pix.height + PAGE_SIZE - 1) & PAGE_MASK;
        break;
    }

    f->fmt.pix.colorspace = V4L2_COLORSPACE_SMPTE170M;
    f->fmt.pix.priv = 0;

    return 0;
}

static int ccomponent_s_fmt(CComponent *This, void *arg)
{
    struct v4l2_format *f = (struct v4l2_format *) arg;
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    int ret=0;

    pl330_trace(PL330_TRACE_VIDIOC, "%s %dx%d 0x%08x\n", __func__,
                f->fmt.pix.width,
                f->fmt.pix.height,
                f->fmt.pix.pixelformat);
    
    if (atomic_read(&This->m_pDevice->stream_started))
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s stream started\n",
                    __func__);
        return -EBUSY;
    }
    if((ret = ccomponent_try_fmt(This, arg))==0)
    {
        
        This->out_width=f->fmt.pix.width;
        This->out_height=f->fmt.pix.height;
        
        frontend_s_pixfmt(frontend,f->fmt.pix.pixelformat);
        
    }
    

    return ret;
}

static int component_enum_framesizes(CComponent *This, void *arg)
{
    struct v4l2_frmsizeenum *fsize = (struct v4l2_frmsizeenum *) arg;
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    const struct pl330_frame *frame = NULL;
    const frontend_frameinfo_t *frameinfo;
    
    int i;

    frameinfo=frontend_g_frameinfo(frontend);
    
    for (i = 0; i < ARRAY_SIZE(formats); i++)
    {
        if (fsize->pixel_format ==
                formats[i].pixelformat)
        {
            break;
        }
    }

    if (i >= ARRAY_SIZE(formats))
        return -EINVAL;
#if 1
    if (fsize->index >= ARRAY_SIZE(frame_formats))
    {
        return -EINVAL;
    }

    frame = &frame_formats[fsize->index];
    fsize->type = V4L2_FRMSIZE_TYPE_DISCRETE;
    fsize->discrete.width = frame->width;
    fsize->discrete.height = frame->height;
#else
    if(frameinfo==NULL)
    {
        if (fsize->index >= ARRAY_SIZE(frame_formats))
        {
            return -EINVAL;
        }

        frame = &frame_formats[fsize->index];
        fsize->type = V4L2_FRMSIZE_TYPE_DISCRETE;
        fsize->discrete.width = frame->width;
        fsize->discrete.height = frame->height;
    }else
    {
        if(fsize->index> 0)
        {
            return -EINVAL;
        }
        fsize->type = V4L2_FRMSIZE_TYPE_DISCRETE;
        fsize->discrete.width = frameinfo->width;
        fsize->discrete.height = frameinfo->height;    
    }

#endif      
            
    return 0;
}

static int component_enum_frameintervals(CComponent *This, void *arg)
{
    struct v4l2_frmivalenum *fival = arg;
    //	struct pl330_frame *frame = NULL;

    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    //        CVideoDecoder *decoder = codeclib->m_pVidDecoder;
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;

    int i;
    int input;

    pl330_trace(PL330_TRACE_VIDIOC, "%s fival->index=%u enum_frameintervals %dx%d\n",
                __func__, fival->index, fival->width, fival->height);

    for (i = 0; i < ARRAY_SIZE(formats); i++)
    {
        if (fival->pixel_format ==
                formats[i].pixelformat)
            break;
    }

    if (i >= ARRAY_SIZE(formats))
        return -EINVAL;

    for (i = 0; i < ARRAY_SIZE(frame_formats); i++)
    {
        if (fival->width == frame_formats[i].width &&
                fival->height == frame_formats[i].height)
            break;
    }

    if (i >= ARRAY_SIZE(frame_formats))
        return -EINVAL;
    input=frontend_g_input(frontend);
    if (frame_formats[i].frame_type)
    {
        if (fival->index >= ((input == C353_HDMI) ?
                frame_formats[i].frame_type :
                frame_formats[i].analog_frame_type))
            return -EINVAL;
        fival->type = V4L2_FRMIVAL_TYPE_DISCRETE;
        fival->discrete.numerator = 1;

        if (input == C353_HDMI)
            fival->discrete.denominator =
                FRAMEINVERVAL(frame_formats[i].frameinterval[fival->index]);
        else
            fival->discrete.denominator =
                FRAMEINVERVAL(frame_formats[i].analog_frameinterval[fival->index]);
    }
    else
        return -EINVAL;

    return 0;
}

static int component_g_parm(CComponent *This, void *arg)
{
    struct v4l2_streamparm *parm = (struct v4l2_streamparm *) arg;
    //uint32_t numerator, denominator;
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    const frontend_frameinfo_t *frameinfo=NULL;
    int framerate;
    

    if (parm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
    {
        return -EINVAL;
    }
    
    framerate=frontend_g_framerate(frontend);
    if(framerate==0)
    {
        frameinfo=frontend_g_frameinfo(frontend);
        if(frameinfo)
            framerate=frameinfo->framerate;
        else
            framerate=30;
 
    }
    parm->parm.capture.capability = V4L2_CAP_TIMEPERFRAME;
    parm->parm.capture.capturemode = 0;
    parm->parm.capture.timeperframe.numerator = 1;
    parm->parm.capture.timeperframe.denominator = framerate;
    parm->parm.capture.extendedmode = 0;
    parm->parm.capture.readbuffers = 0;

    return 0;
}

static int component_s_parm(CComponent *This, void *arg)
{
    struct v4l2_streamparm *parm = (struct v4l2_streamparm *) arg;
    //	uint32_t numerator, denominator;
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    u32 _frame_rate;
    u32 io_frame_rate;

    if (parm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
    {
        return -EINVAL;
    }

//    if (atomic_read(&This->m_pDevice->stream_started))
//    {
//        return -EBUSY;
//    }

    pl330_trace(PL330_TRACE_VIDIOC, "%s set framerate= %u/%u\n",
                __func__, parm->parm.capture.timeperframe.denominator,
                parm->parm.capture.timeperframe.numerator);

    //printk("%s current framerate= %u\n",
    //            __func__, frontend_g_framerate(frontend));
                
    io_frame_rate = frontend_g_framerate(frontend);
    
 
    if (parm->parm.capture.timeperframe.denominator!=0)
    {
    
        _frame_rate =
            parm->parm.capture.timeperframe.denominator / parm->parm.capture.timeperframe.numerator;
            
  
        frontend_s_framerate(frontend,_frame_rate);
        printk("%s set framerate= %u\n", __func__, _frame_rate);
    }
    else
    {
		frontend_s_framerate(frontend,io_frame_rate);
		printk("%s set framerate= %u\n", __func__, io_frame_rate);
	}
    
    pl330_trace(PL330_TRACE_VIDIOC, "%s framerate= %u\n", __func__, _frame_rate);

    return 0;
}

static int ccomponent_dqevent(CComponent *This, void *arg)
{
    return 0;
}

static int ccomponent_streamon(CComponent *This)
{
    EQP_Errors ret = QPERR_FAIL;
    //	CPort				*pPort;
    //	struct v4l2_buffer buf;
    //	EQP_Errors			err;
    //	int i;
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    CQLCodec *codec = codeclib->m_pMpgCodec;
    ENCODER_SETTING *encoder_setting = &codec->encoder_setting;
    //IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend; 

    switch (This->m_dummy_type)
    {
        /* for Video */
    case COMPONENT_TYPE_ENCODE:
    {
        
        ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
        CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
        IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
        TQP_UINT32 pixfmt;
        
        
            
        
        pixfmt=frontend_g_pixfmt(frontend);
        
        if(pixfmt== V4L2_PIX_FMT_MPEG)
        {
            //type = COMPONENT_PORT_COMP_OUT;
            pl330_set_config(This, "SystemControl-StreamType", "ts", false);
            pl330_set_config(This, "SysFunction-Video", "h264", false);
        }else if((pixfmt==V4L2_PIX_FMT_YVU420) || (pixfmt==V4L2_PIX_FMT_YUV420) || (pixfmt == V4L2_PIX_FMT_NV12))
        {
            //type = COMPONENT_PORT_YUV_OUT;
            pl330_set_config(This, "SystemControl-StreamType", "raw", false);
            pl330_set_config(This, "SysFunction-Video", "passthrough", false);
        }


        pl330_default_config(This);
        //LOG_NUM(encoder_setting->bit_rate);
		pl330_set_config(This , "RateControl-AvgBitRate", &encoder_setting->bit_rate, true);
		pl330_set_config(This , "VbrBitRate-MinBitRate", &encoder_setting->bit_rate, true);
		pl330_set_config(This , "VbrBitRate-MaxBitRate", &encoder_setting->bit_rate, true);
        //ccomponent_stop_index(This, type);
        frontend_start_streaming(&codeclib->m_pVidDecoder->m_frontend,true);
        //ret = CComponent_Start_Index(This, This->m_pDevice->v4l2_port_type);
        ret=QPERR_SUCCESS;
        atomic_set(&This->m_pDevice->stream_started, 1);
    }
        break;
        /* open PCM port */
    case COMPONENT_TYPE_PCM:
        ret = CComponent_Start_Index(This, COMPONENT_PORT_PCM_OUT);
        break;
    default:
        pl330_trace(PL330_TRACE_VIDIOC, "%s "
                    "not support component dummy type\n", __func__);
    }

    return ret;
}

static int ccomponent_streamoff(CComponent *This)
{
    EQP_Errors ret = QPERR_FAIL;
    int port_type;

    switch (This->m_dummy_type)
    {
        /* for Video */
    case COMPONENT_TYPE_ENCODE:
    {
        ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
        CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
        IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
        frontend_start_streaming(frontend,false);
        ret = CComponent_Stop_Index(This, This->m_pDevice->v4l2_port_type);
        port_type = This->m_pDevice->v4l2_port_type;
        
        atomic_set(&This->m_pDevice->stream_started, 0);
    }
        break;
        /* stop PCM port */
    case COMPONENT_TYPE_PCM:
        ret = CComponent_Stop_Index(This, COMPONENT_PORT_PCM_OUT);
        port_type = COMPONENT_PORT_PCM_OUT;
        break;
    default:
        pl330_trace(PL330_TRACE_VIDIOC, "%s "
                    "not support component dummy type\n", __func__);
        return ret;
    }

    pl330_port_close(This, port_type);

    return ret;
}


static int component_audio_rate(CComponent *This, void *arg)
{
    unsigned int *rates = (unsigned int *) arg;
    ICodecLib *codeclib_if = CDevice_getCodecLib(This->m_pDevice);
    //	IVideoDecoder *video_dec_if = CDevice_getVideoDecoder(This->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    //	CQLCodec 		*pQLCodec = codeclib->m_pMpgCodec;
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    const frontend_audioinfo_t *audioinfo=frontend_g_audioinfo(frontend);
    //	EQP_Errors	qpStatus;
    //	TQP_UINT32	dwRegister;
    //	TQP_UINT32	dwMask;


    
    pl330_trace(PL330_TRACE_IOCTL, "%s audio samplerate=%u\n",
                __func__, audioinfo->sample_rate);
    *rates = audioinfo->sample_rate;

#if 0
    /* Set GPIO 14, 15 use for I2C to INPUT mode */
    /* set gpio direction */
    qpStatus = QPHCI_RegisterRead(&pQLCodec->m_hci,
                                  0x610, &dwRegister);

    dwMask = ~(1 << 14);
    dwRegister = (dwRegister & dwMask) |
            (GPIO_INPUT << 14);

    dwMask = ~(1 << 15);
    dwRegister = (dwRegister & dwMask) |
            (GPIO_INPUT << 15);
    QPHCI_RegisterWrite(&pQLCodec->m_hci,
                        0x610, dwRegister);

    return 0;
#endif

    return 0;
}

static EQP_Errors CComponent_Start_Index(CComponent *This, COMPONENT_PORT_TYPE type)
{
    EQP_Errors qpStatus = QPERR_SUCCESS;
    CPort *pPort;
    //	int			i;

    pl330_trace(PL330_TRACE_VIDIOC, "%s type=%u\n", __func__, type);
    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s [\n",
                 __FUNCTION__
                 ));

    ENTER_CRITICAL(&This->m_Object)

    // from stop to acquire
    //
    pPort = &This->m_Ports[type];
    if (pPort->m_valid)
    {
        if (PORT_STATE_STOP == pPort->m_State)
        {
            qpStatus = CPort_SetState(pPort,
                                      PORT_STATE_ACQUIRE,
                                      PORT_STATE_STOP
                                      );
            if (!QP_SUCCESS(qpStatus))
            {
                goto CComponent_Start_Exit;
            }
        }
    }
    // from acquire to pause
    //
    pPort = &This->m_Ports[type];
    if (pPort->m_valid)
    {
        if (PORT_STATE_ACQUIRE == pPort->m_State)
        {
            qpStatus = CPort_SetState(pPort,
                                      PORT_STATE_PAUSE,
                                      PORT_STATE_ACQUIRE
                                      );
            if (!QP_SUCCESS(qpStatus))
            {
                goto CComponent_Start_Exit;
            }
        }
    }

    // from pause to run
    //
    pPort = &This->m_Ports[type];
    if (pPort->m_valid)
    {
        qpStatus = CPort_SetState(pPort,
                                  PORT_STATE_RUN,
                                  pPort->m_State
                                  );
        if (!QP_SUCCESS(qpStatus))
        {
            goto CComponent_Start_Exit;
        }
    }

CComponent_Start_Exit:

    if (!QP_SUCCESS(qpStatus))
    {
        pPort = &This->m_Ports[type];
        if (pPort->m_valid)
        {
            CPort_SetState(pPort,
                           PORT_STATE_STOP,
                           pPort->m_State
                           );
        }
    }

    LEAVE_CRITICAL(&This->m_Object)

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s status(%d) ]\n",
                 __FUNCTION__,
                 qpStatus
                 ));

    return (qpStatus);
}

static EQP_Errors CComponent_Start(CComponent *This)
{
    EQP_Errors qpStatus = QPERR_SUCCESS;
    CPort *pPort;
    int i;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s [\n",
                 __FUNCTION__
                 ));

    ENTER_CRITICAL(&This->m_Object)

            // from stop to acquire
            //
    for (i = 0; i < MAX_COMPONENT_PORT; i++)
    {
        pPort = &This->m_Ports[i];
        if (pPort->m_valid)
        {
            if (PORT_STATE_STOP == pPort->m_State)
            {
                qpStatus = CPort_SetState(pPort,
                                          PORT_STATE_ACQUIRE,
                                          PORT_STATE_STOP
                                          );
                if (!QP_SUCCESS(qpStatus))
                {
                    goto CComponent_Start_Exit;
                }
            }
        }
    }

    // from acquire to pause
    //
    for (i = 0; i < MAX_COMPONENT_PORT; i++)
    {
        pPort = &This->m_Ports[i];
        if (pPort->m_valid)
        {
            if (PORT_STATE_ACQUIRE == pPort->m_State)
            {
                qpStatus = CPort_SetState(pPort,
                                          PORT_STATE_PAUSE,
                                          PORT_STATE_ACQUIRE
                                          );
                if (!QP_SUCCESS(qpStatus))
                {
                    goto CComponent_Start_Exit;
                }
            }
        }
    }

    // from pause to run
    //
    for (i = 0; i < MAX_COMPONENT_PORT; i++)
    {
        pPort = &This->m_Ports[i];
        if (pPort->m_valid)
        {
            qpStatus = CPort_SetState(pPort,
                                      PORT_STATE_RUN,
                                      pPort->m_State
                                      );
            if (!QP_SUCCESS(qpStatus))
            {
                goto CComponent_Start_Exit;
            }
        }
    }

CComponent_Start_Exit:

    if (!QP_SUCCESS(qpStatus))
    {
        for (i = 0; i < MAX_COMPONENT_PORT; i++)
        {
            pPort = &This->m_Ports[i];
            if (pPort->m_valid)
            {
                CPort_SetState(pPort,
                               PORT_STATE_STOP,
                               pPort->m_State
                               );
            }
        }
    }

    LEAVE_CRITICAL(&This->m_Object)

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s status(%d) ]\n",
                 __FUNCTION__,
                 qpStatus
                 ));

    return (qpStatus);
}

static EQP_Errors CComponent_Stop_Index(CComponent *This, COMPONENT_PORT_TYPE type)
{
    CPort *pPort;
    //	int		i;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s [\n",
                 __FUNCTION__
                 ));

    ENTER_CRITICAL(&This->m_Object)

    // stop all ports
    //
    pPort = &This->m_Ports[type];
    if (pPort->m_valid)
    {
        CPort_SetState(pPort,
                       PORT_STATE_STOP,
                       pPort->m_State
                       );
    }

    LEAVE_CRITICAL(&This->m_Object)

            // always succeed
    return (QPERR_SUCCESS);

}

static EQP_Errors CComponent_Stop(CComponent *This)
{
    CPort *pPort;
    int i;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s [\n",
                 __FUNCTION__
                 ));

    ENTER_CRITICAL(&This->m_Object)

            // stop all ports
            //
    for (i = 0; i < MAX_COMPONENT_PORT; i++)
    {
        pPort = &This->m_Ports[i];
        if (pPort->m_valid)
        {
            CPort_SetState(pPort,
                           PORT_STATE_STOP,
                           pPort->m_State
                           );
        }
    }

    LEAVE_CRITICAL(&This->m_Object)

            // always succeed
    return (QPERR_SUCCESS);
}

int ccomponent_start_index(CComponent *This, COMPONENT_PORT_TYPE type)
{
    return CComponent_Start_Index(This, type);
}

int ccomponent_start(CComponent *This)
{
    return CComponent_Start(This);
}

int ccomponent_stop_index(CComponent *This, COMPONENT_PORT_TYPE type)
{
    return CComponent_Stop_Index(This, type);
}

int ccomponent_stop(CComponent *This)
{
    return CComponent_Stop(This);
}

static EQP_Errors CComponent_Acquire_Index(CComponent *This,
                                           COMPONENT_PORT_TYPE type)
{
    EQP_Errors qpStatus = QPERR_SUCCESS;
    CPort *pPort;
    //	int			i;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s [\n",
                 __FUNCTION__
                 ));

    if (type > MAX_COMPONENT_PORT)
        return -EINVAL;

    ENTER_CRITICAL(&This->m_Object)

    // from stop to acquire
    //
    pPort = &This->m_Ports[type];
    if (pPort->m_valid)
    {
        if (PORT_STATE_STOP == pPort->m_State)
        {
            qpStatus = CPort_SetState(pPort,
                                      PORT_STATE_ACQUIRE,
                                      PORT_STATE_STOP
                                      );
            if (!QP_SUCCESS(qpStatus))
            {
                goto CComponent_Acquire_Exit;
            }
        }
    }

    // from acquire to pause
    //
    pPort = &This->m_Ports[type];
    if (pPort->m_valid)
    {
        if (PORT_STATE_ACQUIRE == pPort->m_State)
        {
            qpStatus = CPort_SetState(pPort,
                                      PORT_STATE_PAUSE,
                                      PORT_STATE_ACQUIRE
                                      );
            if (!QP_SUCCESS(qpStatus))
            {
                goto CComponent_Acquire_Exit;
            }
        }
    }

CComponent_Acquire_Exit:

    if (!QP_SUCCESS(qpStatus))
    {
        pPort = &This->m_Ports[type];
        if (pPort->m_valid)
        {
            CPort_SetState(pPort,
                           PORT_STATE_STOP,
                           pPort->m_State
                           );
        }
    }

    LEAVE_CRITICAL(&This->m_Object)

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s status(%d) ]\n",
                 __FUNCTION__,
                 qpStatus
                 ));

    return (qpStatus);
}

static EQP_Errors CComponent_Acquire(CComponent *This)
{
    EQP_Errors qpStatus = QPERR_SUCCESS;
    CPort *pPort;
    int i;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s [\n",
                 __FUNCTION__
                 ));

    ENTER_CRITICAL(&This->m_Object)

            // from stop to acquire
            //
    for (i = 0; i < MAX_COMPONENT_PORT; i++)
    {
        pPort = &This->m_Ports[i];
        if (pPort->m_valid)
        {
            if (PORT_STATE_STOP == pPort->m_State)
            {
                qpStatus = CPort_SetState(pPort,
                                          PORT_STATE_ACQUIRE,
                                          PORT_STATE_STOP
                                          );
                if (!QP_SUCCESS(qpStatus))
                {
                    goto CComponent_Acquire_Exit;
                }
            }
        }
    }

    // from acquire to pause
    //
    for (i = 0; i < MAX_COMPONENT_PORT; i++)
    {
        pPort = &This->m_Ports[i];
        if (pPort->m_valid)
        {
            if (PORT_STATE_ACQUIRE == pPort->m_State)
            {
                qpStatus = CPort_SetState(pPort,
                                          PORT_STATE_PAUSE,
                                          PORT_STATE_ACQUIRE
                                          );
                if (!QP_SUCCESS(qpStatus))
                {
                    goto CComponent_Acquire_Exit;
                }
            }
        }
    }

CComponent_Acquire_Exit:

    if (!QP_SUCCESS(qpStatus))
    {
        for (i = 0; i < MAX_COMPONENT_PORT; i++)
        {
            pPort = &This->m_Ports[i];
            if (pPort->m_valid)
            {
                CPort_SetState(pPort,
                               PORT_STATE_STOP,
                               pPort->m_State
                               );
            }
        }
    }

    LEAVE_CRITICAL(&This->m_Object)

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s status(%d) ]\n",
                 __FUNCTION__,
                 qpStatus
                 ));

    return (qpStatus);
}

int ccomponent_acquire_index(CComponent *This,
                             COMPONENT_PORT_TYPE type)
{
    return CComponent_Acquire_Index(This, type);
}

int ccomponent_acquire(CComponent *This)
{
    return CComponent_Acquire(This);
}

/////////////////////////////////////////////////////////////////////////////
//
//


// constructor
//

CComponent * CComponent_Constructor(CComponent *pComponent,
                                    struct video_device *pVideoDev,
                                    CDevice *pDevice,
                                    COMPONENT_TYPE type
                                    )
{
    if (pComponent)
    {
        int i;

        // CObject
        //
        CObject_Constructor(&pComponent->m_Object,
                            QP_NULL,
                            OBJECT_CRITICAL_LIGHT
                            );

        // initialize members
        //
        pComponent->m_pVideoDev = pVideoDev;
        pComponent->m_pDevice = pDevice;
        pComponent->m_hTask = TQP_NULL_HANDLE;
        pComponent->m_hObject = TQP_NULL_HANDLE;
        pComponent->m_type = type;
        pComponent->out_width=0;
        pComponent->out_height=0;

        for (i = 0; i < MAX_COMPONENT_PORT; i++)
        {
            memset(&pComponent->m_Ports[i],
                   0,
                   sizeof (CPort)
                   );
            pComponent->m_Ports[i].m_valid = QP_FALSE;
        }

        if (type == COMPONENT_TYPE_PCM && pDevice->m_pAudioDev)
        {
            pl330_trace(PL330_TRACE_CALL,
                        "%s component@%p type %d assign audio device"
                        " on_buffer_done callback\n",
                        __func__,
                        pComponent,
                        type);
            pComponent->m_pAudioDev = pDevice->m_pAudioDev;
            pComponent->m_Ports[COMPONENT_PORT_PCM_OUT].
                    on_buffer_done = pComponent->m_pAudioDev->on_buffer_done;
        }
        if (type == COMPONENT_TYPE_ENCODE)
        {
            ICodecLib *codeclib_if = CDevice_getCodecLib(pComponent->m_pDevice);
            CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
            IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;

            
            frontend_register_event_callback(frontend,CComponent_Event_CB,pComponent);
        }    
        

        init_waitqueue_head(&pComponent->wait_frame);
    }

    return (pComponent);
}



// destructor
//

void CComponent_Destructor(CComponent *This)
{
    CObject_Destructor(&This->m_Object);
}

EQP_Errors CComponent_Open(CComponent *This)
{
    IMpegCodec *pMpegCodec = CDevice_getMpegCodec(This->m_pDevice);
    EQP_Errors qpStatus;

    // allocate task
    //
    if (COMPONENT_TYPE_ENCODE == This->m_type)
    {
        qpStatus = pMpegCodec->AllocEncodeTask(pMpegCodec,
                                               &This->m_hTask
                                               );
    }
    else if (COMPONENT_TYPE_PCM == This->m_type)
    {
        qpStatus = pMpegCodec->AllocPCMTask(pMpegCodec,
                                            &This->m_hTask);
    }
    else
    {
        qpStatus = pMpegCodec->AllocDecodeTask(pMpegCodec,
                                               &This->m_hTask
                                               );
    }

    if (!QP_SUCCESS(qpStatus))
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                     "%s AllocTask(%d) failed status(%d)\n",
                     __FUNCTION__,
                     This->m_type,
                     qpStatus
                     ));
    }
    else
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
                     "%s AllocTask(%d) hTask(%d)\n",
                     __FUNCTION__,
                     This->m_type,
                     This->m_hTask
                     ));
        // add to the device's component manager
        //
        This->m_hObject = CObjectMgr_AddObject(This->m_pDevice->m_pComponentMgr,
                                               This
                                               );
    }

    return (qpStatus);

}

EQP_Errors CComponent_Close(CComponent *This)
{
    IMpegCodec *pMpegCodec = CDevice_getMpegCodec(This->m_pDevice);
    int i;

    // make sure close all the opened ports
    //
    for (i = 0; i < MAX_COMPONENT_PORT; i++)
    {
        CComponent_PortClose(This,
                             (COMPONENT_PORT_TYPE) i
                             );
    }

    // release the allocated task
    //
    if (TQP_NULL_HANDLE != This->m_hTask)
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
                     "%s ReleaseTask(%d) hTask(%d)\n",
                     __FUNCTION__,
                     This->m_type,
                     This->m_hTask
                     ));

        pMpegCodec->ReleaseTask(pMpegCodec,
                                This->m_hTask
                                );
        This->m_hTask = TQP_NULL_HANDLE;
    }

    // remove ourself from the device's component manager
    //
    if (TQP_NULL_HANDLE != This->m_hObject)
    {
        CObjectMgr_RemoveObject(This->m_pDevice->m_pComponentMgr,
                                This->m_hObject
                                );
        This->m_hObject = TQP_NULL_HANDLE;
    }

    return (QPERR_SUCCESS);
}

int snd_pl330_dqbuf(CComponent *component, struct v4l2_buffer *buf)
{
    CPort *pPort;
    EQP_Errors err;

    buf->sequence = component->m_pDevice->audio_port_type;

    pl330_trace(PL330_TRACE_PCM,
                "%s dqbuf index=%d, port type=%d, byteused=%d\n",
                __func__,
                buf->index,
                buf->sequence,
                buf->bytesused);

    pPort = &component->m_Ports[buf->sequence];
    err = CPort_DQBuf(pPort, buf);

    pl330_trace(PL330_TRACE_PCM,
                "%s end of dqbuf status=%d, index=%d, offset=%u, byteused=%d\n",
                __func__, err,
                buf->index,
                buf->m.offset, buf->bytesused);
    return QpToLinuxKerStatus(err);
}

int snd_pl330_qbuf(CComponent *component, struct v4l2_buffer *buf)
{
    CPort *pPort;
    EQP_Errors err;

    pl330_trace(PL330_TRACE_PCM,
                "%s qbuf index=%d, port type=%d\n", __func__,
                buf->index,
                buf->sequence);

    buf->sequence = component->m_pDevice->audio_port_type;

    pPort = &component->m_Ports[buf->sequence];
    err = CPort_QBuf(pPort, buf);

    return QpToLinuxKerStatus(err);
}

static void pl330_default_config(struct CComponent *component)
{
	pl330_set_config(component, "SystemControl-StreamData", "video+audio", false);
	pl330_set_config(component, "SystemControl-XferMode", "frame", false);
	pl330_set_config(component, "SystemControl-SpsrFreq", "1", false);
	pl330_set_config(component, "SystemControl-FFMode", "frame", false);
	pl330_set_config(component, "SystemControl-VMode", "cavlc", false);

	pl330_set_config(component, "InputControl-SrcMode", "hdmi", false);
	pl330_set_config(component, "InputControl-DataType", "raw", false);

	pl330_set_config(component, "InputControl-CkEdge", "positive", false);

	pl330_set_config(component, "RateControl-Vbr", "0", false);
	pl330_set_config(component, "RateControl-Mode", "viu", false);

	pl330_set_config(component, "FilterControl-StartPixel", "1", false);
	pl330_set_config(component, "FilterControl-StartLine", "0", false);

	pl330_set_config(component, "GopLoopFilter-IntraPeriod", "30", false);
	pl330_set_config(component, "GopLoopFilter-BNum", "0", false);
	pl330_set_config(component, "GopLoopFilter-Idr", "close", false);

	pl330_set_config(component, "SysFunction-Function", "encode", false);
	pl330_set_config(component, "SysLink-VideoInput", "viu", false);
	pl330_set_config(component, "SysLink-VideoOutput", "host", false);


	pl330_set_config(component, "SysFunction-Audio", "audio", false);
	pl330_set_config(component, "SysLink-AudioInput", "aiu", false);
	pl330_set_config(component, "SysLink-AudioOutput", "host", false);

	pl330_set_config(component, "AudioControlParam-AudioType", "aac", false);
	pl330_set_config(component, "AudioControlParam-ChNum", "2", false);
	pl330_set_config(component, "AudioControlEx-AacVer", "mpeg2", false);
	pl330_set_config(component, "AudioControlEx-HType", "adts", false);
	pl330_set_config(component, "AudioControlEx-CutoffFreq", "18000", false);
	pl330_set_config(component, "AudioControlEx-TNS", "1", false);
	pl330_set_config(component, "AudioControlEx-IS", "1", false);
	pl330_set_config(component, "AudioControlEx-PNS", "1", false);
	pl330_set_config(component, "AudioControlEx-MS", "1", false);

}

static void CComponent_Event_CB(void *cxt,IFrontend_event_e event)
{
    CComponent *pComponent=cxt;
    ICodecLib *codeclib_if = CDevice_getCodecLib(pComponent->m_pDevice);
    CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
    IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
    u32 io_framerate;
    pl330_trace(PL330_TRACE_VIDIOC,"%s event %d\n",__func__,event);
    printk(" event %d\n",event);
    
    
    switch(event)
    {
    case FRONTEND_EVENT_NO_SIGNAL:
    case FRONTEND_EVENT_NO_SYNC:
        pl330_acquire_index(pComponent,pComponent->m_pDevice->v4l2_port_type);
        //ccomponent_stop_index(pComponent,pComponent->m_pDevice->v4l2_port_type);
        break;
    case FRONTEND_EVENT_READY_STREAMING:
    case FRONTEND_EVENT_LOCK:
    {
        const frontend_frameinfo_t *frameinfo;
        const frontend_audioinfo_t *audioinfo;
        u32 sync_mode;
        u32 framerate,in_framerate;
        int width,height;
        int out_width,out_height;
        
        
        frameinfo=frontend_g_frameinfo(frontend);
        audioinfo=frontend_g_audioinfo(frontend);
        
        width=frameinfo->width;
        height=frameinfo->height;
        
        if (width ==0 || height ==0)
        {
			printk("%s detect input --> not stable\n",__func__);
			width = 720;
			height = 480;
		}
        
        out_width=pComponent->out_width;
        out_height=pComponent->out_height;
        
        if(frameinfo==NULL)
        {
            pl330_trace(PL330_TRACE_VIDIOC,"Conflict Lock but frameinfo NULL\n");
            return;
        }
        
        if(out_width==0 || out_height==0)
        {
            out_width=width;
            out_height=height;
        }
        
        framerate=frameinfo->framerate;
        if(frameinfo->is_interlace)
            framerate >>= 1;
        #if 0 // fix signal plug error
        if(framerate==60)
            in_framerate=IN_FRAME_RATE_NTSC;
    	else if(framerate==50)
            in_framerate=IN_FRAME_RATE_PAL;
    	else
        #endif
        
       // in_framerate=framerate | IN_FRAME_RATE_ABSOLUTE;
        in_framerate=framerate; // add ABSOLUTE later
        
        printk("%s frameinfo->polarities=%d\n",__func__,frameinfo->polarities);
        
        switch(frameinfo->polarities)
        {
            case H_POS_V_POS:
            sync_mode=SYNC_MODE_POS_POS;
            break;
            case H_POS_V_NEG:
            sync_mode=SYNC_MODE_POS_NEG;
            break;
            case H_NEG_V_POS:
            sync_mode=SYNC_MODE_NEG_POS;
            break;
            case H_NEG_V_NEG:
            sync_mode=SYNC_MODE_NEG_NEG;
            break;
            default:
            sync_mode=SYNC_MODE_POS_POS;
            break;
		}
		
		if ((frameinfo->frontend_select == gv7601_f) && (frameinfo->width == 720))
		{
			pl330_set_config(pComponent, "InputControl-SyncMode", "3", false);
		}
		else
            pl330_set_config(pComponent, "InputControl-SyncMode", &sync_mode, true);        
            
#if 0        
        if (frameinfo->frontend_select==adv7604_f)
        {
            switch(frameinfo->polarities)
            {
                case H_POS_V_POS:
                sync_mode=SYNC_MODE_POS_POS;
                break;
                case H_POS_V_NEG:
                sync_mode=SYNC_MODE_POS_NEG;
                break;
                case H_NEG_V_POS:
                sync_mode=SYNC_MODE_NEG_POS;
                break;
                case H_NEG_V_NEG:
                sync_mode=SYNC_MODE_NEG_NEG;
                break;
                default:
                sync_mode=SYNC_MODE_POS_POS;
                break;
			}
            pl330_set_config(pComponent, "InputControl-SyncMode", &sync_mode, true);
        }
        else
        {
			switch(frameinfo->polarities) 
			{
		        case 0:
			    pl330_set_config(pComponent, "InputControl-SyncMode", "0", false);
			    break;
		        case 1:
			    pl330_set_config(pComponent, "InputControl-SyncMode", "1", false);
			    break;
		        case 2:
			    pl330_set_config(pComponent, "InputControl-SyncMode", "2", false);
			    break;
		        case 3:
			    pl330_set_config(pComponent, "InputControl-SyncMode", "3", false);
			    break;
			    default:
			    pl330_set_config(pComponent, "InputControl-SyncMode", "3", false);
			    break;
	        }
			
		}
#endif        
#ifdef CONFIG_DISABLE_DE_INTERLACE
        {
            pl330_set_config(pComponent, "InputControl-ScanFormat", "progressive", false);
            pl330_set_config(pComponent, "InputControl-Fmt", "progressive", false);
            pl330_set_config(pComponent, "DeInterlace-Mode", "none", false);
        }
#else
        if(frameinfo->is_interlace)
        {
            pl330_set_config(pComponent, "InputControl-ScanFormat", "interlace", false);
            pl330_set_config(pComponent, "InputControl-Fmt", "interlace", false);
            pl330_set_config(pComponent, "DeInterlace-Mode", "topfield", false);
	    }else
	    {
            pl330_set_config(pComponent, "InputControl-ScanFormat", "progressive", false);
            pl330_set_config(pComponent, "InputControl-Fmt", "progressive", false);
            pl330_set_config(pComponent, "DeInterlace-Mode", "none", false);
        }
#endif 

    io_framerate=frontend_g_framerate(frontend);
        
    //printk("%s get framerate= %u\n", __func__, io_framerate);
    if(io_framerate==0)
    {
        frameinfo=frontend_g_frameinfo(frontend);
        if(frameinfo)
            io_framerate=frameinfo->framerate;
        else
            io_framerate=30;
 
    }
    
    if (io_framerate >60)
        io_framerate=60;
    //printk("%s >>>set io_framerate= %u\n", __func__, io_framerate);
    //io_framerate |= IN_FRAME_RATE_ABSOLUTE;
	//pl330_set_config(pComponent, "InputControl-SyncMode", &sync_mode, true);
	
	//pl330_set_config(pComponent, "InputControl-SyncMode","3",false);
	
	//pl330_set_config(pComponent, "InputControl-InFrameRate", &io_framerate, true);
	in_framerate = in_framerate | IN_FRAME_RATE_ABSOLUTE;
	pl330_set_config(pComponent, "InputControl-InFrameRate", &in_framerate, true);
	
	pl330_set_config(pComponent, "PictureResolution-InPicWidth",&width, true);
	pl330_set_config(pComponent, "PictureResolution-InPicHeight",&height, true);
	pl330_set_config(pComponent, "OutPictureResolution-OutPicWidth",&out_width, true);
	pl330_set_config(pComponent, "OutPictureResolution-OutPicHeight",&out_height, true);
    printk("====sync mode %d in_framerate %d In %dx%d out %dx%d\n",sync_mode,in_framerate,width,height,out_width,out_height);
	//framerate |=  IN_FRAME_RATE_ABSOLUTE;
	//printk("%s >>>get out_framerate= %u\n", __func__, framerate);
	//pl330_set_config(pComponent, "InputControl-OutFrameRate",&framerate, true);
	io_framerate |= IN_FRAME_RATE_ABSOLUTE;
	pl330_set_config(pComponent, "InputControl-OutFrameRate",&io_framerate, true);
	pl330_acquire(pComponent);
	ccomponent_start_index(pComponent,pComponent->m_pDevice->v4l2_port_type);
        
    }
        break;
    default:
        break;
    }
        
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
long CComponent_Ioctl(struct file *filp,
                      unsigned int cmd,
                      void *arg
                      )
#else // LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)

long CComponent_Ioctl(struct inode *inode,
                      struct file *filp,
                      unsigned int cmd,
                      void *arg
                      )
#endif //LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
{
    CComponent *pComponent = (CComponent *) filp->private_data;
    unsigned int nSize = _IOC_SIZE(cmd);
    //	unsigned int		nDir = _IOC_DIR(cmd);
    unsigned int nType = _IOC_TYPE(cmd);
    unsigned int nr = _IOC_NR(cmd);
    EQP_Errors qpStatus = QPERR_PARMS;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_EXTRACE),
                 "%s() cmd(0x%08x) type(%c) nr(%d) size(%d)\n",
                 __FUNCTION__,
                 cmd,
                 nType,
                 nr,
                 nSize
                 ));

    switch (cmd)
    {
        /////////////////////////////////////////////////////////////////////////////
        //
        //
        // v4l ioctl
        /* Control handling */
    case VIDIOC_QUERYCTRL:
        pl330_trace(PL330_TRACE_VIDIOC, "%s queryctrl\n", __func__);
        return ccomponent_queryctrl(pComponent, arg);
        break;
    case VIDIOC_G_CTRL:
        pl330_trace(PL330_TRACE_VIDIOC, "%s g_ctrl\n", __func__);
        return ccomponent_g_ctrl(pComponent, arg);
        break;
    case VIDIOC_S_CTRL:
        pl330_trace(PL330_TRACE_VIDIOC, "%s s_ctrl\n", __func__);
        return ccomponent_s_ctrl(pComponent, arg);
        break;

        /* input handling */
    case VIDIOC_ENUMINPUT:
        pl330_trace(PL330_TRACE_VIDIOC, "%s enuminput\n", __func__);
        return ccomponent_enum_input(pComponent, arg);
        break;
    case VIDIOC_G_INPUT:
        pl330_trace(PL330_TRACE_VIDIOC, "%s g_input\n", __func__);
        return ccomponent_g_input(pComponent, arg);
        break;
    case VIDIOC_S_INPUT:
        pl330_trace(PL330_TRACE_VIDIOC, "%s s_input\n", __func__);
        return ccomponent_s_input(pComponent, arg);
        break;

        /* querycap handler */
    case VIDIOC_QUERYCAP:
        pl330_trace(PL330_TRACE_VIDIOC, "%s querycap\n", __func__);
        return ccomponent_querycap(pComponent, arg);
        break;

        /* fmt handler */
    case VIDIOC_ENUM_FMT:
        pl330_trace(PL330_TRACE_VIDIOC, "%s enum_fmt\n", __func__);
        return ccomponent_enum_fmt(pComponent, arg);
        break;
    case VIDIOC_G_FMT:
        pl330_trace(PL330_TRACE_VIDIOC, "%s g_fmt\n", __func__);
        return ccomponent_g_fmt(pComponent, arg);
        break;
    case VIDIOC_S_FMT:
        pl330_trace(PL330_TRACE_VIDIOC, "%s s_fmt\n", __func__);
        return ccomponent_s_fmt(pComponent, arg);
        break;
    case VIDIOC_TRY_FMT:
        pl330_trace(PL330_TRACE_VIDIOC, "%s try_fmt\n", __func__);
        return ccomponent_try_fmt(pComponent, arg);
        break;

    case VIDIOC_ENUMSTD:
        return -EINVAL;
        break;

        /* stream on/off */
    case VIDIOC_STREAMON:
        pl330_trace(PL330_TRACE_VIDIOC, "%s "
                    "component type=%d dummy type=%d streamon\n",
                    __func__, pComponent->m_type,
                    pComponent->m_dummy_type);
        qpStatus = ccomponent_streamon(pComponent);
        break;
    case VIDIOC_STREAMOFF:
        pl330_trace(PL330_TRACE_VIDIOC, "%s "
                    "component type=%d dummy type=%d streamoff\n",
                    __func__, pComponent->m_type,
                    pComponent->m_dummy_type);

        qpStatus = ccomponent_streamoff(pComponent);
        break;

        /* buffer handler */
    case VIDIOC_QBUF:
    {
        CPort *pPort;
        struct v4l2_buffer *pv4lbuf;

        pv4lbuf = (struct v4l2_buffer *) arg;

        pv4lbuf->sequence = (pComponent->m_dummy_type == COMPONENT_TYPE_ENCODE) ?
                pComponent->m_pDevice->v4l2_port_type :
                pComponent->m_pDevice->audio_port_type;

        pl330_trace(PL330_TRACE_VIDIOC, "%s qbuf index=%d, port type=%d bytesused=%d\n", __func__,
                    pv4lbuf->index,
                    pv4lbuf->sequence,
                    pv4lbuf->bytesused);


        if (!pv4lbuf ||
                (pv4lbuf->sequence >= MAX_COMPONENT_PORT)
                )
        {
            QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                         "%s() VIDIOC_QBUF unknown sequence(%d)!!\n",
                         __FUNCTION__,
                         pv4lbuf->sequence
                         ));
            qpStatus = QPERR_PARMS;
        }
        else
        {
            pPort = &pComponent->m_Ports[pv4lbuf->sequence];
            qpStatus = CPort_QBuf(pPort,
                                  pv4lbuf
                                  );
        }
        pl330_trace(PL330_TRACE_VIDIOC, "%s qbuf status=%d\n", __func__,
                    qpStatus);
        if (qpStatus == QPERR_SUCCESS)
        {
            pComponent->m_pDevice->qbuf_event = true;
            wake_up_interruptible(&pComponent->m_pDevice->wq_qbuf);

            /* wait for next dqbuf */
            queue_work(pComponent->m_pDevice->dqbuf_queues, &pComponent->m_pDevice->wq_wait_dqbuf);
        }
        break;
    }

    case VIDIOC_DQBUF:
    {
        CPort *pPort;
        struct v4l2_buffer *pv4lbuf;

        pv4lbuf = (struct v4l2_buffer *) arg;
        pv4lbuf->sequence = (pComponent->m_dummy_type == COMPONENT_TYPE_ENCODE) ?
                pComponent->m_pDevice->v4l2_port_type :
                pComponent->m_pDevice->audio_port_type;

        pl330_trace(PL330_TRACE_VIDIOC, "%s dqbuf index=%d, port type=%d, byteused=%d\n",
                    __func__,
                    pv4lbuf->index,
                    pv4lbuf->sequence,
                    pv4lbuf->bytesused);
        if (!pv4lbuf ||
                (pv4lbuf->sequence >= MAX_COMPONENT_PORT)
                )
        {
            QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                         "%s() VIDIOC_DQBUF unknown sequence(%d)!!\n",
                         __FUNCTION__,
                         pv4lbuf->sequence
                         ));

            pl330_trace(PL330_TRACE_VIDIOC, "%s dqbuf unknown sequence\n", __func__);
            qpStatus = QPERR_PARMS;
        }
        else
        {
            pPort = &pComponent->m_Ports[pv4lbuf->sequence];
            if (pPort->m_valid)
            {
                qpStatus = CPort_DQBuf(pPort, pv4lbuf);
            }
            else
            {
                qpStatus = QPERR_AGAIN;
                break;
            }
        }

        if (qpStatus == QPERR_SUCCESS)
        {
            pl330_trace(PL330_TRACE_VIDIOC,
                        "%s end of dqbuf status=%d, index=%d, offset=%u, byteused=%d\n",
                        __func__, qpStatus,
                        pv4lbuf->index,
                        pv4lbuf->m.offset, pv4lbuf->bytesused);

            pComponent->m_pDevice->dqbuf_event = true;
            wake_up_interruptible(&pComponent->m_pDevice->wq_dqbuf);

            /* wait for next qbuf */
            queue_work(pComponent->m_pDevice->qbuf_queues, &pComponent->m_pDevice->wq_wait_qbuf);
        }
        //			schedule_work(&pComponent->m_pDevice->wq_wait_qbuf);

        break;
    }

    case VIDIOC_REQBUFS:
    {
        struct v4l2_requestbuffers *reqbuf = (struct v4l2_requestbuffers *) arg;
        ICodecLib *codeclib_if = CDevice_getCodecLib(pComponent->m_pDevice);
        CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
        IFrontend *frontend=&codeclib->m_pVidDecoder->m_frontend;
         COMPONENT_PORT_OPEN port_open;
        TQP_UINT32 framerate;
        TQP_UINT32 type;
        int out_width, out_height;
        TQP_UINT32 pixfmt;
        
        const frontend_frameinfo_t * frameinfo;
        CPort *pPort;
        if(reqbuf->count == 0)
            return 0;
        
        
        frameinfo=frontend_g_frameinfo(frontend);
        
        out_width = pComponent->out_width;
        out_height = pComponent->out_height;
        
        pixfmt=frontend_g_pixfmt(frontend);
        
        if(frameinfo)
        {
            framerate = frameinfo->framerate;
            if (frameinfo->is_interlace)
                framerate /= 2;
        }else
        {
            framerate=30;
        }
                  
        if (out_width == 0 || out_height == 0)
        {
            if(frameinfo)
            {
                out_width=frameinfo->width;
                out_height=frameinfo->height;
                framerate=frameinfo->framerate;
                
            }else
            {
                out_width = 1920;
                out_height = 1080;
                framerate = 30;
            }
        }
        memset(&port_open, 0, sizeof (COMPONENT_PORT_OPEN));
        if(pixfmt== V4L2_PIX_FMT_MPEG)
        {
            pl330_set_config(pComponent, "SystemControl-StreamType", "ts", false);
            pl330_set_config(pComponent, "SysFunction-Video", "h264", false);
            port_open.type = COMPONENT_PORT_COMP_OUT;
            port_open.format.vid.nDataType = QP_COMP_VID_DATA_TYPE_H264;
        }else if(pixfmt==V4L2_PIX_FMT_YVU420)
        {
            pl330_set_config(pComponent, "SystemControl-StreamType", "raw", false);
            pl330_set_config(pComponent, "SysFunction-Video", "passthrough", false);
            port_open.format.yuv.nWidth = out_width;
            port_open.format.yuv.nHeight = out_height;
            port_open.type = COMPONENT_PORT_YUV_OUT;
            port_open.format.yuv.nBitCount = 12;
            port_open.format.yuv.nDataType = QP_YUV_DATA_TYPE_YV12;
            if (framerate)
                port_open.format.yuv.nFrameRate = framerate;
            else
                port_open.format.yuv.nFrameRate = 30;
            pl330_trace(PL330_TRACE_VIDEO,"COMPONENT_PORT_YUV_OUT %dx%d %d\n", port_open.format.yuv.nWidth, port_open.format.yuv.nHeight, port_open.format.yuv.nFrameRate);
        }else if(pixfmt==V4L2_PIX_FMT_YUV420)
        {
            pl330_set_config(pComponent, "SystemControl-StreamType", "raw", false);
            pl330_set_config(pComponent, "SysFunction-Video", "passthrough", false);
            port_open.format.yuv.nWidth = out_width;
            port_open.format.yuv.nHeight = out_height;
            port_open.type = COMPONENT_PORT_YUV_OUT;
            port_open.format.yuv.nBitCount = 12;
            port_open.format.yuv.nDataType = QP_YUV_DATA_TYPE_I420;
            if (framerate)
                port_open.format.yuv.nFrameRate = framerate;
            else
                port_open.format.yuv.nFrameRate = 30;
            pl330_trace(PL330_TRACE_VIDEO,"COMPONENT_PORT_YUV_OUT i420 %dx%d %d\n", port_open.format.yuv.nWidth, port_open.format.yuv.nHeight, port_open.format.yuv.nFrameRate);
        }else if(pixfmt==V4L2_PIX_FMT_NV12)
        {
            pl330_set_config(pComponent, "SystemControl-StreamType", "raw", false);
            pl330_set_config(pComponent, "SysFunction-Video", "passthrough", false);
            port_open.format.yuv.nWidth = out_width;
            port_open.format.yuv.nHeight = out_height;
            port_open.type = COMPONENT_PORT_YUV_OUT;
            port_open.format.yuv.nBitCount = 12;
            port_open.format.yuv.nDataType = QP_YUV_DATA_TYPE_NV12;
            if (framerate)
                port_open.format.yuv.nFrameRate = framerate;
            else
                port_open.format.yuv.nFrameRate = 30;
            pl330_trace(PL330_TRACE_VIDEO,"COMPONENT_PORT_YUV_OUT NV12 %dx%d %d\n", port_open.format.yuv.nWidth, port_open.format.yuv.nHeight, port_open.format.yuv.nFrameRate);
        }

        port_open.memory = V4L2_MEMORY_MMAP;
        if((qpStatus=pl330_port_open(pComponent, &port_open))==QPERR_SUCCESS)
        {
            if (pixfmt == V4L2_PIX_FMT_MPEG)
            {
                type = COMPONENT_PORT_COMP_OUT;
            }    
            else
            {
                type = COMPONENT_PORT_YUV_OUT;
            }
            pPort = &pComponent->m_Ports[type];
            qpStatus = CPort_ReqBuf(pPort,reqbuf);
        }
        
        pl330_trace(PL330_TRACE_VIDIOC, "%s reqbuf count=%d\n", __func__, reqbuf->count);
        
	
    }
    break;
#if 0        
    {
        
        CPort *pPort;
        struct v4l2_requestbuffers *reqbuf;
        COMPONENT_PORT_OPEN port_open;
        //			int i;


        /* workaround for gstreamer/qv4l2 */
#if 1
        ICodecLib *codeclib_if = CDevice_getCodecLib(pComponent->m_pDevice);
        CQLCodecLib *codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
        CVideoDecoder *decoder = codeclib->m_pVidDecoder;
        CADV7604 *adv7604 = (CADV7604 *) codeclib->m_pVidDecoder;
        struct adv7604_state *state = &adv7604->state;

        memset(&port_open, 0, sizeof (COMPONENT_PORT_OPEN));
        reqbuf = (struct v4l2_requestbuffers *) arg;


        if (decoder->v4l2_ops.s_input)
            decoder->v4l2_ops.s_input(decoder, adv7604->input);

        if (decoder->v4l2_ops.try_frame)
        {
            if (decoder->v4l2_ops.try_frame(decoder,
                                            &state->cur_out_frame, &state->format))
            {
                return -EINVAL;
            }
        }

#if 0
        port_open.format.yuv.nWidth = state->timings.bt.width;
        port_open.format.yuv.nHeight = state->timings.bt.height;
        port_open.type = COMPONENT_PORT_YUV_OUT;
        port_open.format.yuv.nBitCount = 12;
        port_open.format.yuv.nDataType = QP_YUV_DATA_TYPE_YV12;
        port_open.format.yuv.nFrameRate = 30;

        port_open.memory = V4L2_MEMORY_MMAP;

        pl330_port_open(pComponent, &port_open);

        pl330_acquire(pComponent);
#endif
#endif

        pl330_trace(PL330_TRACE_VIDIOC, "%s reqbuf type=%d, memory=%d, count=%d\n",
                    __func__, reqbuf->type, reqbuf->memory, reqbuf->count);

        reqbuf->count = (pComponent->m_dummy_type == COMPONENT_TYPE_ENCODE) ?
                pComponent->m_pDevice->v4l2_port_type :
                pComponent->m_pDevice->audio_port_type;

        if (!reqbuf ||
                (reqbuf->count >= MAX_COMPONENT_PORT)
                )
        {
            QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                         "%s() VIDIOC_REQBUFS unknown sequence(%d)!!\n",
                         __FUNCTION__,
                         reqbuf->count
                         ));
            qpStatus = QPERR_PARMS;
        }
        else
        {
            pPort = &pComponent->m_Ports[reqbuf->count];
            qpStatus = CPort_ReqBuf(pPort,
                                    reqbuf
                                    );
        }
        pl330_trace(PL330_TRACE_VIDIOC, "%s reqbuf count=%d\n", __func__, reqbuf->count);
        break;
    }
#endif
    case VIDIOC_QUERYBUF:
    {
        CPort *pPort;
        struct v4l2_buffer *pv4lbuf;

        pv4lbuf = (struct v4l2_buffer *) arg;
        pl330_trace(PL330_TRACE_VIDIOC, "%s querybuf index=%d\n", __func__, pv4lbuf->index);

        pv4lbuf->sequence = (pComponent->m_dummy_type == COMPONENT_TYPE_ENCODE) ?
                pComponent->m_pDevice->v4l2_port_type :
                pComponent->m_pDevice->audio_port_type;

        if (!pv4lbuf ||
                (pv4lbuf->sequence >= MAX_COMPONENT_PORT)
                )
        {
            QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                         "%s() VIDIOC_QUERYBUF unknown sequence(%d)!!\n",
                         __FUNCTION__,
                         pv4lbuf->sequence
                         ));
            qpStatus = QPERR_PARMS;
        }
        else
        {
            pPort = &pComponent->m_Ports[pv4lbuf->sequence];
            qpStatus = CPort_QueryBuf(pPort,
                                      pv4lbuf
                                      );
            pv4lbuf->memory = V4L2_MEMORY_MMAP;
            
            pl330_trace(PL330_TRACE_VIDIOC, "%s querybuf status%d\n", __func__, qpStatus);
            
        }
        break;
        
    }
    case VIDIOC_ENUM_FRAMESIZES:
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s enum_framesizes\n", __func__);
        return component_enum_framesizes(pComponent, arg);
    }
    case VIDIOC_ENUM_FRAMEINTERVALS:
    {
        return component_enum_frameintervals(pComponent, arg);
    }
    case VIDIOC_G_PARM:
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s g_parm\n", __func__);
        return component_g_parm(pComponent, arg);
    }
    case VIDIOC_S_PARM:
    {
        pl330_trace(PL330_TRACE_VIDIOC, "%s s_parm\n", __func__);
        return component_s_parm(pComponent, arg);
    }
    case VIDIOC_DQEVENT:
    {
        return ccomponent_dqevent(pComponent, arg);
    }


        /////////////////////////////////////////////////////////////////////////////
        //
        //

        // stream extension interface
        //
#if 0
    case C353_STRM_IOCTL_FMT_CHANGE:
        pl330_trace(PL330_TRACE_IOCTL, "%s fmt_change\n", __func__);
        return component_fmt_change(pComponent, arg);
#endif
    case C353_STRM_IOCTL_AUDIO_RATE:
        return component_audio_rate(pComponent, arg);
    case PLDEV_STRM_IOCTL_PORT_OPEN:
    {
        PCOMPONENT_PORT_OPEN pOpen = (PCOMPONENT_PORT_OPEN) arg;

        pl330_trace(PL330_TRACE_IOCTL, "%s port_open\n", __func__);
        qpStatus = CComponent_PortOpen(pComponent,
                                       pOpen
                                       );
        break;
    }

    case PLDEV_STRM_IOCTL_PORT_CLOSE:
    {
        PCOMPONENT_PORT_TYPE pType = (PCOMPONENT_PORT_TYPE) arg;

        pl330_trace(PL330_TRACE_IOCTL, "%s port[%d] close\n",
                    __func__, pType);
        qpStatus = CComponent_PortClose(pComponent,
                                        *pType
                                        );
        break;
    }

    case PLDEV_STRM_IOCTL_DEVICE_INFO:
    {
        PPLDEV_INFO pInfo = (PPLDEV_INFO) arg;

        pl330_trace(PL330_TRACE_IOCTL, "%s\n", __func__);
        /* special checking for driver and api version */
        if ((pInfo->busType != PLDRV_VERSION_CODE) || (pInfo->type != PLIOCTLAPI_VERSION_CODE))
        {
            QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                         "%s WARNING: API Version not match! App (DRV:0x%x API:0x%x), Driver (DRV:0x%x API:0x%x)!!\n",
                         __FUNCTION__, pInfo->busType, pInfo->type, PLDRV_VERSION_CODE, PLIOCTLAPI_VERSION_CODE));
        }
        if (((pInfo->busType & 0xFFFFFF00) != (PLDRV_VERSION_CODE & 0xFFFFFF00)) ||
                ((pInfo->type & 0XFFFFFF00) != (PLIOCTLAPI_VERSION_CODE & 0xFFFFFF00)))
        {
            QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                         "%s ERROR: Found but PLDEV APP and Drvier version not match! App (DRV:0x%x API:0x%x), Driver (DRV:0x%x API:0x%x)!!\n",
                         __FUNCTION__, pInfo->busType, pInfo->type, PLDRV_VERSION_CODE, PLIOCTLAPI_VERSION_CODE));
            return QPERR_CTRL_VERSION;
        }
        switch (pComponent->m_pDevice->m_type)
        {
        case PLV4LDEV_TYPE_QL300_PCIE_EVK:
            pInfo->busType = QPHCI_BUS_PCI;
            break;

        default:
            pInfo->busType = -1;
            break;
        }
        pInfo->devInstance = pComponent->m_pDevice->m_id;
        pInfo->type = pComponent->m_dummy_type;
        pInfo->hTask = pComponent->m_hTask;
        qpStatus = QPERR_SUCCESS;
        break;
    }

    case PLDEV_STRM_IOCTL_ACQUIRE:
    {
        pl330_trace(PL330_TRACE_IOCTL, "%s\n", __func__);
        qpStatus = CComponent_Acquire(pComponent);
        break;
    }

    case PLDEV_STRM_IOCTL_GET_DRIVER_VER:
    {
        PTQP_UINT32 pVersion = (PTQP_UINT32) arg;

        pl330_trace(PL330_TRACE_IOCTL, "%s\n", __func__);
        *pVersion = PLDRV_VERSION_CODE;
        qpStatus = QPERR_SUCCESS;
        break;
    }

    case PLDEV_STRM_IOCTL_GET_IOCTLAPI_VER:
    {
        PTQP_UINT32 pVersion = (PTQP_UINT32) arg;

        pl330_trace(PL330_TRACE_IOCTL, "%s\n", __func__);
        *pVersion = PLIOCTLAPI_VERSION_CODE;
        qpStatus = QPERR_SUCCESS;
        break;
    }

        /////////////////////////////////////////////////////////////////////////////
        //
        //

        // device control interface
        //
    case PLDEV_CNTL_IOCTL_SET_SYSTEM_CONTROL:
    case PLDEV_CNTL_IOCTL_SET_PICTURE_RESOLUTION:
    case PLDEV_CNTL_IOCTL_SET_INPUT_CONTROL:
    case PLDEV_CNTL_IOCTL_SET_RATE_CONTROL:
    case PLDEV_CNTL_IOCTL_SET_VBR_BIT_RATE:
    case PLDEV_CNTL_IOCTL_SET_GOP_LOOP_FILTER:
    case PLDEV_CNTL_IOCTL_SET_ET_CONTROL:
    case PLDEV_CNTL_IOCTL_SET_FILTER_CONTROL:
    case PLDEV_CNTL_IOCTL_SET_OUT_PICTURE_RESOLUTION:
    case PLDEV_CNTL_IOCTL_SET_BLOCK_SIZE:
    case PLDEV_CNTL_IOCTL_SET_AUDIO_CONTROL_PARAM:
    case PLDEV_CNTL_IOCTL_SET_AUDIO_CONTROL_EX:
    case PLDEV_CNTL_IOCTL_SET_REL_CUR_VID_BUF:
    case PLDEV_CNTL_IOCTL_SET_INSERT_I_FRAME:
    case PLDEV_CNTL_IOCTL_SET_INSERT_USER_DATA:
    case PLDEV_CNTL_IOCTL_SET_MP4_VID_NUM_BLOCK:
    case PLDEV_CNTL_IOCTL_SET_AUDIO_ENHANCEMENT:
    case PLDEV_CNTL_IOCTL_SET_ENC_STOP_MODE:
    case PLDEV_CNTL_IOCTL_SET_INDEX_CAP_FREQ:
    case PLDEV_CNTL_IOCTL_SET_ENC_VID_PADDING:
    case PLDEV_CNTL_IOCTL_SET_ENC_FREEZE_VIDEO:
    case PLDEV_CNTL_IOCTL_SET_ENC_STILL_VIDEO_INPUT:
    case PLDEV_CNTL_IOCTL_SET_ENC_CAP_MODE:
    case PLDEV_CNTL_IOCTL_SET_ENC_GET_A_FRAME:
    case PLDEV_CNTL_IOCTL_SET_ENC_MJPEG_QUALITY:
    case PLDEV_CNTL_IOCTL_SET_ENC_MJPEG_FRAME_BUF:
    case PLDEV_CNTL_IOCTL_SET_ENC_ALL_SETTINGS:
    case PLDEV_CNTL_IOCTL_SET_ENC_AUD_INPUT_VOL:

    case PLDEV_CNTL_IOCTL_SET_DEC_STREAM_ID:
    case PLDEV_CNTL_IOCTL_SET_DEC_OUTPUT_MODE:
    case PLDEV_CNTL_IOCTL_SET_DEC_RESOLUTION:
    case PLDEV_CNTL_IOCTL_SET_DEC_STOP_MODE:
    case PLDEV_CNTL_IOCTL_SET_DEC_TRICK_MODE:
    case PLDEV_CNTL_IOCTL_SET_DEC_DISP_ATTRIB:
    case PLDEV_CNTL_IOCTL_SET_DEC_DISP_MODE:
    case PLDEV_CNTL_IOCTL_SET_DEC_AUDIO_CONTROL:
    case PLDEV_CNTL_IOCTL_SET_DEC_DATA_XFER_CONTROL:
    case PLDEV_CNTL_IOCTL_SET_DEC_ALL_SETTINGS:
    case PLDEV_CNTL_IOCTL_SET_DEC_AUD_OUTPUT_VOL:

    case PLDEV_CNTL_IOCTL_SET_BEGIN_CHANNEL_CHANGE:
    case PLDEV_CNTL_IOCTL_SET_END_CHANNEL_CHANGE:
    case PLDEV_CNTL_IOCTL_SET_656_OUT_CNTL:
    case PLDEV_CNTL_IOCTL_SET_I2S_OUT_CNTL:
    case PLDEV_CNTL_IOCTL_SET_VOU_CONFIG:
    case PLDEV_CNTL_IOCTL_SET_AOU_CONFIG:
    case PLDEV_CNTL_IOCTL_SET_POWER_STATE:
    case PLDEV_CNTL_IOCTL_SET_ERROR_RECOVERY:
    case PLDEV_CNTL_IOCTL_SET_OSD_MEMORY:
#ifdef SUPPORT_VI_OSD
    case PLDEV_CNTL_IOCTL_SET_ENC_OSD_SETTINGS:
#endif
    case PLDEV_CNTL_IOCTL_SET_DECIMATION:
    case PLDEV_CNTL_IOCTL_SET_DEINTERLACE:
    case PLDEV_CNTL_IOCTL_SET_EXTERNAL_TRIGGER:
    case PLDEV_CNTL_IOCTL_SET_PTS_COUNTER_RESET:
    case PLDEV_CNTL_IOCTL_SET_RATE_CONTROL_EX:
    case PLDEV_CNTL_IOCTL_SET_LARGE_COMPRESS_BUFFER_CONTROL:
    case PLDEV_CNTL_IOCTL_SET_ENABLE_LOW_BITRATE_MODE:
    case PLDEV_CNTL_IOCTL_SET_ENABLE_AV_DISCARD_CONTROL:
    case PLDEV_CNTL_IOCTL_SET_USE_SW_PTS:

    case PLDEV_CNTL_IOCTL_SET_GPIO_BIT_DIRECTION:
    case PLDEV_CNTL_IOCTL_SET_GPIO_BIT_VALUE:

    case PLDEV_CNTL_IOCTL_SET_SYS_FUNCTION:
    case PLDEV_CNTL_IOCTL_SET_SYS_LINK:
    case PLDEV_CNTL_IOCTL_SET_SYS_CONFIG:
    {
        IMpegCodec *pMpegCodec = CDevice_getMpegCodec(pComponent->m_pDevice);

        pl330_trace(PL330_TRACE_IOCTL, "%s pMpegCodec->Set cmd=%u\n", __func__, cmd);
        /* call codec/cqlcodec.c CQLCodec_Set() */
        qpStatus = pMpegCodec->Set(pMpegCodec,
                                   (TQP_REFGUID) & PROPSETID_QPCODEC_CONTROL,
                                   pComponent->m_hTask,
                                   nr - PLDEV_CNTL_IOCTL_SET,
                                   (TQP_PVOID) arg,
                                   QP_NULL,
                                   nSize
                                   );
        break;
    }

    case PLDEV_CNTL_IOCTL_GET_SYSTEM_CONTROL:
    case PLDEV_CNTL_IOCTL_GET_PICTURE_RESOLUTION:
    case PLDEV_CNTL_IOCTL_GET_INPUT_CONTROL:
    case PLDEV_CNTL_IOCTL_GET_RATE_CONTROL:
    case PLDEV_CNTL_IOCTL_GET_VBR_BIT_RATE:
    case PLDEV_CNTL_IOCTL_GET_GOP_LOOP_FILTER:
    case PLDEV_CNTL_IOCTL_GET_ET_CONTROL:
    case PLDEV_CNTL_IOCTL_GET_FILTER_CONTROL:
    case PLDEV_CNTL_IOCTL_GET_OUT_PICTURE_RESOLUTION:
    case PLDEV_CNTL_IOCTL_GET_BLOCK_SIZE:
    case PLDEV_CNTL_IOCTL_GET_AUDIO_CONTROL_PARAM:
    case PLDEV_CNTL_IOCTL_GET_AUDIO_CONTROL_EX:
    case PLDEV_CNTL_IOCTL_GET_VIOSD_TABLE_ADDR:
    case PLDEV_CNTL_IOCTL_GET_GET_CUR_VID_BUF_INFO:
    case PLDEV_CNTL_IOCTL_GET_MP4_VID_NUM_BLOCK:
    case PLDEV_CNTL_IOCTL_GET_AUDIO_ENHANCEMENT:
    case PLDEV_CNTL_IOCTL_GET_ENC_STOP_MODE:
    case PLDEV_CNTL_IOCTL_GET_INDEX_CAP_FREQ:
    case PLDEV_CNTL_IOCTL_GET_ENC_VID_PADDING:
    case PLDEV_CNTL_IOCTL_GET_ENC_FREEZE_VIDEO:
    case PLDEV_CNTL_IOCTL_GET_ENC_STILL_VIDEO_INPUT:
    case PLDEV_CNTL_IOCTL_GET_ENC_CAP_MODE:
    case PLDEV_CNTL_IOCTL_GET_ENC_MJPEG_QUALITY:
    case PLDEV_CNTL_IOCTL_GET_ENC_MJPEG_FRAME_BUF:
    case PLDEV_CNTL_IOCTL_GET_ENC_ALL_SETTINGS:
    case PLDEV_CNTL_IOCTL_GET_ENC_AUD_INPUT_VOL:

    case PLDEV_CNTL_IOCTL_GET_DEC_STREAM_ID:
    case PLDEV_CNTL_IOCTL_GET_DEC_OUTPUT_MODE:
    case PLDEV_CNTL_IOCTL_GET_DEC_RESOLUTION:
    case PLDEV_CNTL_IOCTL_GET_DEC_STOP_MODE:
    case PLDEV_CNTL_IOCTL_GET_DEC_TRICK_MODE:
    case PLDEV_CNTL_IOCTL_GET_DEC_DISP_ATTRIB:
    case PLDEV_CNTL_IOCTL_GET_DEC_DISP_MODE:
    case PLDEV_CNTL_IOCTL_GET_DEC_GET_PLAY_INFO:
    case PLDEV_CNTL_IOCTL_GET_DEC_AUDIO_CONTROL:
    case PLDEV_CNTL_IOCTL_GET_DEC_DATA_XFER_CONTROL:
    case PLDEV_CNTL_IOCTL_GET_DEC_ALL_SETTINGS:
    case PLDEV_CNTL_IOCTL_GET_DEC_AUD_OUTPUT_VOL:

    case PLDEV_CNTL_IOCTL_GET_ENCODER_STATUS:
    case PLDEV_CNTL_IOCTL_GET_DECODER_STATUS:
    case PLDEV_CNTL_IOCTL_GET_656_OUT_CNTL:
    case PLDEV_CNTL_IOCTL_GET_I2S_OUT_CNTL:
    case PLDEV_CNTL_IOCTL_GET_VOU_CONFIG:
    case PLDEV_CNTL_IOCTL_GET_AOU_CONFIG:
    case PLDEV_CNTL_IOCTL_GET_POWER_STATE:
    case PLDEV_CNTL_IOCTL_GET_ERROR_RECOVERY:
    case PLDEV_CNTL_IOCTL_GET_OSD_MEMORY:
    case PLDEV_CNTL_IOCTL_GET_DECIMATION:
    case PLDEV_CNTL_IOCTL_GET_DEINTERLACE:
    case PLDEV_CNTL_IOCTL_GET_EXTERNAL_TRIGGER:
    case PLDEV_CNTL_IOCTL_GET_PTS_COUNTER_RESET:
    case PLDEV_CNTL_IOCTL_GET_RATE_CONTROL_EX:
    case PLDEV_CNTL_IOCTL_GET_LARGE_COMPRESS_BUFFER_CONTROL:
    case PLDEV_CNTL_IOCTL_GET_ENABLE_LOW_BITRATE_MODE:
    case PLDEV_CNTL_IOCTL_GET_ENABLE_AV_DISCARD_CONTROL:
    case PLDEV_CNTL_IOCTL_GET_USE_SW_PTS:

    case PLDEV_CNTL_IOCTL_GET_GPIO_BIT_DIRECTION:
    case PLDEV_CNTL_IOCTL_GET_GPIO_BIT_VALUE:

    case PLDEV_CNTL_IOCTL_GET_SYS_FUNCTION:
    case PLDEV_CNTL_IOCTL_GET_SYS_LINK:
    case PLDEV_CNTL_IOCTL_GET_SYS_CONFIG:
    case PLDEV_CNTL_IOCTL_GET_CODEC_ERROR:
    {
        IMpegCodec *pMpegCodec = CDevice_getMpegCodec(pComponent->m_pDevice);
        TQP_UINT32 dwGot = 0;

        pl330_trace(PL330_TRACE_IOCTL, "%s pMpegCodec->Get cmd=%u\n", __func__, cmd);
        qpStatus = pMpegCodec->Get(pMpegCodec,
                                   (TQP_REFGUID) & PROPSETID_QPCODEC_CONTROL,
                                   pComponent->m_hTask,
                                   nr - PLDEV_CNTL_IOCTL_GET,
                                   (TQP_PVOID) arg,
                                   (TQP_PVOID) arg,
                                   &dwGot
                                   );
        break;
    }


        /////////////////////////////////////////////////////////////////////////////
        //
        //
        // diagnostic interface
        //
    case PLDEV_DIAG_IOCTL_FIRMWARE_DOWNLOAD_PTR:
    {
        PQPCODEC_DIAG_FW_DOWNLOAD_PTR pFW = (PQPCODEC_DIAG_FW_DOWNLOAD_PTR) arg;
        QPCODEC_DIAG_FW_DOWNLOAD_PTR FW;
        ICodecLib *pCodecLib = CDevice_getCodecLib(pComponent->m_pDevice);

        pl330_trace(PL330_TRACE_IOCTL, "%s firmware download\n", __func__);
        memcpy(&FW,
               pFW,
               sizeof (QPCODEC_DIAG_FW_DOWNLOAD_PTR)
               );

#ifdef USE_KMALLOC_INSTEAD_OF_VMALLOC
        FW.pFW = kmalloc(FW.FWSize, GFP_KERNEL);
#else
        FW.pFW = vmalloc(FW.FWSize);
#endif
        if (!FW.pFW)
        {
            qpStatus = QPERR_NOMEM;
        }
        else
        {
            if (copy_from_user(FW.pFW, pFW->pFW, pFW->FWSize))
            {
                qpStatus = QPERR_NOMEM;
            }
            else
            {
                qpStatus = pCodecLib->Set(pCodecLib,
                                          (TQP_REFGUID) & PROPSETID_QPCODEC_DIAG,
                                          pComponent->m_hTask,
                                          nr - PLDEV_DIAG_IOCTL_BASE,
                                          (TQP_PVOID) & FW,
                                          QP_NULL,
                                          nSize
                                          );
            }
#ifdef USE_KMALLOC_INSTEAD_OF_VMALLOC
            kfree(FW.pFW);
#else
            vfree(FW.pFW);
#endif
        }
        break;
    }

    case PLDEV_DIAG_IOCTL_DMA_WRITE_PTR:
    {
        PQPCODEC_DIAG_DMA_PTR_STRUCT pDMA = (PQPCODEC_DIAG_DMA_PTR_STRUCT) arg;
        QPCODEC_DIAG_DMA_PTR_STRUCT DMA;
        ICodecLib *pCodecLib = CDevice_getCodecLib(pComponent->m_pDevice);
        pl330_trace(PL330_TRACE_IOCTL, "%s DMA_WRITE_PTR\n", __func__);

        memcpy(&DMA,
               pDMA,
               sizeof (QPCODEC_DIAG_DMA_PTR_STRUCT)
               );

#ifdef USE_KMALLOC_INSTEAD_OF_VMALLOC
        DMA.pData = kmalloc(DMA.size, GFP_KERNEL);
#else
        DMA.pData = vmalloc(DMA.size);
#endif
        if (!DMA.pData)
        {
            qpStatus = QPERR_NOMEM;
        }
        else
        {
            if (copy_from_user(DMA.pData,
                               pDMA->pData,
                               pDMA->size
                               ))
            {
                qpStatus = QPERR_NOMEM;
            }
            else
            {
                qpStatus = pCodecLib->Set(pCodecLib,
                                          (TQP_REFGUID) & PROPSETID_QPCODEC_DIAG,
                                          pComponent->m_hTask,
                                          nr - PLDEV_DIAG_IOCTL_BASE,
                                          (TQP_PVOID) & DMA,
                                          QP_NULL,
                                          nSize
                                          );
            }
#ifdef USE_KMALLOC_INSTEAD_OF_VMALLOC
            kfree(DMA.pData);
#else
            vfree(DMA.pData);
#endif
        }
        break;
    }

    case PLDEV_DIAG_IOCTL_I2C_WRITE:
    case PLDEV_DIAG_IOCTL_I2C_WRITE_EX:
    case PLDEV_DIAG_IOCTL_MEM_WRITE:
    case PLDEV_DIAG_IOCTL_REG_WRITE:
    case PLDEV_DIAG_IOCTL_HCI_REG_WRITE:
    case PLDEV_DIAG_IOCTL_ENABLE_DRIVER_I2C:
    case PLDEV_DIAG_IOCTL_DISABLE_DRIVER_I2C:
    case PLDEV_DIAG_IOCTL_VIDEO_DECODER_SET:
    {
        ICodecLib *pCodecLib = CDevice_getCodecLib(pComponent->m_pDevice);

        pl330_trace(PL330_TRACE_IOCTL, "%s pCodecLib->Set cmd=%u\n", __func__, cmd);
        qpStatus = pCodecLib->Set(pCodecLib,
                                  (TQP_REFGUID) & PROPSETID_QPCODEC_DIAG,
                                  pComponent->m_hTask,
                                  nr - PLDEV_DIAG_IOCTL_BASE,
                                  (TQP_PVOID) arg,
                                  QP_NULL,
                                  nSize
                                  );
        break;
    }

    case PLDEV_DIAG_IOCTL_DMA_READ_PTR:
    {
        PQPCODEC_DIAG_DMA_PTR_STRUCT pDMA = (PQPCODEC_DIAG_DMA_PTR_STRUCT) arg;
        QPCODEC_DIAG_DMA_PTR_STRUCT DMA;
        ICodecLib *pCodecLib = CDevice_getCodecLib(pComponent->m_pDevice);
        TQP_UINT32 dwGot = 0;

        pl330_trace(PL330_TRACE_IOCTL, "%s DMA_READ_PTR\n", __func__);
        memcpy(&DMA,
               pDMA,
               sizeof (QPCODEC_DIAG_DMA_PTR_STRUCT)
               );

#ifdef USE_KMALLOC_INSTEAD_OF_VMALLOC
        DMA.pData = kmalloc(DMA.size, GFP_KERNEL);
#else
        DMA.pData = vmalloc(DMA.size);
#endif
        if (!DMA.pData)
        {
            qpStatus = QPERR_NOMEM;
        }
        else
        {
            qpStatus = pCodecLib->Get(pCodecLib,
                                      (TQP_REFGUID) & PROPSETID_QPCODEC_DIAG,
                                      pComponent->m_hTask,
                                      nr - PLDEV_DIAG_IOCTL_BASE,
                                      (TQP_PVOID) & DMA,
                                      (TQP_PVOID) & DMA,
                                      &dwGot
                                      );
            if (QP_SUCCESS(qpStatus))
            {
                if (copy_to_user(pDMA->pData,
                                 DMA.pData,
                                 pDMA->size
                                 ))
                {
                    qpStatus = QPERR_NOMEM;
                }
            }
#ifdef USE_KMALLOC_INSTEAD_OF_VMALLOC
            kfree(DMA.pData);
#else
            vfree(DMA.pData);
#endif
        }
        break;
    }

    case PLDEV_DIAG_IOCTL_I2C_READ:
    case PLDEV_DIAG_IOCTL_I2C_READ_EX:
    case PLDEV_DIAG_IOCTL_I2C_WRITE_THEN_READ:
    case PLDEV_DIAG_IOCTL_I2C_WRITE_THEN_READ_EX:
    case PLDEV_DIAG_IOCTL_MEM_READ:
    case PLDEV_DIAG_IOCTL_REG_READ:
    case PLDEV_DIAG_IOCTL_HCI_REG_READ:
    case PLDEV_DIAG_IOCTL_FIRMWARE_COMMAND:
    case PLDEV_DIAG_IOCTL_VIDEO_DECODER_GET:
    {

        ICodecLib *pCodecLib = CDevice_getCodecLib(pComponent->m_pDevice);
        TQP_UINT32 dwGot = 0;

        pl330_trace(PL330_TRACE_IOCTL, "%s pCodecLib->Get cmd=%u\n", __func__, cmd);
        qpStatus = pCodecLib->Get(pCodecLib,
                                  (TQP_REFGUID) & PROPSETID_QPCODEC_DIAG,
                                  pComponent->m_hTask,
                                  nr - PLDEV_DIAG_IOCTL_BASE,
                                  (TQP_PVOID) arg,
                                  (TQP_PVOID) arg,
                                  &dwGot
                                  );
        break;
    }

#ifdef SUPPORT_PL3XX_OSD
    case PLDEV_DIAG_IOCTL_OSD_CONFIG:
        qpStatus = qldiag_osd_param(pComponent->m_pDevice, arg);
        break;

    case PLDEV_DIAG_IOCTL_OSD_CONTROL:
        qpStatus = qldiag_osd_control(pComponent->m_pDevice, arg);
        break;

#ifdef SUPPORT_VI_OSD
    case PLDEV_DIAG_IOCTL_VIOSD_CONTROL:
        qpStatus = qldiag_viosd_control(pComponent->m_pDevice, arg);
        break;
#endif
#endif


    case PLDEV_DIAG_IOCTL_VIDEOPROP_SET:
        pl330_trace(PL330_TRACE_IOCTL, "%s videoprop_set\n", __func__);
        qpStatus = qldiag_videoprop_set(pComponent->m_pDevice, arg);
        break;
    case PLDEV_DIAG_IOCTL_VIDEOPROP_GET:
        pl330_trace(PL330_TRACE_IOCTL, "%s videoprop_get\n", __func__);
        qpStatus = qldiag_videoprop_get(pComponent->m_pDevice, arg);
        break;
    case PLDEV_DIAG_IOCTL_VIDEO_DECODER_SOURCE_SET:
    {
        ICodecLib *pCodecLib = CDevice_getCodecLib(pComponent->m_pDevice);
        //			TQP_UINT32	dwGot = 0;
        TQP_UINT32 *p = (TQP_UINT32 *) arg;
        VIDEO_SOURCE newarg;
        pl330_trace(PL330_TRACE_IOCTL, "%s decoder_source_set\n", __func__);
        /* map dwValue to VIDEO_SOURCE */
        switch (*p)
        {
        case 0: /* AUTO/HDMI */
            newarg = VIDEO_SOURCE_DIGITAL1;
            break;
        case 1: /* Composite */
            newarg = VIDEO_SOURCE_CVIDEO1;
            break;
        case 2: /* S-Video */
            newarg = VIDEO_SOURCE_SVIDEO1;
            break;
        case 3: /* Component */
            newarg = VIDEO_SOURCE_NVIDEO1;
            break;
        default:
            qpStatus = QPERR_PARMS;
            return (QpToLinuxKerStatus(qpStatus));
        }
        qpStatus = pCodecLib->Set(pCodecLib,
                                  (TQP_REFGUID) & PROPSETID_QPCODEC_DIAG,
                                  pComponent->m_hTask,
                                  nr - PLDEV_DIAG_IOCTL_BASE,
                                  (TQP_PVOID) & newarg,
                                  QP_NULL,
                                  nSize
                                  );
        break;
    }
    case PLDEV_DIAG_IOCTL_VIDEO_DECODER_SOURCE_GET:
    {
        ICodecLib *pCodecLib = CDevice_getCodecLib(pComponent->m_pDevice);
        TQP_UINT32 dwGot = 0;
        TQP_UINT32 *p = (TQP_UINT32 *) arg;
        VIDEO_SOURCE newarg;

        pl330_trace(PL330_TRACE_IOCTL, "%s decoder_source_get\n", __func__);
        qpStatus = pCodecLib->Get(pCodecLib,
                                  (TQP_REFGUID) & PROPSETID_QPCODEC_DIAG,
                                  pComponent->m_hTask,
                                  nr - PLDEV_DIAG_IOCTL_BASE,
                                  QP_NULL,
                                  (TQP_PVOID) & newarg,
                                  &dwGot
                                  );
        /* map VIDEO_SOURCE to dwValue*/
        switch (newarg)
        {
        case VIDEO_SOURCE_DIGITAL1: /* AUTO/HDMI */
            *p = 0;
            break;
        case VIDEO_SOURCE_CVIDEO1: /* Composite */
            *p = 1;
            break;
        case VIDEO_SOURCE_SVIDEO1: /* S-Video */
            *p = 2;
            break;
        case VIDEO_SOURCE_NVIDEO1: /* Component */
            *p = 3;
            break;
        default:
            *p = 0;
            qpStatus = QPERR_PARMS;
            return (QpToLinuxKerStatus(qpStatus));
        }
        break;
    }
    case PLDEV_DIAG_IOCTL_FIRMWARE_RELOAD:
    case PLDEV_DIAG_IOCTL_RESET_ARM:
    {
        ICodecLib *pCodecLib = CDevice_getCodecLib(pComponent->m_pDevice);
        qpStatus = pCodecLib->Set(pCodecLib,
                                  (TQP_REFGUID) & PROPSETID_QPCODEC_DIAG,
                                  pComponent->m_hTask,
                                  nr - PLDEV_DIAG_IOCTL_BASE,
                                  QP_NULL,
                                  QP_NULL,
                                  nSize
                                  );
        break;
    }

    case PLDEV_DIAG_IOCTL_DEBUGLEVEL_SET:
    {
        TQP_UINT32 *p = (TQP_UINT32 *) arg;
        TQP_UINT32 func, value;

        pl330_trace(PL330_TRACE_IOCTL, "%s debuglevel_set\n", __func__);

        func = ((*p) & 0xFFFF0000) >> 16;
        value = (*p) & 0xFFFF;

        switch (func)
        {
        case 0:
            qpStatus = QPDbgMSetCompLevel(g_PLV4LDevDBGCompID, value);
            break;
        case 1:
            qpStatus = QPDbgMSetCompLevel(g_QPCodecDBGCompID, value);
            break;
        case 2:
            qpStatus = QPDbgMSetCompLevel(g_QPUsbDBGCompID, value);
            break;
        case 3:
            qpStatus = QPDbgMSetCompLevel(g_QPPCIeDBGCompID, value);
            break;
        case 4:
            qpStatus = QPDbgMSetCompLevel(g_QPPeripheralDBGCompID, value);
            break;
#ifdef QP_DBG_TO_BUFFER
        case 10:
            qpStatus = QPDbgMSetBufCompLevel(g_PLV4LDevDBGCompID, value);
            break;
        case 11:
            qpStatus = QPDbgMSetBufCompLevel(g_QPCodecDBGCompID, value);
            break;
        case 12:
            qpStatus = QPDbgMSetBufCompLevel(g_QPUsbDBGCompID, value);
            break;
        case 13:
            qpStatus = QPDbgMSetBufCompLevel(g_QPPCIeDBGCompID, value);
            break;
        case 14:
            qpStatus = QPDbgMSetBufCompLevel(g_QPPeripheralDBGCompID, value);
            break;
            //case 15:
            //	qpStatus = QPDbgMSetBufCompLevel(PaulDBGCompID, value);
            //	break;					
        case 20:
            QPDbgMTimeStampPrinting(1);
            qpStatus = QPDbgMSetBufInit();
            break;
        case 21:
            qpStatus = QPDbgMSetBufDump("/tmp/kernellog.txt");
            break;
#endif
        default:
            qpStatus = QPERR_PARMS;
            break;
        }
        break;
    }
    case PLDEV_DIAG_IOCTL_DEBUGLEVEL_GET:
    {
        TQP_UINT32 *p = (TQP_UINT32 *) arg;
        TQP_UINT32 func;
        unsigned long value;

        pl330_trace(PL330_TRACE_IOCTL, "%s debuglevel_get\n", __func__);

        func = ((*p) & 0xFFFF0000) >> 16;
        switch (func)
        {
        case 0:
            qpStatus = QPDbgMGetCompLevel(g_PLV4LDevDBGCompID, &value);
            break;
        case 1:
            qpStatus = QPDbgMGetCompLevel(g_QPCodecDBGCompID, &value);
            break;
        case 2:
            qpStatus = QPDbgMGetCompLevel(g_QPUsbDBGCompID, &value);
            break;
        case 3:
            qpStatus = QPDbgMGetCompLevel(g_QPPCIeDBGCompID, &value);
            break;
        case 4:
            qpStatus = QPDbgMGetCompLevel(g_QPPeripheralDBGCompID, &value);
            break;
#ifdef QP_DBG_TO_BUFFER
        case 10:
            qpStatus = QPDbgMGetBufCompLevel(g_PLV4LDevDBGCompID, &value);
            break;
        case 11:
            qpStatus = QPDbgMGetBufCompLevel(g_QPCodecDBGCompID, &value);
            break;
        case 12:
            qpStatus = QPDbgMGetBufCompLevel(g_QPUsbDBGCompID, &value);
            break;
        case 13:
            qpStatus = QPDbgMGetBufCompLevel(g_QPPCIeDBGCompID, &value);
            break;
        case 14:
            qpStatus = QPDbgMGetBufCompLevel(g_QPPeripheralDBGCompID, &value);
            break;
            //case 15:
            //	qpStatus = QPDbgMGetBufCompLevel(PaulDBGCompID, &value);
            //	break;		
#endif
        default:
            qpStatus = QPERR_PARMS;
            break;
        }
        if (qpStatus == QPERR_SUCCESS)
            *p = (func << 16) | (value & 0xFFFF);
        break;
    }
    default:
        pl330_trace(PL330_TRACE_VIDIOC,"%s ioctl not support %x\n", __func__, cmd);
        qpStatus = QPERR_NOTSUPP;
        break;
    }

    return (QpToLinuxKerStatus(qpStatus));
}

int CComponent_Mmap(CComponent *This,
                    struct vm_area_struct *vma
                    )
{
    CPort *pPort;
    COMPONENT_PORT_TYPE type;
    struct v4l2_buffer *pV4l2Buf;
    long lErr = 0;
    unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
    int nIndex;
    unsigned long ulStart = 0;
    unsigned long ulAddr = 0;
    unsigned long ulSize = 0;

    // find the port this buffer belongs to
    //
    type = offset >> PORT_MMAP_BUF_TYPE_SHIFT;
    if (type >= MAX_COMPONENT_PORT)
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                     "%s vma->vm_pgoff(0x%x) not found!!!\n",
                     __FUNCTION__,
                     vma->vm_pgoff
                     ));
        return (-EINVAL);
    }

    pPort = &This->m_Ports[type];
    if (V4L2_MEMORY_MMAP != pPort->m_memoryType)
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                     "%s type(%d) invalid memory type(%d)!!!\n",
                     __FUNCTION__,
                     type,
                     pPort->m_memoryType
                     ));
        return (-EINVAL);
    }

    offset &= ~PORT_MMAP_BUF_TYPE_MASK;
    offset = offset >> PAGE_SHIFT;

    for (nIndex = 0; nIndex < pPort->m_frame_nbs; nIndex++)
    {
        pV4l2Buf = &pPort->m_v4l2Buffers[nIndex];
        if ((pV4l2Buf->m.offset >> PAGE_SHIFT) == offset)
        {
            break;
        }
    }

    if (nIndex >= pPort->m_frame_nbs)
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                     "%s vma->vm_pgoff(0x%x) not found!!!\n",
                     __FUNCTION__,
                     vma->vm_pgoff
                     ));
        return (-EINVAL);
    }

    // VM_IO marks the area as being an mmaped region for I/O to a
    // device. It also prevents the region from being core dumped.
    //
    vma->vm_flags |= VM_IO;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
#define VM_RESERVED  (VM_DONTEXPAND | VM_DONTDUMP)
#endif
    vma->vm_flags |= VM_RESERVED;

#ifdef USE_KMALLOC_INSTEAD_OF_VMALLOC
    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif

    ulStart = vma->vm_start;
    ulAddr = (unsigned long) (pPort->m_pBufMem + pV4l2Buf->m.offset);
    ulSize = vma->vm_end - vma->vm_start;

    while (ulSize > 0)
    {
#ifdef USE_KMALLOC_INSTEAD_OF_VMALLOC
        if ((lErr = remap_pfn_range(vma,
                                    ulStart,
                                    __pa((void *) ulAddr) >> PAGE_SHIFT,
                                    PAGE_SIZE,
                                    PAGE_SHARED
                                    )) < 0
                )
#else
        if ((lErr = vm_insert_page(vma,
                                   ulStart,
                                   vmalloc_to_page((void *) ulAddr)
                                   )) < 0
                )
#endif
        {
            QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
                         "%s vm_insert_page() start(%d) addr(%d) failed(%d)!!!\n",
                         __FUNCTION__,
                         ulStart,
                         ulAddr,
                         lErr
                         ));
            return (lErr);
        }

        ulStart += PAGE_SIZE;
        ulAddr += PAGE_SIZE;
        ulSize -= PAGE_SIZE;
    }

    vma->vm_ops = &plv4l_mmap_ops;
    vma->vm_private_data = pV4l2Buf;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_TRACE),
                 "%s vma->vm_pgoff(0x%x) found buf(%d) lErr(%d)\n",
                 __FUNCTION__,
                 vma->vm_pgoff,
                 nIndex,
                 lErr
                 ));

    return (lErr);
}

int CComponent_Poll(CComponent *This,
                    struct file *filp, poll_table *wait
                    )
{
    CPort *pPort;
    //	COMPONENT_PORT_TYPE	type;
    EQP_Errors qpStatus;
    int i, mask = 0, IsEmpty;

    if (TQP_NULL_HANDLE == This->m_hObject)
        return POLLERR;

    poll_wait(filp, &(This->wait_frame), wait);

    for (i = 0; i < MAX_COMPONENT_PORT; i++)
    {
        pPort = &This->m_Ports[i];
        if (pPort->m_valid)
        {
            qpStatus = CPort_CheckBuf(pPort, &IsEmpty);
            if (qpStatus == QPERR_SUCCESS)
            {
                if (IsEmpty == 0)
                {
                    //						printk("===> [Port %d dir %d ] is not empty\n", i, pPort->m_dir);
                    switch (pPort->m_dir)
                    {
                    case PORT_DIR_READ:
                        mask |= POLLIN | POLLRDNORM;
                        break;

                    case PORT_DIR_WRITE:
                        mask |= POLLOUT | POLLWRNORM;
                        break;

                    case PORT_DIR_NONE:
                    default:
                        mask |= 0;
                    }
                }
                else
                {
                    //						printk("===> [Port %d dir %d ] is empty or error\n", i, pPort->m_dir);
                }
            }
            else
            {
                //					printk("===> [Port %d dir %d ] is empty or error\n", i, pPort->m_dir);
            }
        }
        else
        {
            continue;
        }
    }
    return (mask);
}

