///////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2011 Sigma Designs, Inc.  All Rights Reserved.
//
// Drventry.c
//
// Description:
//
//  PCIe driver entry point
//
// History:
//	Jan 27, 2010	(dt) created
//
// Authors: (dt) David Tao
//
///////////////////////////////////////////////////////////////////////////////


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/slab.h>

#include "qptypes.h"
#include "qpdbgmgr.h"
#include "qpmemmgr.h"
#include "qposmgr.h"
#include "qptimemgr.h"
#include "qpmgrspriv.h"

#include "pci_config.h"

#include "plv4ldevversion.h"
#include "cdevice.h"

#include "pl330.h"
#include "pl330-dbg.h"

unsigned int pl330_trace_param = 0x0f;
unsigned int adv7604_trace_level = 0;
/////////////////////////////////////////////////////////////////////////////
//
//

#define PL330_PCI_VENDER_ID		0x1AF2
#define PL330_PCI_DEVICE_ID		0xA001


/////////////////////////////////////////////////////////////////////////////
//
//

struct semaphore g_module_lock;

static struct pci_device_id	plpci_table[] =
{
	{
		.vendor = PL330_PCI_VENDER_ID, .device = PL330_PCI_DEVICE_ID,
		.subvendor = 0x1af2, .subdevice = 0xa001,
	},
	{}
};


MODULE_DEVICE_TABLE(pci, plpci_table);

#ifndef MODULE_NAME
#define MODULE_NAME "AVerMedia PCIe H.264 codec"
#endif


static unsigned char plpci_get_revision(struct pci_dev *pdev)
{
    TQP_UINT8 revision;

    pci_read_config_byte(pdev, PCI_REVISION_ID, &revision);
    return revision;
}



/////////////////////////////////////////////////////////////////////////////
//
//

static int plpci_probe(struct pci_dev *pdev,
					   const struct pci_device_id *id
					   )
{
	long				lErr = 0;
	PLV4LDEV_TYPE		devType;
    CDevice				*pDevice = QP_NULL;
	PLV4LDEV_PARENT_DEV	parent_dev;
	EQP_Errors			qpStatus;
	TQP_UINT32			i = 0;
	unsigned char		revision;
        resource_size_t mmio_base;
	resource_size_t mmio_length;
	unsigned long mmio_flags;

	// register the component for debug output
	//
	QPDbgMRegComp(g_PLV4LDevDBGCompID,
				  "QP V4LDev"
				  );
	QPDbgMRegComp(g_QPCodecDBGCompID,
				  "QP CodecLib"
				  );
	QPDbgMRegComp(g_QPPCIeDBGCompID,
				  "QP PCIe"
				  );
	QPDbgMRegComp(g_QPPeripheralDBGCompID,
				  "QP Peripheral"
				  );

	QPDbgMSetCompLevel(g_PLV4LDevDBGCompID,
					   QPDBGM_LEVEL_ERROR
					   );
	QPDbgMSetCompLevel(g_QPCodecDBGCompID,
					   QPDBGM_LEVEL_ERROR
					   );
	QPDbgMSetCompLevel(g_QPPCIeDBGCompID,
					   QPDBGM_LEVEL_ERROR
					   );
	QPDbgMSetCompLevel(g_QPPeripheralDBGCompID,
					   QPDBGM_LEVEL_ERROR
					   );

	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
				 "%s() (pci_dev=%08x)\n",
				 __FUNCTION__,
				 pdev
				 ));

	pci_enable_device(pdev);

	revision = plpci_get_revision(pdev);
	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
				 "%s() (revision=%x)\n",
				 __FUNCTION__,
				 revision
				 ));

	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
				 "%s() driver v%d.%d.%d\n",
				 __FUNCTION__,
				 PLDRV_VERSION_MAJOR,
				 PLDRV_VERSION_MINOR,
				 PLDRV_VERSION_BRANCH
				 ));

	if (revision == 0x42)
	{
		return (-ENODEV);
	}

	lErr = pci_set_mwi(pdev);
	if (lErr)
	{
		goto ERROR;
	}

	lErr = pci_request_regions(pdev,
							   MODULE_NAME
							   );
	if (lErr < 0)
	{
		goto ERROR1;
	}

	mmio_base = pci_resource_start(pdev,  0);
	mmio_length = pci_resource_len(pdev,  0);
	mmio_flags = pci_resource_flags(pdev,  0);
	pl330_trace(PL330_TRACE_PROBE, "%s mmio_base:%x\n", __func__, mmio_base);
	pl330_trace(PL330_TRACE_PROBE, "%s mmio_length:%x\n", __func__, mmio_length);
	pl330_trace(PL330_TRACE_PROBE, "%s mmio_flags:%x\n", __func__, mmio_flags);

	pci_set_master(pdev);

	parent_dev.p_pciedev = pdev;
    devType = PLV4LDEV_TYPE_QL300_PCIE_EVK;

	switch (devType)
	{
		case PLV4LDEV_TYPE_QL300_PCIE_EVK:

			down(&g_module_lock);
			// find an empty slot for this device
			//
			for (i = 0; i < PLV4LDEV_MAX_DEVICES; i++)
			{
				if (QP_NULL == g_pPlV4lDevices[i])
				{
					// create and int the device
					//
					pDevice = (CDevice *)kmalloc(sizeof(CDevice),
												 GFP_KERNEL
												 );
					if (pDevice)
					{
						memset(pDevice,
							   0,
							   sizeof(CDevice)
							   );

						if (!CDevice_Constructor(pDevice,
												 i,
												 devType,
												 &parent_dev,
												 &pdev->dev,
												 (TQP_PVOID)0x100000,
												 (TQP_PVOID)0,
												 0x100000,
												 0
												 ))
						{
							kfree((void *)pDevice);
							pDevice = QP_NULL;
							lErr = -ENOMEM;
						}
						else
						{
							qpStatus = CDevice_Create(pDevice);
							if (QP_SUCCESS(qpStatus))
							{
								g_pPlV4lDevices[i] = pDevice;
								pci_set_drvdata(pdev,
												(void *)pDevice
												);
							}
							else
							{
								CDevice_Destructor(pDevice);
								kfree((void *)pDevice);
								pDevice = QP_NULL;
								lErr = -ENOMEM;
							}
						}
					}
					else
					{
						lErr = -ENOMEM;
					}
					break;
				}
			}

			if (PLV4LDEV_MAX_DEVICES == i)
			{
				// no empty slot available
				//
				lErr = -ENOMEM;
			}

			up(&g_module_lock);
			break;

		default:
			lErr = -ENODEV;
			goto ERROR1;
    }

	if (lErr != 0)
	{
		goto ERROR1;
	}

END:
	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
				 "%s() lErr(%d)\n",
				 __FUNCTION__,
				 lErr
				 ));
	return (lErr);

ERROR1:
	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
				 "%s() Err1\n",
				 __FUNCTION__
				 ));
    pci_clear_mwi(pdev);

ERROR:
	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_ERROR),
				 "%s() Err\n",
				 __FUNCTION__
				 ));
	pci_release_regions (pdev);
	pci_disable_device(pdev);
	goto END;
}



static void plpci_remove(struct pci_dev *pdev)
{
    CDevice	*pDevice = (CDevice *)pci_get_drvdata(pdev);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
				 "%s() (pdev=%08x) (pDevice=%08x)\n",
				 __FUNCTION__,
				 pdev,
				 pDevice
				 ));

	down(&g_module_lock);

	if (pDevice)
	{
		g_pPlV4lDevices[pDevice->m_id] = QP_NULL;
		CDevice_Close(pDevice);
		CDevice_Destructor(pDevice);
		kfree((void *)pDevice);
    }

	up(&g_module_lock);

	pci_release_regions(pdev);
	pci_disable_device(pdev);

	pci_set_drvdata(pdev,
					NULL
					);
}



int plpci_suspend(struct pci_dev *pdev,
				  pm_message_t state
				  )
{
    CDevice	*pDevice = (CDevice *)pci_get_drvdata(pdev);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
				 "%s() \n",
				 __FUNCTION__
				 ));
	if (pDevice)
	{
		CDevice_PowerDown(pDevice);
    }

	pci_save_state(pdev);

	pci_set_power_state(pdev,
						PCI_D3hot
						);

    return 0;
}


int plpci_resume(struct pci_dev *pdev)
{
    CDevice	*pDevice = (CDevice *)pci_get_drvdata(pdev);

	QPDbgMPrintf((QPDBGMCOMPLVL(g_PLV4LDevDBGCompID, QPDBGM_LEVEL_WARN),
				 "%s() \n",
				 __FUNCTION__
				 ));

	pci_restore_state(pdev);

	pci_set_power_state(pdev,
						PCI_D0
						);
	if (pDevice)
	{
		CDevice_PowerUp(pDevice);
	}
	return (0);
}



static struct pci_driver pci_driver =
{
	.name		= MODULE_NAME,
	.id_table	= plpci_table,
	.probe		= plpci_probe,
	.suspend	= plpci_suspend,
	.resume		= plpci_resume,

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0))
	.remove		= __devexit_p(plpci_remove),
#else
	.remove		= plpci_remove,
#endif
};



static int __init plpci_init_module(void)
{
	long	lErr;

	QPMgrsInit();

	printk("%s() \n",
		   __FUNCTION__
		   );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37)
	sema_init(&g_module_lock, 1);
#else
	init_MUTEX(&g_module_lock);
#endif

    lErr = pci_register_driver(&pci_driver);

	return (lErr);
}


static void __exit plpci_fini_module(void)
{
	pci_unregister_driver(&pci_driver);

	printk("%s() \n",
		   __FUNCTION__
		   );

	// terminate OS manager
	//
	QPMgrsTerm();
	return;
}

MODULE_LICENSE("GPL");

//module_param_named(hdcp, hdcp_a0, uint, S_IRUGO | S_IWUSR);
//MODULE_PARM_DESC(hdcp, "1: HDCP on, 0: HDCP off");

//module_param_named(trace, pl330_trace_param, uint, S_IRUGO | S_IWUSR);
//MODULE_PARM_DESC(trace, "Trace level bitmask");

//module_param_named(viu_trace, adv7604_trace_level, uint, S_IRUGO | S_IWUSR);
//MODULE_PARM_DESC(viu_trace, "VIU trace level");

//module_param_named(de_h_start, de_h_start, uint, S_IRUGO | S_IWUSR);
//MODULE_PARM_DESC(de_h_start, "A control to vary the leading edge positoin of the DE signal output by the CP core. \
//This value in a 2's complement format.The unit of de_h_start (10 bits) is one pixel clock.");

//module_param_named(de_h_end, de_h_end, uint, S_IRUGO | S_IWUSR);
//MODULE_PARM_DESC(de_h_end, "A control to vary the trailing edge positoin of the DE signal output by the CP core. \
//This value in a 2's complement format.The unit of de_h_end (10 bits) is one pixel clock.");

//module_param_named(de_v_start, de_v_start, uint, S_IRUGO | S_IWUSR);
//MODULE_PARM_DESC(de_v_start, "A control to vary the start position of the VBI region.\
//This value represented in a 2's complement format.Ths unit of de_v_start (4 bits) is one pixel clock.");

//module_param_named(de_v_end, de_v_end, uint, S_IRUGO | S_IWUSR);
//MODULE_PARM_DESC(de_v_end, "A control to vary the end position of the VBI region.\
//This value represented in a 2's complement format.The unit of de_v_end (4 bits) is one pixel clock.");


module_init(plpci_init_module);
module_exit(plpci_fini_module);
