#include "pl330.h"
#include "pl330-dbg.h"

#include "qpmpgcodec.h"
#include "qplinuxkermap.h"
#include "pldevioctl.h"
#include "cadv7604.h"

#define SNDRV_PORT_TYPE			COMPONENT_PORT_PCM_OUT
#define SNDRV_PORT_OPEN_TYPE	QPMPGCODEC_ENC_PCM_OUT

#define SNDRV_DRIVER_STR	"C353"

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/workqueue.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/control.h>
#include <sound/initval.h>

static void snd_pl330_on_buffer_done(CComponent *component)
{
	struct pl330_audio_dev *chip =
		component->m_pDevice->m_pAudioDev;
	struct v4l2_buffer *buf;

	struct snd_pcm_runtime *runtime;
	unsigned char *cp;
	int done_size;
	int frames;
	unsigned int oldptr;
	unsigned int stride;
	unsigned int cnt;
	bool period_elapsed = false;

	pl330_trace(PL330_TRACE_PCM, "%s()\n", __func__);

	if(!chip->buf && !chip->running)
		return;

	if(!atomic_read(&chip->stream_started))
		return;

	if(!chip->substream)
		return;

	runtime = chip->substream->runtime;
	buf = chip->buf;

	snd_pl330_dqbuf(component, buf);

	done_size = buf->bytesused;
	frames = bytes_to_frames(runtime, done_size);
	stride = runtime->frame_bits >> 3;
	oldptr = chip->hwptr;

	/* buffer address */
	cp = component->m_Ports[buf->sequence].m_pBufMem +
		buf->m.offset;

	if(oldptr + frames >= runtime->buffer_size)
	{
		cnt = runtime->buffer_size - oldptr;
		memcpy(runtime->dma_area + oldptr * stride,
				cp, cnt * stride);
		memcpy(runtime->dma_area, cp + cnt * stride,
				done_size - cnt * stride);
	}
	else
	{
		memcpy(runtime->dma_area + oldptr * stride,
				cp, done_size);
	}

	chip->hwptr += frames;
	if(chip->hwptr >= runtime->buffer_size)
		chip->hwptr -= runtime->buffer_size;

	chip->capture_transfer_done += frames;
	if(chip->capture_transfer_done >= runtime->period_size)
	{
		chip->capture_transfer_done -= runtime->period_size;
		period_elapsed = true;
	}

	if(period_elapsed)
		snd_pcm_period_elapsed(chip->substream);

	snd_pl330_qbuf(component, buf);
}

static int snd_pl330_port_open(struct pl330_audio_dev *chip)
{
	COMPONENT_PORT_OPEN port_open;

	pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);

	memset(&port_open, 0, sizeof(COMPONENT_PORT_OPEN));

	/* PCM port setting */
	port_open.type = SNDRV_PORT_TYPE;
	port_open.memory = V4L2_MEMORY_MMAP;

	port_open.format.pcm.nChannels = 2;
	port_open.format.pcm.nSamplesPerSec = 48000;
	port_open.format.pcm.nBlockAlign = 4;
	port_open.format.pcm.nBitsPerSample = 16;
	port_open.format.pcm.nAvgBytesPerSec = 48000 * 4;

	/* open pcm port */
	return pl330_port_open(chip->component, &port_open);
}

static void snd_pl330_set_hw(CComponent *component)
{
	ICodecLib *codeclib_if = CDevice_getCodecLib(component->m_pDevice);
	CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_iCodecLib, codeclib_if);
        CADV7604 *adv7604 = (CADV7604 *)codeclib->m_pVidDecoder;

	struct adv7604_state *state = &adv7604->state;
	struct v4l2_bt_timings *bt = &state->timings.bt;

	pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);

	pl330_set_config(component, "SystemControl-StreamType",
			"raw", false);
	pl330_set_config(component, "SystemControl-StreamData",
			"video+audio", false);
	
	pl330_set_config(component, "SysFunction-Function",
			"encode", false);
	pl330_set_config(component, "SysFunction-Audio",
			"passthrough", false);

#if 1
	/* just default settings for video */
	pl330_set_config(component, "SystemControl-XferMode",
			"frame", false);
	pl330_set_config(component, "SystemControl-SpsrFreq",
			"1", false);

	pl330_set_config(component, "SystemControl-FFMode",
			"frame", false);
	pl330_set_config(component, "SystemControl-VMode",
			"cavlc", false);


	if(bt->interlaced)
	{
		pl330_set_config(component, "InputControl-ScanFormat",
			"interlace", false);
		pl330_set_config(component, "InputControl-Fmt",
			"interlace", false);
		pl330_set_config(component, "DeInterlace-Mode",
			"topfield", false);
	}
	else
	{
		pl330_set_config(component, "InputControl-ScanFormat",
			"progressive", false);
		pl330_set_config(component, "InputControl-Fmt",
			"progressive", false);
		pl330_set_config(component, "DeInterlace-Mode",
			"none", false);
	}

	pl330_set_config(component, "InputControl-SrcMode",
			"hdmi", false);
	pl330_set_config(component, "InputControl-SyncMode",
			"0", false);
	pl330_set_config(component, "InputControl-DataType",
			"raw", false);

	pl330_set_config(component, "InputControl-CkEdge",
			"positive", false);
	pl330_set_config(component, "RateControl-Vbr",
			"0", false);
	pl330_set_config(component, "RateControl-Mode", "viu", false);

	pl330_set_config(component, "FilterControl-StartPixel", "1", false);
	pl330_set_config(component, "FilterControl-StartLine", "0", false);

	pl330_set_config(component, "GopLoopFilter-IntraPeriod", "30", false);
	pl330_set_config(component, "GopLoopFilter-BNum", "0", false);
	pl330_set_config(component, "GopLoopFilter-Idr", "close", false);

	pl330_set_config(component, "SysLink-VideoInput", "viu", false);
	pl330_set_config(component, "SysLink-VideoOutput", "host", false);
#endif

	/* PCM format configuration */
	pl330_set_config(component, "AudioControlParam-AudioType",
			"lpcm", false);
	pl330_set_config(component, "AudioControlParam-BitRate",
			"128", false);
	pl330_set_config(component, "AudioControlParam-ChNum",
			"2", false);
	pl330_set_config(component, "AudioControlParam-AiI2s",
			"program", false);
	pl330_set_config(component, "AudioControlParam-AiMsb",
			"msb", false);
	pl330_set_config(component, "AudioControlParam-AiM",
			"slave", false);
	pl330_set_config(component, "AudioControlEx-RawType",
			"lpcm", false);

	switch(state->rates)
	{
		case AUDIO_RATE_32000:
			pl330_set_config(component, "AudioControlParam-SampleRate",
					"32k", false);
			break;
		case AUDIO_RATE_44100:
			pl330_set_config(component, "AudioControlParam-SampleRate",
					"44.1k", false);
			break;
		case AUDIO_RATE_48000:
			pl330_set_config(component, "AudioControlParam-SampleRate",
					"48k", false);
			break;
		case AUDIO_RATE_88200:
			pl330_set_config(component, "AudioControlParam-SampleRate",
					"88.2k", false);
			break;
		case AUDIO_RATE_96000:
			pl330_set_config(component, "AudioControlParam-SampleRate",
					"96k", false);
			break;
		case AUDIO_RATE_176400:
			pl330_set_config(component, "AudioControlParam-SampleRate",
					"176.4k", false);
			break;
		case AUDIO_RATE_192000:
			pl330_set_config(component, "AudioControlParam-SampleRate",
					"192k", false);
			break;
		default:
			break;
	}
}

static int snd_pl330_start_audio(struct pl330_audio_dev *chip)
{
	CComponent *component = chip->component;
	EQP_Errors err;
        
        pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);
#if 0
	if(is_first_open)
	{
		is_first_open = false;
		return -EBUSY;
	}
#endif

	/* set PL330 audio */
	snd_pl330_set_hw(component);

	/* open pcm port */
	err = snd_pl330_port_open(chip);

	if (!QP_SUCCESS(err))
		return -EIO;

	/* start pcm port */
	err = ccomponent_start_index(component, SNDRV_PORT_TYPE);

	chip->running = true;

	return QpToLinuxKerStatus(err);
}

static int snd_pl330_stop_audio(struct pl330_audio_dev *chip)
{
	CComponent *component = chip->component;
	EQP_Errors err;
        
        pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);

	/* stop pcm port */
	err = ccomponent_stop_index(component, SNDRV_PORT_TYPE);

	/* close pcm port */
	pl330_port_close(component, SNDRV_PORT_TYPE);

	chip->running = false;

	return QpToLinuxKerStatus(err);
}

static struct snd_pcm_hardware snd_pl330_hw = {
	.info = SNDRV_PCM_INFO_MMAP |
		SNDRV_PCM_INFO_INTERLEAVED |
		SNDRV_PCM_INFO_BLOCK_TRANSFER |
		SNDRV_PCM_INFO_MMAP_VALID,
	.formats = SNDRV_PCM_FMTBIT_S16_LE,

	.rates =		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_44100 |
					SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_88200 |
					SNDRV_PCM_RATE_96000 | SNDRV_PCM_RATE_176400 |
					SNDRV_PCM_RATE_192000,
	.rate_max =		192000,
	.rate_min =		32000,
	.channels_max =	2,
	.channels_min =	2,
	.period_bytes_min = (4*2048),
	.period_bytes_max = (16*2048),
	.periods_min = 128,
	.periods_max = 512,
	.buffer_bytes_max = (512*16*2048),
};

static int snd_pl330_buffer_free(struct pl330_audio_dev *chip)
{
	pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);

	kfree(chip->buf);
	chip->buf = NULL;

	return 0;
}

static int snd_pl330_component_init(struct pl330_audio_dev *chip)
{
	CComponent *component;
//	int err = 0;

        pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);
	component = snd_pl330_component_open(chip->dev);

	if(!component)
		return -ENOMEM;
	chip->component = component;

	return 0;

}

static int snd_pl330_pcm_open(struct snd_pcm_substream *substream)
{
	struct pl330_audio_dev *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;

	pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);

	if(snd_pl330_component_init(chip) < 0)
		return -EIO;

	chip->substream = substream;
	runtime->hw = snd_pl330_hw;
	snd_pcm_hw_constraint_integer(runtime, SNDRV_PCM_HW_PARAM_PERIODS);

	return 0;
}

static int snd_pl330_pcm_close(struct snd_pcm_substream *substream)
{
	struct pl330_audio_dev *chip = snd_pcm_substream_chip(substream);

//	CDevice *dev = chip->dev;
	CComponent *component = chip->component;

	pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);

	snd_pl330_component_close(chip->dev, component);

	return 0;
}

static int snd_pl330_hw_params(struct snd_pcm_substream *substream,
			struct snd_pcm_hw_params *hw_params)
{
	struct pl330_audio_dev *chip = snd_pcm_substream_chip(substream);
	struct v4l2_buffer *buf;

	pl330_trace(PL330_TRACE_SND, "%s() alloc buffer size = %u"
			" periods = %u period_bytes = %u\n",
			__func__,
			params_buffer_bytes(hw_params),
			params_periods(hw_params),
			params_period_bytes(hw_params));

	if(substream->runtime->dma_area)
	{
		snd_pl330_buffer_free(chip);
		substream->runtime->dma_area = NULL;
	}

	buf = kzalloc(sizeof(struct v4l2_buffer), GFP_KERNEL);
	if(!buf)
		return -ENOMEM;

	chip->buf = buf;

	return snd_pcm_lib_alloc_vmalloc_buffer(substream,
				params_buffer_bytes(hw_params));
}

static int snd_pl330_hw_free(struct snd_pcm_substream *substream)
{
	struct pl330_audio_dev *chip = snd_pcm_substream_chip(substream);
        
        pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);

	if(substream->runtime->dma_area)
	{
		snd_pl330_buffer_free(chip);
		substream->runtime->dma_area = NULL;
	}

	return snd_pcm_lib_free_vmalloc_buffer(substream);
}

static int snd_pl330_prepare(struct snd_pcm_substream *substream)
{
	struct pl330_audio_dev *chip = snd_pcm_substream_chip(substream);

	pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);

	chip->capture_transfer_done = 0;
	chip->hwptr = 0;

	pl330_acquire_index(chip->component, SNDRV_PORT_TYPE);

	return 0;
}


static void pl330_audio_trigger(struct work_struct *work)
{
	struct pl330_audio_dev *chip = container_of(work, struct pl330_audio_dev, wq_trigger);

	if(atomic_read(&chip->stream_started))
	{
		pl330_trace(PL330_TRACE_SND, "%s() start\n", __func__);
		snd_pl330_start_audio(chip);
	}
	else {
		pl330_trace(PL330_TRACE_SND, "%s() stop\n", __func__);
		snd_pl330_stop_audio(chip);
	}
}

static int snd_pl330_card_trigger(struct snd_pcm_substream *substream,
			int cmd)
{
	struct pl330_audio_dev *chip = snd_pcm_substream_chip(substream);
	int err = 0;

	switch(cmd)
	{
	case SNDRV_PCM_TRIGGER_START:
		atomic_set(&chip->stream_started, 1);
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		atomic_set(&chip->stream_started, 0);
		break;
	default:
		err = -EINVAL;
	}

	schedule_work(&chip->wq_trigger);

	return err;
}

static snd_pcm_uframes_t snd_pl330_pointer(
		struct snd_pcm_substream *substream)
{
	struct pl330_audio_dev *chip = snd_pcm_substream_chip(substream);
//	struct snd_pcm_runtime *runtime = substream->runtime;

	pl330_trace(PL330_TRACE_SND,
			"%s() hwptr=%u\n", __func__,
			chip->hwptr);

	return chip->hwptr;
}

static __attribute__((__unused__)) struct page *snd_pl330_page(struct snd_pcm_substream *substream,
			unsigned long offset)
{
	void *pageptr = substream->runtime->dma_area + offset;
	return vmalloc_to_page(pageptr);
}

static struct snd_pcm_ops snd_pl330_pcm_ops = {
	.open = snd_pl330_pcm_open,
	.close = snd_pl330_pcm_close,
	.ioctl = snd_pcm_lib_ioctl,
	.hw_params = snd_pl330_hw_params,
	.hw_free = snd_pl330_hw_free,
	.prepare = snd_pl330_prepare,
	.trigger = snd_pl330_card_trigger,
	.pointer = snd_pl330_pointer,
	.page = snd_pcm_lib_get_vmalloc_page,
//	.page = snd_pl330_page,
};

static int snd_pl330_pcm(struct pl330_audio_dev *chip, int device,
		char *name)
{
	int err;
	struct snd_pcm *pcm;

	pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);

	err = snd_pcm_new(chip->card, name, device, 0, 1, &pcm);
	if(err < 0)
		return err;
	pcm->private_data = chip;
	strcpy(pcm->name, name);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &snd_pl330_pcm_ops);

	/* on buffer done callback */
	chip->on_buffer_done = &snd_pl330_on_buffer_done;

	return 0;
}

struct pl330_audio_dev *pl330_audio_register(CDevice *dev)
{
	struct snd_card *card;
	struct pl330_audio_dev *chip;

	int err;

	pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);
#if LINUX_VERSION_CODE  < KERNEL_VERSION(3,15,0)
	err = snd_card_create(SNDRV_DEFAULT_IDX1, SNDRV_DRIVER_STR,
			THIS_MODULE, sizeof(struct pl330_audio_dev), &card);
#else
	err = snd_card_new(dev->m_pGenericDevice,SNDRV_DEFAULT_IDX1, SNDRV_DRIVER_STR,
			THIS_MODULE, sizeof(struct pl330_audio_dev), &card);
#endif	
	if(err < 0)
		goto error;

	chip = (struct pl330_audio_dev *) card->private_data;
	chip->dev = dev;
	chip->pci = dev->m_device.p_pciedev;
	chip->card = card;

	spin_lock_init(&chip->lock);

	snd_card_set_dev(card, &chip->pci->dev);

	err = snd_pl330_pcm(chip, 0, "C353 LPCM");
	if(err < 0)
		goto error;

	strcpy(card->driver, "C353-Audio");
	strcpy(card->shortname, "C353 Audio");
	strcpy(card->longname, "AVerMedia C353 LPCM Audio");

	INIT_WORK(&chip->wq_trigger, pl330_audio_trigger);

	err = snd_card_register(card);
	if(err < 0)
		goto error;

	pl330_trace(PL330_TRACE_SND, "%s() registered ALSA audio device\n",
			__func__);

	return chip;
error:
	snd_card_free(card);
	pl330_trace(PL330_TRACE_SND, "%s() Failed to register "
			"audio adapter\n", __func__);

	return NULL;
}

void pl330_audio_unregister(CDevice *dev)
{
	struct pl330_audio_dev *chip = dev->m_pAudioDev;

	pl330_trace(PL330_TRACE_SND, "%s()\n", __func__);

	snd_card_free(chip->card);
}
