///////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2011 Sigma Designs, Inc.  All Rights Reserved.   
//
//		OS Manager - Linux Kernel Mode specific code
//
// History:
//	April 10, 2008
//
// Authors: Catherine Cheng, Enrico Cadorin
//
///////////////////////////////////////////////////////////////////////////////



#include "qptypes.h"
#include "qposmgr.h"
#include "qpdbgmgr.h"
#include "qpmemmgr.h"
#include "qposmgrint.h"
#include "qptimemgr.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
#include <linux/sched/signal.h>
#else
#include <linux/sched.h>
#endif
#include <linux/kthread.h>

// The IRQL level used to protect the event group
#define EVTG_IRQL	HIGH_LEVEL

/*
static BOOLEAN lock (PRKSEMAPHORE mutexP);
static void unlock (PRKSEMAPHORE mutexP);*/


EQP_Errors
OSMInit (void)
{
	EQP_Errors	rv = QPERR_SUCCESS;

	return (rv);
}


void
OSMTerm (void)
{
}


void
OSMGetCaps (PTQP_OSMCaps capsP)
{
	capsP->osType = QPOS_TYPE_LINUXKERNEL;
	capsP->osVersion = 0;
}


EQP_Errors
OSMCreateThread (
		PTThreadBlock twsP, PTQP_OSMThreadParms parmsP)
{
	EQP_Errors	rv = QPERR_SUCCESS;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0))
	twsP->threadT = kernel_thread((void *)parmsP->procP, parmsP->parm, CLONE_FS|CLONE_FILES);
#else
	parmsP->task = kthread_run((void *)parmsP->procP, parmsP->parm,
			parmsP->name);
#endif

	return (rv);
}


EQP_Errors
OSMDeleteThread (PTThreadBlock twsP)
{
	EQP_Errors		rv = QPERR_SUCCESS;

#if defined(__LINUX24__) || (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,24))
        kill_proc(twsP->threadT, SIGKILL, 1);
#else
        kill_pid(find_get_pid(twsP->threadT), SIGKILL, 1);
#endif
	return (rv);
}


EQP_Errors
OSMCreateSem (PTSemaphoreBlock semBlkP,
 			unsigned long	maxCount,
			unsigned long	initCount)
{
	EQP_Errors	rv = QPERR_SUCCESS;

	sema_init( &semBlkP->sem, initCount );

	return (rv);
}

EQP_Errors
OSMDeleteSem (PTSemaphoreBlock semBlkP)
{
	EQP_Errors	rv = QPERR_SUCCESS;

	return (rv);
}

EQP_Errors
OSMSignalSem (PTSemaphoreBlock semBlkP)
{
	up( &semBlkP->sem );

	return QPERR_SUCCESS;
}

EQP_Errors
OSMWaitSem (PTSemaphoreBlock semBlkP, long timeoutMS)
{
	down( &semBlkP->sem ); 
	
	return QPERR_SUCCESS;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,15,0)
static void _Evtgrp_timeout(struct timer_list *t)
{
	PTEventBlock evtBlkP = from_timer(evtBlkP, t, timer);//(PTEventBlock)evt;

	up(&evtBlkP->sem);
}
#else
static void _Evtgrp_timeout(unsigned long evt)
{
	PTEventBlock evtBlkP = (PTEventBlock)evt;

	up(&evtBlkP->sem);
}
#endif


EQP_Errors
OSMCreateEvtgrp (PTEventBlock evtBlkP, TQP_UINT32 initState)
{
	unsigned long flags;

	spin_lock_init(&evtBlkP->lock);

	spin_lock_irqsave(&evtBlkP->lock, flags);

	evtBlkP->bits = initState;

	spin_unlock_irqrestore(&evtBlkP->lock, flags);

	sema_init( &evtBlkP->sem, 0 );
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,15,0)
        timer_setup(&(evtBlkP->timer), _Evtgrp_timeout, 0);
#else
	init_timer(&evtBlkP->timer);
#endif

	return QPERR_SUCCESS;
}

EQP_Errors
OSMDeleteEvtgrp (PTEventBlock evtBlkP)
{
	del_timer_sync(&evtBlkP->timer);
	return QPERR_SUCCESS;
}

EQP_Errors
OSMWaitEvtgrp (PTEventBlock evtBlkP,
				TQP_UINT32 mask,
				TQP_UINT32 * evtStatusP,
				long timeoutMS)
{
	EQP_Errors	rv = QPERR_TIMEOUT;
	unsigned long	ulTimeoutJiffies = 0;

        unsigned long    flags;

	//printk("%s()-> evtBlkP=0x%08x, mask=0x%08x\n", __FUNCTION__, evtBlkP, mask);
	
	if (timeoutMS > 0) {
		ulTimeoutJiffies = jiffies + ((timeoutMS*HZ+999)/1000);
		evtBlkP->timer.expires = ulTimeoutJiffies;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,15,0)
		evtBlkP->timer.function = _Evtgrp_timeout;
		evtBlkP->timer.data = (unsigned long)evtBlkP;
#endif
		add_timer(&evtBlkP->timer);
	}
	*evtStatusP = 0;

RECHECK:
	if (0 != timeoutMS) {
		down(&evtBlkP->sem);
	}

	spin_lock_irqsave(&evtBlkP->lock, flags);

		*evtStatusP = (mask & evtBlkP->bits);

		if (*evtStatusP) {
			evtBlkP->bits &= ~mask;
			rv = QPERR_SUCCESS;
		}

		spin_unlock_irqrestore(&evtBlkP->lock, flags);

	if (timeoutMS > 0) {
		if ((rv != QPERR_SUCCESS) && (jiffies < ulTimeoutJiffies)) {
			/* The case here is because semaphore could be stay in high 
                         * status in the end of previous task streaming.
                         * And that could make the first down(sem) operation in new 
                         * task streaming actually does not wait given timeout parameter. 
                         */
			goto RECHECK;
		} else {	
			del_timer_sync(&evtBlkP->timer);
		}
	}


	if ((QPERR_SUCCESS != rv) && (timeoutMS < 0)) {
		rv = QPERR_SUCCESS;
	}
	
	return (rv);
}


EQP_Errors
OSMSetEvtgrp (PTEventBlock evtBlkP, TQP_UINT32 mask)
{
	unsigned long flags; 

	spin_lock_irqsave(&evtBlkP->lock, flags);

	evtBlkP->bits |= mask;

	spin_unlock_irqrestore(&evtBlkP->lock, flags);

	up(&evtBlkP->sem);

	return QPERR_SUCCESS;
}


EQP_Errors
OSMClearEvtgrp (PTEventBlock evtBlkP, TQP_UINT32 mask)
{
	unsigned long flags; 

	spin_lock_irqsave(&evtBlkP->lock, flags);

	evtBlkP->bits &= ~mask;

	spin_unlock_irqrestore(&evtBlkP->lock, flags);

	return QPERR_SUCCESS;
}


EQP_Errors
OSMRegisterISR (TQP_UINT32 isrNumber)
{
	EQP_Errors	rv = QPERR_SUCCESS;

	// only the codec ISR number is allwoed
	if (isrNumber != QPOS_ISR_CODEC) {
	//if (isrNumber > QPOS_ISR_CODEC_MAX) {
		rv = QPERR_FAIL;
		QPDbgMPrintf((QPDBGM_LEVEL_ERROR,"OSMRegisterISR(WINKER): invalid ISR number %i\n",isrNumber));
	}

	return (rv);
}

/*
static BOOLEAN
lock (PRKSEMAPHORE mutexP)
{
	return (KeWaitForSingleObject(mutexP,Executive,KernelMode,FALSE,NULL) == STATUS_SUCCESS);
}

static void
unlock (PRKSEMAPHORE mutexP)
{
	KeReleaseSemaphore(mutexP,0,1,FALSE);
}
*/


