///////////////////////////////////////////////////////////////////////////////
//
// cadv7604.c
//
// Description: 
//   ADV7604 implementation
// 
// History:
//	 Aug 15, 2013 created
//
// Authors: Shihpo Wang
//
///////////////////////////////////////////////////////////////////////////////

#include "cadv7604.h"
#include "qptimemgr.h"
#include "qpdbgmgr.h"
#include "util.h"
#include "cgpio_i2c.h"
#include "plgpio.h"
#include "boardinfo-probe.h"

#include "cqlcodeclib.h"
#include "cqlcodec.h"
#include "c353.h"
#include "c353-ctrl.h"
#include "pl330.h"
#include "pl330-dbg.h"
#include "tlv320aic3101.h"
#include "cadv7604_edid.h"
#include "pl330_hwctl.h"

#include <linux/device.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/kthread.h>
#include <linux/time.h>
#include <linux/list.h>
#include <linux/slab.h>

#define debug adv7604_trace_level
extern unsigned int adv7604_trace_level;

static void adv7604_configure_pl330_audio(CADV7604 *This,
		struct adv7604_state *state);
static void adv7604_audio_sample_rate(CADV7604 *This);
static int adv7604_configure_pl330(CADV7604 *This,
		struct v4l2_dv_timings *timings);
static int cadv7604_v4l2_s_input(void *frontend, int input);
#define adv7604_printk(level, This, fmt, arg...)  \
	printk(level "Device: %02x " fmt,  This->m_cI2CDevAddr, ## arg)

#define adv7604_dbg(level, debug, This, fmt, arg...)						\
	do {																\
		if(adv7604_trace_level >= (level))											\
			adv7604_printk(KERN_DEBUG, This, fmt, ## arg);					\
	} while(0)

#define adv7604_info(This, fmt, arg...)  \
	do {							\
		if(adv7604_trace_level > 0) \
			printk(KERN_INFO "Device: %02x " fmt,  This->m_cI2CDevAddr, ## arg); \
	} while(0)

#define ADV7604_fsc (28636360)

/////////////////////////////////////////////////////////////////////////////

#define DIGITAL_INPUT (state->mode == ADV7604_MODE_HDMI)

#define ADV7604_POLLING_TIMEOUT		(1 * HZ)

#define ADV7604_FMT_QUEUE_SIZE 10

#define AIC3101_GPIO_RESET_QUERY 1
//#define AIC3101_DEVICE_ADDR_QUERRY (0x30 >> 1) // add for probe TI3101


#define V4L2_DV_BT_CEA_1920X1080I30 { \
	.type = V4L2_DV_BT_656_1120, \
			V4L2_INIT_BT_TIMINGS(1920, 1080, 1, \
					V4L2_DV_HSYNC_POS_POL | V4L2_DV_VSYNC_POS_POL, \
					74250000, 88, 44, 148, 2, 5, 15, 2, 5, 16, \
					V4L2_DV_BT_STD_CEA861, V4L2_DV_FL_HALF_LINE) \
}


/* Supported CEA and DMT timings */
static const struct v4l2_dv_timings adv7604_timings[] = {
	V4L2_DV_BT_CEA_720X480P59_94,
	V4L2_DV_BT_CEA_720X576P50,
	V4L2_DV_BT_CEA_1280X720P24,
	V4L2_DV_BT_CEA_1280X720P25,
	V4L2_DV_BT_CEA_1280X720P50,
	V4L2_DV_BT_CEA_1280X720P60,
	V4L2_DV_BT_CEA_1920X1080P24,
	V4L2_DV_BT_CEA_1920X1080P25,

	V4L2_DV_BT_CEA_1920X1080I30,

	V4L2_DV_BT_CEA_1920X1080P30,
	V4L2_DV_BT_CEA_1920X1080P50,
	V4L2_DV_BT_CEA_1920X1080I60,
	V4L2_DV_BT_CEA_1920X1080P60,

	/* sorted by DMT ID */
	V4L2_DV_BT_DMT_640X350P85,
	V4L2_DV_BT_DMT_640X400P85,
	V4L2_DV_BT_DMT_720X400P85,
	V4L2_DV_BT_DMT_640X480P60,
	V4L2_DV_BT_DMT_640X480P72,
	V4L2_DV_BT_DMT_640X480P75,
	V4L2_DV_BT_DMT_640X480P85,
	V4L2_DV_BT_DMT_800X600P56,
	V4L2_DV_BT_DMT_800X600P60,
	V4L2_DV_BT_DMT_800X600P72,
	V4L2_DV_BT_DMT_800X600P75,
	V4L2_DV_BT_DMT_800X600P85,
	V4L2_DV_BT_DMT_848X480P60,
	V4L2_DV_BT_DMT_1024X768P60,
	V4L2_DV_BT_DMT_1024X768P70,
	V4L2_DV_BT_DMT_1024X768P75,
	V4L2_DV_BT_DMT_1024X768P85,
	V4L2_DV_BT_DMT_1152X864P75,
	V4L2_DV_BT_DMT_1280X768P60_RB,
	V4L2_DV_BT_DMT_1280X768P60,
	V4L2_DV_BT_DMT_1280X768P75,
	V4L2_DV_BT_DMT_1280X768P85,
	V4L2_DV_BT_DMT_1280X800P60_RB,
	V4L2_DV_BT_DMT_1280X800P60,
	V4L2_DV_BT_DMT_1280X800P75,
	V4L2_DV_BT_DMT_1280X800P85,
	V4L2_DV_BT_DMT_1280X960P60,
	V4L2_DV_BT_DMT_1280X960P85,
	V4L2_DV_BT_DMT_1280X1024P60,
	V4L2_DV_BT_DMT_1280X1024P75,
	V4L2_DV_BT_DMT_1280X1024P85,
	V4L2_DV_BT_DMT_1360X768P60,
	V4L2_DV_BT_DMT_1400X1050P60_RB,
	V4L2_DV_BT_DMT_1400X1050P60,
	V4L2_DV_BT_DMT_1400X1050P75,
	V4L2_DV_BT_DMT_1400X1050P85,
	V4L2_DV_BT_DMT_1440X900P60_RB,
	V4L2_DV_BT_DMT_1440X900P60,
	V4L2_DV_BT_DMT_1600X1200P60,
	V4L2_DV_BT_DMT_1680X1050P60_RB,
	V4L2_DV_BT_DMT_1680X1050P60,
	V4L2_DV_BT_DMT_1792X1344P60,
	V4L2_DV_BT_DMT_1856X1392P60,
	V4L2_DV_BT_DMT_1920X1200P60_RB,
	V4L2_DV_BT_DMT_1366X768P60,
	V4L2_DV_BT_DMT_1920X1080P60,
	{ },
};

struct adv7604_video_standards {
	struct v4l2_dv_timings timings;
	u8 vid_std;
	u8 v_freq;
};

/* sorted by number of lines */
static const struct adv7604_video_standards adv7604_prim_mode_comp[] = {
	/* { V4L2_DV_BT_CEA_720X480P59_94, 0x0a, 0x00 }, TODO flickering */
	{ V4L2_DV_BT_CEA_720X576P50, 0x0b, 0x00 },
	{ V4L2_DV_BT_CEA_1280X720P50, 0x19, 0x01 },
	{ V4L2_DV_BT_CEA_1280X720P60, 0x19, 0x00 },
	{ V4L2_DV_BT_CEA_1920X1080P24, 0x1e, 0x04 },
	{ V4L2_DV_BT_CEA_1920X1080P25, 0x1e, 0x03 },
	{ V4L2_DV_BT_CEA_1920X1080P30, 0x1e, 0x02 },
	{ V4L2_DV_BT_CEA_1920X1080P50, 0x1e, 0x01 },
	{ V4L2_DV_BT_CEA_1920X1080P60, 0x1e, 0x00 },
	/* TODO add 1920x1080P60_RB (CVT timing) */
	{ },
};

/* sorted by number of lines */
static const struct adv7604_video_standards adv7604_prim_mode_gr[] = {
	{ V4L2_DV_BT_DMT_640X480P60, 0x08, 0x00 },
	{ V4L2_DV_BT_DMT_640X480P72, 0x09, 0x00 },
	{ V4L2_DV_BT_DMT_640X480P75, 0x0a, 0x00 },
	{ V4L2_DV_BT_DMT_640X480P85, 0x0b, 0x00 },
	{ V4L2_DV_BT_DMT_800X600P56, 0x00, 0x00 },
	{ V4L2_DV_BT_DMT_800X600P60, 0x01, 0x00 },
	{ V4L2_DV_BT_DMT_800X600P72, 0x02, 0x00 },
	{ V4L2_DV_BT_DMT_800X600P75, 0x03, 0x00 },
	{ V4L2_DV_BT_DMT_800X600P85, 0x04, 0x00 },
	{ V4L2_DV_BT_DMT_1024X768P60, 0x0c, 0x00 },
	{ V4L2_DV_BT_DMT_1024X768P70, 0x0d, 0x00 },
	{ V4L2_DV_BT_DMT_1024X768P75, 0x0e, 0x00 },
	{ V4L2_DV_BT_DMT_1024X768P85, 0x0f, 0x00 },
	//{ V4L2_DV_BT_DMT_1280X800P60, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1280X800P60_RB, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1280X1024P60, 0x05, 0x00 },
	{ V4L2_DV_BT_DMT_1280X1024P75, 0x06, 0x00 },
	{ V4L2_DV_BT_DMT_1360X768P60, 0x12, 0x00 },
	{ V4L2_DV_BT_DMT_1366X768P60, 0x13, 0x00 },
	{ V4L2_DV_BT_DMT_1400X1050P60, 0x14, 0x00 },
	{ V4L2_DV_BT_DMT_1400X1050P75, 0x15, 0x00 },
	//{ V4L2_DV_BT_DMT_1440X900P60, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1440X900P60_RB, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1600X1200P60, 0x16, 0x00 }, /* TODO not tested */
	/* TODO add 1600X1200P60_RB (not a DMT timing) */
	{ V4L2_DV_BT_DMT_1680X1050P60, 0x18, 0x00 },
	{ V4L2_DV_BT_DMT_1920X1200P60_RB, 0x19, 0x00 }, /* TODO not tested */
	{ },
};

/* sorted by number of lines */
static const struct adv7604_video_standards adv7604_prim_mode_hdmi_comp[] = {
	{ V4L2_DV_BT_CEA_720X480P59_94, 0x0a, 0x00 },
	{ V4L2_DV_BT_CEA_720X576P50, 0x0b, 0x00 },
	{ V4L2_DV_BT_CEA_1280X720P30 ,0x13, 0x02 },
	{ V4L2_DV_BT_CEA_1280X720P50, 0x13, 0x01 },
	{ V4L2_DV_BT_CEA_1280X720P60, 0x13, 0x00 },
	{ V4L2_DV_BT_CEA_1920X1080P24, 0x1e, 0x04 },
	{ V4L2_DV_BT_CEA_1920X1080P25, 0x1e, 0x03 },
	{ V4L2_DV_BT_CEA_1920X1080I30, 0x14, 0x02 },
	{ V4L2_DV_BT_CEA_1920X1080P30, 0x1e, 0x02 },
	{ V4L2_DV_BT_CEA_1920X1080I50, 0x1e, 0x01 },
	{ V4L2_DV_BT_CEA_1920X1080P50, 0x1e, 0x01 },
	{ V4L2_DV_BT_CEA_1920X1080I60, 0x1e, 0x00 },
	{ V4L2_DV_BT_CEA_1920X1080P60, 0x1e, 0x00 },
	{ },
};

/* sorted by number of lines */
static const struct adv7604_video_standards adv7604_prim_mode_hdmi_gr[] = {
	{ V4L2_DV_BT_DMT_640X480P60, 0x08, 0x00 },
	{ V4L2_DV_BT_DMT_640X480P72, 0x09, 0x00 },
	{ V4L2_DV_BT_DMT_640X480P75, 0x0a, 0x00 },
	{ V4L2_DV_BT_DMT_640X480P85, 0x0b, 0x00 },
	{ V4L2_DV_BT_DMT_800X600P56, 0x00, 0x00 },
	{ V4L2_DV_BT_DMT_800X600P60, 0x01, 0x00 },
	{ V4L2_DV_BT_DMT_800X600P72, 0x02, 0x00 },
	{ V4L2_DV_BT_DMT_800X600P75, 0x03, 0x00 },
	{ V4L2_DV_BT_DMT_800X600P85, 0x04, 0x00 },
	{ V4L2_DV_BT_DMT_1024X768P60, 0x0c, 0x00 },
	{ V4L2_DV_BT_DMT_1024X768P70, 0x0d, 0x00 },
	{ V4L2_DV_BT_DMT_1024X768P75, 0x0e, 0x00 },
	{ V4L2_DV_BT_DMT_1024X768P85, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1280X800P60, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1280X800P60_RB, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1280X800P75, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1280X1024P60, 0x05, 0x00 },
	{ V4L2_DV_BT_DMT_1280X1024P75, 0x06, 0x00 },
	{ V4L2_DV_BT_DMT_1360X768P60, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1440X900P60, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1440X900P60_RB, 0x0f, 0x00 },
	{ V4L2_DV_BT_DMT_1680X1050P60, 0x18, 0x00 },
	{ },
};

/* ----------------------------------------------------------------------- */

// I2C IO Map
int io_write(void *this, TQP_UINT8 reg, TQP_UINT8 value)
{
    CADV7604 *This=this;
    TQP_UINT8 buffer[2];
    buffer[0] = reg;
    buffer[1] = value;

    return (II2C_Write(This->m_pI2C, ADV7604_DEVICE_ADDR_IO, 2, buffer));
}

int io_read(void *this, TQP_UINT8 reg)
{
    CADV7604 *This=this;
    TQP_UINT8 value = 0;
    EQP_Errors err;
    err = II2C_WriteThenRead(This->m_pI2C, ADV7604_DEVICE_ADDR_IO, 1, &reg, 1, &value);
    //if(err != QPERR_SUCCESS)
        //adv7604_printk(KERN_ERR, This, "%s error %d\n", __func__, err);
    return value;
}

int io_write_and_or(void *This, u8 reg, u8 mask, u8 val)
{
	return io_write(This, reg, (io_read(This, reg) & mask) | val);
}

// I2C CP Map
int cp_write(void *this, u8 reg, u8 value)
{
    CADV7604 *This=this;
    TQP_UINT8 buffer[2];

    buffer[0] = reg;
    buffer[1] = value;

    return (II2C_Write(This->m_pI2C, ADV7604_DEVICE_ADDR_CP, 2, buffer));
}

int cp_read(void *this, u8 reg)
{
    CADV7604 *This=this;
    TQP_UINT8 value = 0;
    EQP_Errors err;
    err = II2C_WriteThenRead(This->m_pI2C, ADV7604_DEVICE_ADDR_CP, 1, &reg, 1, &value);
    //if(err != QPERR_SUCCESS)
        //adv7604_printk(KERN_ERR, This, "%s error %d\n", __func__, err);
    return value;
}

int cp_write_and_or(void *This, u8 reg, u8 mask, u8 val)
{
    return cp_write(This, reg, (cp_read(This, reg) & mask) | val);
}

// I2C AFE_DPLL Map
int afe_write(void *this, u8 reg, u8 value)
{
    CADV7604 *This=this;
    TQP_UINT8 buffer[2];

    buffer[0] = reg;
    buffer[1] = value;

    return (II2C_Write(This->m_pI2C, ADV7604_DEVICE_ADDR_AFE_DPLL, 2, buffer));
}

int afe_read(void *this, u8 reg)
{
    CADV7604 *This=this;
    TQP_UINT8 value = 0;
    EQP_Errors err;
    err = II2C_WriteThenRead(This->m_pI2C, ADV7604_DEVICE_ADDR_AFE_DPLL, 1, &reg, 1, &value);
    //if(err != QPERR_SUCCESS)
        //adv7604_printk(KERN_ERR, This, "%s error %d\n", __func__, err);
    return value;
}

int afe_write_and_or(void *This, u8 reg, u8 mask, u8 val )
{
    return afe_write(This, reg, ( afe_read(This, reg ) & mask ) | val );
}

// I2C REPEATER Map
int rep_write(void *this, u8 reg, u8 value)
{
    CADV7604 *This=this;
    TQP_UINT8 buffer[2];

    buffer[0] = reg;
    buffer[1] = value;

    return (II2C_Write(This->m_pI2C, ADV7604_DEVICE_ADDR_REPEATER, 2, buffer));
}

int rep_read(void *this, u8 reg)
{
    CADV7604 *This=this;
    TQP_UINT8 value = 0;
    EQP_Errors err;
    err = II2C_WriteThenRead(This->m_pI2C, ADV7604_DEVICE_ADDR_REPEATER, 1, &reg, 1, &value);
    //if(err != QPERR_SUCCESS)
        //adv7604_printk(KERN_ERR, This, "%s error %d\n", __func__, err);
    return value;
}

int rep_write_and_or(CADV7604 *This, u8 reg, u8 mask, u8 val)
{
    return rep_write(This, reg, (rep_read(This, reg) & mask) | val);
}

// I2C HDMI Map
int hdmi_write(void *this, u8 reg, u8 value)
{
    CADV7604 *This=this;
    TQP_UINT8 buffer[2];

    buffer[0] = reg;
    buffer[1] = value;

    return (II2C_Write(This->m_pI2C, ADV7604_DEVICE_ADDR_HDMI, 2, buffer));
}

int hdmi_read(void *this, u8 reg)
{
    CADV7604 *This=this;
    TQP_UINT8 value = 0;
    EQP_Errors err;
    err = II2C_WriteThenRead(This->m_pI2C, ADV7604_DEVICE_ADDR_HDMI, 1,	&reg, 1, &value);
    //if(err != QPERR_SUCCESS)
        //adv7604_printk(KERN_ERR, This, "%s error %d\n", __func__, err);
    return value;
}

static inline unsigned hblanking(const struct v4l2_bt_timings *t)
{
	return t->hfrontporch + t->hsync + t->hbackporch;
}

static inline unsigned htotal(const struct v4l2_bt_timings *t)
{
	return t->width + t->hfrontporch + t->hsync + t->hbackporch;
}

static inline unsigned vblanking(const struct v4l2_bt_timings *t)
{
	return t->vfrontporch + t->vsync + t->vbackporch;
}

static inline unsigned vtotal(const struct v4l2_bt_timings *t)
{
	return t->height + t->vfrontporch + t->vsync + t->vbackporch;
}

/* kernel 3.7.10 drivers/media/v4l2-core/v4l2-common.c */
/**
 * v4l_match_dv_timings - check if two timings match
 * @t1 - compare this v4l2_dv_timings struct...
 * @t2 - with this struct.
 * @pclock_delta - the allowed pixelclock deviation.
 *
 * Compare t1 with t2 with a given margin of error for the pixelclock.
 */
bool v4l_match_dv_timings(const struct v4l2_dv_timings *t1,
							const struct v4l2_dv_timings *t2,
							unsigned pclock_delta)
{
	if (t1->type != t2->type || t1->type != V4L2_DV_BT_656_1120)
                return false;
	if (t1->bt.width == t2->bt.width &&
		t1->bt.height == t2->bt.height &&
		t1->bt.interlaced == t2->bt.interlaced &&
		t1->bt.polarities == t2->bt.polarities &&
		t1->bt.pixelclock >= t2->bt.pixelclock - pclock_delta &&
		t1->bt.pixelclock <= t2->bt.pixelclock + pclock_delta &&
		t1->bt.hfrontporch == t2->bt.hfrontporch &&
		t1->bt.vfrontporch == t2->bt.vfrontporch &&
		t1->bt.vsync == t2->bt.vsync &&
		t1->bt.vbackporch == t2->bt.vbackporch &&
		(!t1->bt.interlaced ||
		(t1->bt.il_vfrontporch == t2->bt.il_vfrontporch &&
			t1->bt.il_vsync == t2->bt.il_vsync &&
			t1->bt.il_vbackporch == t2->bt.il_vbackporch)))
		return true;
	return false;
}


static __inline void udelay_loop(TQPTM_Time uS)
{
    while(uS--)
    {
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
        asm volatile("nop;"); asm volatile("nop;");	asm volatile("nop;"); asm volatile("nop;");
    }
};

static void adv7604_audio_sample_rate(CADV7604 *This)
{
    struct adv7604_state *state = &This->state;
    long cts = 0, n = 0;
    int tmds;
    int freq;
    int i;
    long diff, min;
    int table[] = {
                AUDIO_32K,
                AUDIO_44K,
                AUDIO_48K,
                AUDIO_88K,
                AUDIO_96K,
                AUDIO_176K,
                AUDIO_192K
    };
	
    cts = hdmi_read(This, 0x5b) << 12;
    cts |= hdmi_read(This, 0x5c) << 4;
    cts |= ((hdmi_read(This, 0x5d) >> 4) & 0x0f);

    if(cts == 0)
    {
        return;
    }

    n = (hdmi_read(This, 0x5d) & 0x0f) << 16;
    n |= hdmi_read(This, 0x5e) << 8;
    n |= hdmi_read(This, 0x5f);
    tmds = hdmi_read(This, 0x06);
    //adv7604_info(This, "cts = %x, n = %x, tmds = %x\n",cts, n, tmds);

    if(cts == 0)
        return;

    freq = ((((long)tmds * n * 10) / cts) * 1000) >> 7;
    //adv7604_info(This, "freq = %x\n", freq);

    for(i = 0, min = 0xff; i < ARRAY_SIZE(table); i++)
    {
        diff = abs(freq - table[i]);
        //adv7604_info(This, "diff = %d\n", diff);
        if(diff <= min)
        {
            min = diff;
            state->rates = i;
        }
    }
    //adv7604_info(This, "sample_rate = %x\n", state->rates);
}


/* get input status. Same as the status field in the v4l2_input struct */
static int adv7604_g_input_status(void *this, int n, u32 *status)
{
    CADV7604 *This=this;
    //struct adv7604_state *state = &This->state;

    *status = 0;

    if(n == C353_HDMI)
        /* cable power */
        *status |= (io_read(This, 0x6f) & 0x12) ? 0 : V4L2_IN_ST_NO_POWER;
    else
        *status |= (io_read(This, 0x0c) & 0x24) ? V4L2_IN_ST_NO_POWER : 0;

    *status |= no_signal(This, n == C353_HDMI ? 1 : 2) ? V4L2_IN_ST_NO_SIGNAL : 0;

    if(n == C353_VGA) 
    {
        *status |= no_lock_stdi(This, 2) ? V4L2_IN_ST_NO_H_LOCK : 0;
    }
    else if(n == C353_HDMI)
        *status |= no_lock_stdi(This, 1) ? V4L2_IN_ST_NO_SYNC: 0;

    //if (no_lock_cp(This))
        //adv7604_dbg(1, debug, This, "%s: status = 0x%x\n", __func__, *status);

    return 0;
}

static void set_rgb_quantization_range(CADV7604 *This)
{
	struct adv7604_state *state = &This->state;

	adv7604_info(This, "rgb range:%d\n", state->rgb_quantization_range);
	switch (state->rgb_quantization_range) {
	case V4L2_DV_RGB_RANGE_AUTO:
		/* automatic */
		if (DIGITAL_INPUT && !(hdmi_read(This, 0x05) & 0x80)) {
			/* receiving DVI-D signal */

			/* ADV7604 selects RGB limited range regardless of
			   input format (CE/IT) in automatic mode */
			if (state->standards & V4L2_DV_BT_STD_CEA861) {
				/* RGB limited range (16-235) */
				io_write_and_or(This, 0x02, 0x0f, 0x00);

			} else {
				/* RGB full range (0-255) */
				io_write_and_or(This, 0x02, 0x0f, 0x10);
			}
		} else {
			/* receiving HDMI or analog signal, set automode */
			io_write_and_or(This, 0x02, 0x0f, 0xf0);
		}
		break;
	case V4L2_DV_RGB_RANGE_LIMITED:
		/* RGB limited range (16-235) */
		io_write_and_or(This, 0x02, 0x0f, 0x00);
		break;
	case V4L2_DV_RGB_RANGE_FULL:
		/* RGB full range (0-255) */
		io_write_and_or(This, 0x02, 0x0f, 0x10);
		break;
	}
}

static void adv7604_cable_select(CADV7604 *This)
{
    int cable_detect;
    int i;
    int cdet;
    //int det_mask = 0x1e;
	
    cable_detect = io_read(This, 0x6f);
    //printk("::cable_detect %x\n",cable_detect);
	
    for (i=0; i <= PORT_D; i++)
    {
        cdet = ((cable_detect >> (i + 1)) & 0x01 ); /*000ABCD0*/
        if (cdet)
        {
            hdmi_write( This, 0x00, PORT_D - i ); //r10
            //printk("::enable port %d\n",PORT_D - i);
        }	
    }
}


static void adv7604_select_input(CADV7604 *This)
{
    struct adv7604_state *state = &This->state;
    switch (state->mode) 
    {
        case ADV7604_MODE_COMP:
        case ADV7604_MODE_GR:
            /* channel 2 */
            io_write_and_or(This, 0x07, 0xbf, 0x00);

            /* reset ADI recommended settings for HDMI: */
            /* "ADV7604 Register Settings Recommendations (rev. 2.5, June 2010)" p. 4. */
            hdmi_write(This, 0x0d, 0x04); /* HDMI filter optimization */
            hdmi_write(This, 0x3d, 0x00); /* DDC bus active pull-up control */
            hdmi_write(This, 0x3e, 0x74); /* TMDS PLL optimization */
            hdmi_write(This, 0x4e, 0x3b); /* TMDS PLL optimization */
            hdmi_write(This, 0x57, 0x74); /* TMDS PLL optimization */
            hdmi_write(This, 0x58, 0x63); /* TMDS PLL optimization */
            hdmi_write(This, 0x8d, 0x18); /* equaliser */
            hdmi_write(This, 0x8e, 0x34); /* equaliser */
            hdmi_write(This, 0x93, 0x88); /* equaliser */
            hdmi_write(This, 0x94, 0x2e); /* equaliser */
            hdmi_write(This, 0x96, 0x00); /* enable automatic EQ changing */

#if 0
            /* set RGB Pin */
            /* ADC0 Pin7 = 0x07, ADC1 Pin8 = 0x08, ADC2 Pin9 = 0x09 */
            afe_write(This, 0x03, 0x78);
            afe_write(This, 0x04, 0x90);

#endif
            /* enable voltage clamp */
            io_write_and_or(This, 0x30, 0x7f, 0x80);

            /* set ADI recommended settings for digitizer */
            /* "ADV7604 Register Settings Recommendations (rev. 2.5, June 2010)" p. 17. */
            afe_write(This, 0x12, 0x7b); /* ADC noise shaping filter controls */
            afe_write(This, 0x0c, 0x1f); /* CP core gain controls */
            cp_write_and_or(This, 0x3e, 0xfb, 0x04); /* CP core pre-gain control */
            cp_write(This, 0xc3, 0x39); /* CP coast control. Graphics mode */
            cp_write(This, 0x40, 0x5c); /* CP core pre-gain control. Graphics mode */
            break;

        case ADV7604_MODE_HDMI:
            /* channel 1 */
            io_write_and_or(This, 0x07, 0xbf, 0x40);

            /* set ADI recommended settings for HDMI: */
            /* "ADV7604 Register Settings Recommendations (rev. 2.5, June 2010)" p. 4. */
            hdmi_write(This, 0x0d, 0x84); /* HDMI filter optimization */
            hdmi_write(This, 0x3d, 0x10); /* DDC bus active pull-up control */
            hdmi_write(This, 0x3e, 0x39); /* TMDS PLL optimization */
            hdmi_write(This, 0x4e, 0x3b); /* TMDS PLL optimization */
            hdmi_write(This, 0x57, 0xb6); /* TMDS PLL optimization */
            hdmi_write(This, 0x58, 0x03); /* TMDS PLL optimization */
            hdmi_write(This, 0x8d, 0x18); /* equaliser */
            hdmi_write(This, 0x8e, 0x34); /* equaliser */
            hdmi_write(This, 0x93, 0x8b); /* equaliser */
            hdmi_write(This, 0x94, 0x2d); /* equaliser */
            hdmi_write(This, 0x96, 0x01); /* enable automatic EQ changing */

            /* windows driver 0x00b */
            hdmi_write(This, 0x9d, 0x01); /* THDS PLL Gear Optimization */
            afe_write(This, 0xc8, 0x40); /* phase control */

            /* reset ADI recommended settings for digitizer */
            /* "ADV7604 Register Settings Recommendations (rev. 2.5, June 2010)" p. 17. */

            /* windows driver 0x7b  ADI recommended write */
            afe_write(This, 0x12, 0xfb); /* ADC noise shaping filter controls */

            /* windows driver 0x1f  ADI recommended write */
            afe_write(This, 0x0c, 0x0d); /* CP core gain controls */
            cp_write_and_or(This, 0x3e, 0xfb, 0x00); /* CP core pre-gain control */
            cp_write(This, 0xc3, 0x39); /* CP coast control. Graphics mode */
            cp_write(This, 0x40, 0x80); /* CP core pre-gain control. Graphics mode */

            break;
        default:
            //adv7604_printk(KERN_INFO, This, "%s: Unknown mode %d\n",__func__, mode);
            break;
    }
}


static void adv7604_print_timings(CADV7604 *This,
	struct v4l2_dv_timings *timings, const char *txt, bool detailed)
{
    struct v4l2_bt_timings *bt = &timings->bt;
    u32 htot, vtot;

    if (timings->type != V4L2_DV_BT_656_1120)
        return;

    htot = htotal(bt);
    vtot = vtotal(bt);

    if (detailed)
    {
        adv7604_info(This, "%s %dx%d%s%d (%dx%d)",
                     txt, bt->width, bt->height, bt->interlaced ? "i" : "p",
                     (htot * vtot) > 0 ? ((u32) bt->pixelclock /
                     (htot * vtot)) : 0,
                     htot, vtot);

        adv7604_info(This, "    horizontal: fp = %d, %ssync = %d, bp = %d\n",
                     bt->hfrontporch,
                     (bt->polarities & V4L2_DV_HSYNC_POS_POL) ? "+" : "-",
                     bt->hsync, bt->hbackporch);
        adv7604_info(This, "    vertical: fp = %d, %ssync = %d, bp = %d\n",
                     bt->vfrontporch,
                     (bt->polarities & V4L2_DV_VSYNC_POS_POL) ? "+" : "-",
                     bt->vsync, bt->vbackporch);
        if (bt->interlaced)
            adv7604_info(This, "    interlace: il_vfrontporch = %d, il_vsync = %d, il_vbackporch = %d\n",
                         bt->il_vfrontporch, bt->il_vsync, bt->il_vbackporch);

        adv7604_info(This, "    pixelclock: %lld, flags: 0x%x, standards: 0x%x\n",
                     bt->pixelclock, bt->reserved[0], bt->reserved[1]);

    }else{
        adv7604_info(This, "%s %dx%d%s%d (%dx%d)\n",
                     txt, bt->width, bt->height, bt->interlaced ? "i" : "p",
                     (htot * vtot) > 0 ? ((u32) bt->pixelclock /
                     (htot * vtot)) : 0,
                     htot, vtot);
    }
}

struct stdi_readback {
	u16 bl, lcf, lcvs;
	u8 hs_pol, vs_pol;
	bool interlaced;
};


/* ----------------------------------------------------------------------- */

int read_stdi(void *This, struct stdi_readback *stdi, int ch)
{
    TQP_UINT8 polarity;
    TQP_UINT8 regs[5];
    int val[5];
    int i;

    if(ch == 1) 
    {
        regs[0] = 0xb1;
        regs[1] = 0xb2;
        regs[2] = 0xb3;
        regs[3] = 0xb4;
        regs[4] = 0xb5;
    }
    else if(ch == 2)
    {
        regs[0] = 0x49;
        regs[1] = 0x4a;
        regs[2] = 0x4b;
        regs[3] = 0x4c;
        regs[4] = 0x4f;
    }

    if (no_lock_stdi(This, ch) || no_lock_sspd(This, ch)) 
    {
        //adv7604_dbg(2, debug, This, "%s: channel %d STDI and/or SSPD not locked\n", __func__, ch);
        return -1;
    }

    for(i = 0; i < 5; i++)
        val[i] = cp_read(This, regs[i]);

    /* read STDI */
    stdi->bl = ((val[0] & 0x3f) << 8) | val[1];
    stdi->lcf = ((val[2] & 0x7) << 8) | val[3];
    stdi->lcvs = val[2] >> 3;
    stdi->interlaced = val[0] & 0x40;

    polarity = val[4];

    /* read SSPD */
    /* activity detected on Hsync and VSync or embedded sync to Sync Channel 1 SSPD */
    if ((polarity & 0x01) == 0x01) 
    {
        /* Activity detected on embedded sync to Sync CH1 SSPD */
        if((polarity & 0x04) == 0) {
        }
        else {
        }

        stdi->hs_pol = (polarity & 0x08) ? '+' : '-';
        stdi->vs_pol = (polarity & 0x20) ? '+' : '-';

        //adv7604_dbg(2, debug, This, "%s: STDI hs_pol = %c, vs_pol = %c\n", __func__, stdi->hs_pol, stdi->vs_pol);
    }else{
        stdi->hs_pol = 'x';
        stdi->vs_pol = 'x';
    }

    if (no_lock_stdi(This, ch) || no_lock_sspd(This, ch)) 
    {
        //adv7604_dbg(2, debug, This,"%s: signal lost during readout of STDI/SSPD\n", __func__);
        return -1;
    }

    if (stdi->lcf < 239 || stdi->bl < 8 || stdi->bl == 0x3fff) 
    {
        //adv7604_dbg(2, debug, This, "%s: invalid signal\n", __func__);
        memset(stdi, 0, sizeof(struct stdi_readback));
        return -1;
    }

    return 0;
}

static int stdi2dv_timings(CADV7604 *This,
		struct stdi_readback *stdi,
		struct v4l2_dv_timings *timings)
{
    //struct adv7604_state *state = &This->state;
    u32 hfreq = (ADV7604_fsc * 8) / stdi->bl;
    u32 pix_clk;
    int i;

    for (i = 0; adv7604_timings[i].bt.height; i++) 
    {
        if (vtotal(&adv7604_timings[i].bt) != stdi->lcf + 1)
            continue;
        if (adv7604_timings[i].bt.vsync != stdi->lcvs)
            continue;

        pix_clk = hfreq * htotal(&adv7604_timings[i].bt);
        if ((pix_clk < adv7604_timings[i].bt.pixelclock + 1000000) &&
            (pix_clk > adv7604_timings[i].bt.pixelclock - 1000000)) 
        {
            *timings = adv7604_timings[i];
            return 0;
        }
    }

    adv7604_info(This, "%s stdi->lcf = %u, hfreq = %u, stdi->lcvs = %u\n",
            __func__, stdi->lcf, hfreq, stdi->lcvs);

    adv7604_dbg(2, debug, This,
		"%s: No format candidate found for lcvs = %d, lcf=%d, bl = %d, %chsync, %cvsync\n",
		__func__, stdi->lcvs, stdi->lcf, stdi->bl,
		stdi->hs_pol, stdi->vs_pol);
    return -1;
}


/* Fill the optional fields .standards and .flags in struct v4l2_dv_timings
   if the format is listed in adv7604_timings[] */
static void adv7604_fill_optional_dv_timings_fields(CADV7604 *This,
		struct v4l2_dv_timings *timings)
{
    struct adv7604_state *state = &This->state;
    int i;

    for (i = 0; adv7604_timings[i].bt.width; i++) 
    {
        if (v4l_match_dv_timings(timings, &adv7604_timings[i], DIGITAL_INPUT ? 250000 : 1000000)) 
        {
            *timings = adv7604_timings[i];
            break;
        }
    }
}

static int adv7604_query_dv_timings(CADV7604 *This,
			struct v4l2_dv_timings *timings, int ch)
{
    struct adv7604_state *state = &This->state;
    struct v4l2_bt_timings *bt = &timings->bt;
    struct stdi_readback stdi;

    if (!timings)
        return -EINVAL;

    memset(timings, 0, sizeof(struct v4l2_dv_timings));

    if (no_signal(This, ch)) 
    {
        adv7604_dbg(1, debug, This, "%s: no valid signal\n", __func__);
        return -ENOLINK;
    }

    /* read STDI */
    if (read_stdi(This, &stdi, ch)) 
    {
        adv7604_dbg(1, debug, This, "%s: STDI/SSPD not locked\n", __func__);
        return -ENOLINK;
    }
    bt->interlaced = stdi.interlaced ? V4L2_DV_INTERLACED : V4L2_DV_PROGRESSIVE;

    if (DIGITAL_INPUT && ch == 1) 
    {
        timings->type = V4L2_DV_BT_656_1120;

        bt->width = (hdmi_read(This, 0x07) & 0x0f) * 256 + hdmi_read(This, 0x08);
        bt->height = (hdmi_read(This, 0x09) & 0x0f) * 256 + hdmi_read(This, 0x0a);
        bt->pixelclock = (hdmi_read(This, 0x06) * 1000000) + ((hdmi_read(This, 0x3b) & 0x30) >> 4) * 250000;
        bt->hfrontporch = (hdmi_read(This, 0x20) & 0x03) * 256 + hdmi_read(This, 0x21);
        bt->hsync = (hdmi_read(This, 0x22) & 0x03) * 256 + hdmi_read(This, 0x23);
        bt->hbackporch = (hdmi_read(This, 0x24) & 0x03) * 256 + hdmi_read(This, 0x25);
        bt->vfrontporch = ((hdmi_read(This, 0x2a) & 0x1f) * 256 + hdmi_read(This, 0x2b)) / 2;
        bt->vsync = ((hdmi_read(This, 0x2e) & 0x1f) * 256 + hdmi_read(This, 0x2f)) / 2;
        bt->vbackporch = ((hdmi_read(This, 0x32) & 0x1f) * 256 + hdmi_read(This, 0x33)) / 2;
        bt->polarities = ((hdmi_read(This, 0x05) & 0x10) ? V4L2_DV_VSYNC_POS_POL : 0) | ((hdmi_read(This, 0x05) & 0x20) ? V4L2_DV_HSYNC_POS_POL : 0);
        if (bt->interlaced == V4L2_DV_INTERLACED) 
        {
            bt->height += (hdmi_read(This, 0x0b) & 0x0f) * 256 + hdmi_read(This, 0x0c);
            bt->il_vfrontporch = ((hdmi_read(This, 0x2c) & 0x1f) * 256 + hdmi_read(This, 0x2d)) / 2;
            bt->il_vsync = ((hdmi_read(This, 0x30) & 0x1f) * 256 + hdmi_read(This, 0x31)) / 2;
            bt->vbackporch = ((hdmi_read(This, 0x34) & 0x1f) * 256 + hdmi_read(This, 0x35)) / 2;

            /* Set vertical back porch for the even field (aka field 2) */
            bt->il_vbackporch = bt->vbackporch + 1;
        }
        adv7604_fill_optional_dv_timings_fields(This, timings);
    } else {
        /* find format
         * Since LCVS values are inaccurate [REF_03, p. 275-276],
         * stdi2dv_timings() is called with lcvs +-1 if the first attempt fails.
         */
        if (!stdi2dv_timings(This, &stdi, timings))
            goto found;
        stdi.lcvs += 1;
        adv7604_dbg(1, debug, This, "%s: lcvs + 1 = %d\n", __func__, stdi.lcvs);
        if (!stdi2dv_timings(This, &stdi, timings))
            goto found;
        stdi.lcvs -= 2;
        adv7604_dbg(1, debug, This, "%s: lcvs - 1 = %d\n", __func__, stdi.lcvs);
        if (!stdi2dv_timings(This, &stdi, timings))
            goto found;
        stdi.lcvs += 1;
        stdi.lcf += 2;
        adv7604_dbg(1, debug, This, "%s: lcf + 2 = %d\n", __func__, stdi.lcf);
        if (stdi2dv_timings(This, &stdi, timings)) 
        {
            /*
             * The STDI block may measure wrong values, especially
             * for lcvs and lcf. If the driver can not find any
             * valid timing, the STDI block is restarted to measure
             * the video timings again. The function will return an
             * error, but the restart of STDI will generate a new
             * STDI interrupt and the format detection process will
             * restart.
             */
            if (state->restart_stdi_once) 
            {
                adv7604_dbg(1, debug, This, "%s: restart STDI\n", __func__);
                /* TODO restart STDI for Sync Channel 2 */
                /* enter one-shot mode */
                cp_write_and_or(This, 0x86, 0xf9, 0x00);
                /* trigger STDI restart */
                cp_write_and_or(This, 0x86, 0xf9, 0x04);
                /* reset to continuous mode */
                cp_write_and_or(This, 0x86, 0xf9, 0x02);
                state->restart_stdi_once = false;
                return -ENOLINK;
            }
            adv7604_dbg(1, debug, This, "%s: format not supported\n", __func__);
            return -ERANGE;
        }
        state->restart_stdi_once = true;
    }
found:

    if (no_signal(This, ch)) 
    {
        adv7604_dbg(1, debug, This, "%s: signal lost during readout\n", __func__);
        memset(timings, 0, sizeof(struct v4l2_dv_timings));
        return -ENOLINK;
    }

    if ((!DIGITAL_INPUT && bt->pixelclock > 170000000) || (DIGITAL_INPUT && bt->pixelclock > 225000000)) 
    {
        adv7604_dbg(1, debug, This, "%s: pixelclock out of range %d\n", __func__, (u32)bt->pixelclock);
        return -ERANGE;
    }

    if (debug > 1)
        adv7604_print_timings(This, timings, "adv7604_query_dv_timings:", true);

    return 0;
}

//////////////////////////////////////////////////

static  __attribute__((__unused__)) void adv7604_detect_std(CADV7604 *This, struct adv7604_video_standards *std)
{
}

//////////////////////////////////////////////////
static int find_and_set_predefined_video_timings(CADV7604 *This,
		u8 prim_mode,
		const struct adv7604_video_standards *predef_vid_timings,
		const struct v4l2_dv_timings *timings)
{
    struct adv7604_state *state = &This->state;
    int i;

    adv7604_info(This, "%s prim_mode=%d\n", __func__, prim_mode);
    for (i = 0; predef_vid_timings[i].timings.bt.width; i++) 
    {
        if (!v4l_match_dv_timings(timings, &predef_vid_timings[i].timings, DIGITAL_INPUT ? 250000 : 1000000))
            continue;

        adv7604_info(This, "%s vid_std=%d\n", __func__, predef_vid_timings[i].vid_std);
        io_write(This, 0x00, predef_vid_timings[i].vid_std); /* video std */
        io_write(This, 0x01, (predef_vid_timings[i].v_freq << 4) + prim_mode); /* v_freq and prim mode */
        return 0;
    }

    return -1;
}

static int configure_predefined_video_timings(CADV7604 *This, struct v4l2_dv_timings *timings)
{
    struct adv7604_state *state = &This->state;
    int err;

    adv7604_info(This, "%s mode=%d\n", __func__, state->mode);

    adv7604_configure_video_timings(This);

    switch (state->mode) 
    {
        case ADV7604_MODE_COMP:
        case ADV7604_MODE_GR:
            err = find_and_set_predefined_video_timings(This, 0x01, adv7604_prim_mode_comp, timings);
            if (err)
                err = find_and_set_predefined_video_timings(This, 0x02, adv7604_prim_mode_gr, timings);

            /* enable adc hdmi simult mode */
            io_write_and_or(This, 0x01, 0x7f, 0x80); 
            break;
        case ADV7604_MODE_HDMI:
            err = find_and_set_predefined_video_timings(This, 0x05, adv7604_prim_mode_hdmi_comp, timings);
            if (err)
                err = find_and_set_predefined_video_timings(This, 0x06, adv7604_prim_mode_hdmi_gr, timings);
            break;
        default:
            adv7604_printk(KERN_INFO, This, "%s: Unknown mode %d\n", __func__, state->mode);
            err = -1;
            break;
    }


    return err;
}



static void _configure_custom_video_timings(CADV7604 *This,
		const struct v4l2_bt_timings *bt)
{
	struct adv7604_state *state = &This->state;
	u32 width = htotal(bt);
	u32 height = vtotal(bt);
	u16 cp_start_sav = bt->hsync + bt->hbackporch - 4;
	u16 cp_start_eav = width - bt->hfrontporch;
	u16 cp_start_vbi = height - bt->vfrontporch;
	u16 cp_end_vbi = bt->vsync + bt->vbackporch;
	u16 ch1_fr_ll = (((u32)bt->pixelclock / 100) > 0) ?
		((width * (ADV7604_fsc / 100)) / ((u32)bt->pixelclock / 100)) : 0;
//	const u8 pll[2] = {
//		0xc0 | ((width >> 8) & 0x1f),
//		width & 0xff
//	};
//	TQP_UINT8 pll_block[] =
//	{
//		ADV7604_DEVICE_ADDR_IO, 0x16, 0xc0 | ((width >> 8) & 0x1f),
//		ADV7604_DEVICE_ADDR_IO, 0x17, width & 0xff,
//	};

	adv7604_info(This, "%s width = %u, height = %u\n", __func__, width, height);

	switch (state->mode) {
	case ADV7604_MODE_COMP:
	case ADV7604_MODE_GR:
		/* auto graphics */
		io_write(This, 0x00, 0x07); /* video std */
		io_write(This, 0x01, 0x02); /* prim mode */

        /* GR_AV_BL_EN */
		/* enable embedded syncs for auto graphics mode */
    	cp_write_and_or(This, 0x81, 0xef, 0x10);

        /* DATA_BLANK_EN AVCODE_INSERT_EN*/
//    	io_write_and_or(This, 0x05, 0xf3, 0x0c);


		/* Should only be set in auto-graphics mode [REF_02, p. 91-92] */
		/* setup PLL_DIV_MAN_EN and PLL_DIV_RATIO */
		/* IO-map reg. 0x16 and 0x17 should be written in sequence */
//		CADV7604_WriteBlock(This, pll_block);
//
    	adv7604_info(This, "%s %x\n", __func__, ((width >> 8) & 0x1f));
    	adv7604_info(This, "%s %x\n", __func__, (width & 0xff));
        io_write(This, 0x16, 0xc0 | ((width >> 8) & 0x1f));
        io_write(This, 0x17, width & 0xff);

		/* active video - horizontal timing */
    	adv7604_info(This, "%s cp_start_sav = %x\n", __func__, cp_start_sav);
    	adv7604_info(This, "%s cp_start_eav = %x\n", __func__, cp_start_eav);
    	adv7604_info(This, "%s cp_start_vbi = %x\n", __func__, cp_start_vbi);
    	adv7604_info(This, "%s cp_end_vbi = %x\n", __func__, cp_end_vbi);
		cp_write(This, 0xa2, (cp_start_sav >> 4) & 0xff);
		cp_write(This, 0xa3, ((cp_start_sav & 0x0f) << 4) |
					((cp_start_eav >> 8) & 0x0f));
		cp_write(This, 0xa4, cp_start_eav & 0xff);

		/* active video - vertical timing */
		cp_write(This, 0xa5, (cp_start_vbi >> 4) & 0xff);
		cp_write(This, 0xa6, ((cp_start_vbi & 0xf) << 4) |
					((cp_end_vbi >> 8) & 0xf));
		cp_write(This, 0xa7, cp_end_vbi & 0xff);
		break;
	case ADV7604_MODE_HDMI:
		/* set default prim_mode/vid_std for HDMI
		   accoring to [REF_03, c. 4.2] */
		io_write(This, 0x00, 0x02); /* video std */
		io_write(This, 0x01, 0x06); /* prim mode */
		break;
	default:
		adv7604_printk(KERN_INFO, This, "%s: Unknown mode %d\n",
				__func__, state->mode);
		break;
	}

	cp_write(This, 0x8f, (ch1_fr_ll >> 8) & 0x7);
	cp_write(This, 0x90, ch1_fr_ll & 0xff);
    adv7604_info(This, "%s %x\n", __func__, ((height >> 4) & 0xff));
    adv7604_info(This, "%s %x\n", __func__, ((height & 0x0f) << 4));
	cp_write(This, 0xab, (height >> 4) & 0xff);
	cp_write(This, 0xac, (height & 0x0f) << 4);
}

#if 1
/* v4l2 ioctl handling */
static int adv7604_s_dv_timings(CADV7604 *This, struct v4l2_dv_timings *timings)
{
	struct adv7604_state *state = &This->state;
	struct v4l2_bt_timings *bt;
	u32 htot, vtot;

	int err;

	if (!timings)
		return -EINVAL;

	bt = &timings->bt;

	if ((!DIGITAL_INPUT && bt->pixelclock > 170000000) ||
			(DIGITAL_INPUT && bt->pixelclock > 225000000)) {
		printk(KERN_INFO "%s: pixelclock out of range %d\n",
				__func__, (u32)bt->pixelclock);
		return -ERANGE;
	}

	adv7604_fill_optional_dv_timings_fields(This, timings);

	state->timings = *timings;

	cp_write(This, 0x91, bt->interlaced ? 0x50 : 0x10);

	/* Use prim_mode and vid_std when available */
	err = configure_predefined_video_timings(This, timings);
	if (err) {
		/* custom settings when the video format
		 does not have prim_mode/vid_std */
		_configure_custom_video_timings(This, bt);
	}

	set_rgb_quantization_range(This);

	htot = htotal(bt);
	vtot = vtotal(bt);
	state->frame_rate = (u32)bt->pixelclock / (htot * vtot);


	if (1)
		adv7604_print_timings(This, timings,
				"adv7604_s_dv_timings:", true);


	return 0;
}

static  __attribute__((__unused__)) int adv7604_g_dv_timings(CADV7604 *This,
		struct v4l2_dv_timings *timings)
{
	struct adv7604_state *state = &This->state;

	*timings = state->timings;
	return 0;
}
#endif

/* v4l2 control operations */
static int adv7604_s_ctrl(void *frontend, struct v4l2_control *ctrl)
{
	CADV7604 *This = (CADV7604 *) container_of(frontend,CVideoDecoder,m_frontend);
	struct adv7604_state *state = &This->state;
    CVideoDecoder *pVideoDecoder = &This->m_VideoDecoder;
	CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib,
			m_Object, pVideoDecoder->m_Object.m_pParent);
	CQLCodec *pQLCodec = codeclib->m_pMpgCodec;
	CDevice *cdevice = (CDevice *)pQLCodec->m_pDevice;
	CComponent *component = cdevice->m_pComponent;
/*
	IAudioCodec *i_audio_codec = CDevice_getAudioCodec(cdevice);
    CAudioCodec *pAudioCodec = GET_INHERITED_OBJECT(CAudioCodec,
			m_iAudCodec, i_audio_codec);
    CAIC3101 *aic3101 = GET_INHERITED_OBJECT(CAIC3101, m_AudioCodec, pAudioCodec);
*/

	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		adv7604_info(This, "%s set brightness value=%d\n", __func__, ctrl->value);
		cp_write(This, 0x3c, ctrl->value);
		return 0;
	case V4L2_CID_CONTRAST:
		adv7604_info(This, "%s set contrast value=%d\n", __func__, ctrl->value);
		cp_write(This, 0x3a, ctrl->value);
		return 0;
	case V4L2_CID_SATURATION:
		adv7604_info(This, "%s set saturation value=%d\n", __func__, ctrl->value);
		cp_write(This, 0x3b, ctrl->value);
		return 0;
	case V4L2_CID_HUE:
		adv7604_info(This, "%s set hue value=%d\n", __func__, ctrl->value);
		cp_write(This, 0x3d, ctrl->value);
		return 0;
	case V4L2_CID_MPEG_VIDEO_BITRATE:
		adv7604_info(This, "%s set video bitrate=%d\n", __func__, ctrl->value);
		state->bit_rate = ctrl->value / 1000;
		pl330_set_config(component, "RateControl-AvgBitRate", &state->bit_rate, true);
		pl330_set_config(component, "VbrBitRate-MinBitRate", &state->bit_rate, true);
		pl330_set_config(component, "VbrBitRate-MaxBitRate", &state->bit_rate, true);
		return 0;
/*
	case V4L2_CID_AUDIO_MUTE:
		adv7604_info(This, "%s set audio mute=%d\n", __func__, ctrl->value);
		if(DIGITAL_INPUT)
			hdmi_write(This, 0x1a, (ctrl->value << 4) | 0xa);
		else
			aic3101_mute(aic3101, ctrl->value);
		return 0;
*/
#if 0
	case  V4L2_CID_DV_RX_RGB_RANGE:
		state->rgb_quantization_range = ctrl->val;
		set_rgb_quantization_range(This);
		return 0;
	case V4L2_CID_ADV_RX_ANALOG_SAMPLING_PHASE:
		/* Set the analog sampling phase. This is needed to find the
		   best sampling phase for analog video: an application or
		   driver has to try a number of phases and analyze the picture
		   quality before settling on the best performing phase. */
		afe_write(This, 0xc8, ctrl->val);
		return 0;
	case V4L2_CID_ADV_RX_FREE_RUN_COLOR_MANUAL:
		/* Use the default blue color for free running mode,
		   or supply your own. */
		cp_write_and_or(This, 0xbf, ~0x04, (ctrl->val << 2));
		return 0;
	case V4L2_CID_ADV_RX_FREE_RUN_COLOR:
		cp_write(This, 0xc0, (ctrl->val & 0xff0000) >> 16);
		cp_write(This, 0xc1, (ctrl->val & 0x00ff00) >> 8);
		cp_write(This, 0xc2, (u8)(ctrl->val & 0x0000ff));
		return 0;
#endif
	}
	return -EINVAL;
}

static int adv7604_g_ctrl(void *frontend, struct v4l2_control *ctrl)
{
	CADV7604 *This = (CADV7604 *) container_of(frontend,CVideoDecoder,m_frontend);
	struct adv7604_state *state = &This->state;

	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		ctrl->value = cp_read(This, 0x3c);
		return 0;
	case V4L2_CID_CONTRAST:
		ctrl->value = cp_read(This, 0x3a);
		return 0;
	case V4L2_CID_SATURATION:
		ctrl->value = cp_read(This, 0x3b);
		return 0;
	case V4L2_CID_HUE:
		ctrl->value = cp_read(This, 0x3d);
		return 0;
	case V4L2_CID_MPEG_VIDEO_BITRATE:
		ctrl->value = state->bit_rate * 1000;
		return 0;
	}
	return -EINVAL;
}

static void adv7604_configure_resolution(CADV7604 *This, CComponent *component,
		struct pl330_frame *frame, struct pl330_format *format,
		COMPONENT_PORT_OPEN *port_open)
{
	struct adv7604_state *state = &This->state;
	struct v4l2_dv_timings *timings = &state->timings;
	struct v4l2_bt_timings *bt = &timings->bt;
	
	int w = frame->width;
	int h = frame->height;

	adv7604_info(This, "%s %d:%d\n", __func__,
		w, h);

//    bt->width = 1280;
//    bt->height = 800;
	pl330_set_config(component, "PictureResolution-InPicWidth",
			&bt->width, true);
	pl330_set_config(component, "PictureResolution-InPicHeight",
			&bt->height, true);

	if(w == 0 || h == 0 || w > bt->width || h > bt->height ||
	   (format->pixelformat == V4L2_PIX_FMT_YVU420) ||
	   (format->pixelformat == V4L2_PIX_FMT_YUV420) ||
	   (format->pixelformat == V4L2_PIX_FMT_NV12) )
	{
		pl330_set_config(component, "OutPictureResolution-OutPicWidth",
				&bt->width, true);
		pl330_set_config(component, "OutPictureResolution-OutPicHeight",
				&bt->height, true);

		w = bt->width;
		h = bt->height;
	}
	else
	{
		pl330_set_config(component, "OutPictureResolution-OutPicWidth",
				&w, true);
		pl330_set_config(component, "OutPictureResolution-OutPicHeight",
				&h, true);
	}

	state->cur_out_frame.width = w;
	state->cur_out_frame.height = h;
	adv7604_info(This, "%s current out frame %ux%u\n", __func__,
		w, h);

	if((format->pixelformat == V4L2_PIX_FMT_YVU420) || 
    (format->pixelformat == V4L2_PIX_FMT_YUV420) || 
    (format->pixelformat == V4L2_PIX_FMT_NV12))
	{
		port_open->format.yuv.nWidth = w;
		port_open->format.yuv.nHeight = h;
	}
}

static int adv7604_try_frame(void *frontend, struct pl330_frame *frame,
			struct pl330_format *format)
{
    CADV7604 *This = (CADV7604 *) container_of(frontend,CVideoDecoder,m_frontend);
    CVideoDecoder *pVideoDecoder = &This->m_VideoDecoder;
    CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_Object, pVideoDecoder->m_Object.m_pParent);
    CQLCodec *pQLCodec = codeclib->m_pMpgCodec;
    CDevice *cdevice = (CDevice *)pQLCodec->m_pDevice;
    CComponent *component = cdevice->m_pComponent;
    struct adv7604_state *state = &This->state;

    COMPONENT_PORT_OPEN port_open;
    memset(&port_open, 0, sizeof(COMPONENT_PORT_OPEN));

    adv7604_info(This, "%s %ux%u input=%ux%u@%u%c.\n",
			__func__,
			frame->width, frame->height,
			state->timings.bt.width, state->timings.bt.height,
			state->frame_rate,
			state->timings.bt.interlaced ? 'i' : 'p');

    if(state->timings.bt.width == 0 && state->timings.bt.height == 0)
    {
        adv7604_info(This, "input no signal\n");
        return -EIO;
    }

    if(!state->frame_rate)
        return -EINVAL;

    /* configure PL330 */
    adv7604_configure_pl330(This, &state->timings);

    /* timing 1280x800 VGA need to use AutoGraphic Mode */
    if(state->timings.bt.width == 1280 && state->timings.bt.height == 800)
    {
        adv7604_print_timings(This, &state->timings,"adv7604_try_frame", true);
        _configure_custom_video_timings(This, &state->timings.bt);
    }
    if(state->timings.bt.width == 1440 && state->timings.bt.height == 900)
    {
        adv7604_print_timings(This, &state->timings, "adv7604_try_frame", true);
        _configure_custom_video_timings(This, &state->timings.bt);
    }
    if(state->timings.bt.width == 1680 && state->timings.bt.height == 1050)
    {
        adv7604_print_timings(This, &state->timings, "adv7604_try_frame", true);
        _configure_custom_video_timings(This, &state->timings.bt);
    }

	if((format->pixelformat == V4L2_PIX_FMT_YVU420) || 
    (format->pixelformat == V4L2_PIX_FMT_YUV420) || 
    (format->pixelformat == V4L2_PIX_FMT_NV12))
	{
		if(frame->width != state->timings.bt.width ||
		   frame->height != state->timings.bt.height)
		{
			adv7604_info(This, "%s pixelformat is YV12 frame resolution %ux%u "
					"must be equal to source.\n",
					__func__, frame->width, frame->height);
		}
	}

	adv7604_configure_resolution(This, component, frame, format, &port_open);

	adv7604_info(This, "frame_rate=%u\n", state->out_frame_rate);

    if(state->out_frame_rate == 0)
        state->out_frame_rate = state->frame_rate;

	/* if resolution > 1920x1080p set frame rate to 30 */
    if(state->out_frame_rate >= 30) {
    	if(state->cur_out_frame.width == 1920)
		    if(state->timings.bt.interlaced == V4L2_DV_PROGRESSIVE)
			    state->out_frame_rate = 30;
	    else if(state->cur_out_frame.width == 1680)
	    	if(state->timings.bt.interlaced == V4L2_DV_PROGRESSIVE)
		    	state->out_frame_rate = 30;
    }

//    state->out_frame_rate = 60;
    adv7604_info(This, "%s frame_rate set to %d\n", __func__, state->out_frame_rate);
    pl330_set_config(component, "InputControl-OutFrameRate",
            &state->out_frame_rate, true);

	adv7604_info(This, "frame_rate=%u\n", state->out_frame_rate);

	switch(format->pixelformat)
	{
		case V4L2_PIX_FMT_MPEG:
			pl330_set_config(component, "SystemControl-StreamType", "ts", false);
			pl330_set_config(component, "SysFunction-Video", "h264", false);
			port_open.type = COMPONENT_PORT_COMP_OUT;
			port_open.format.vid.nDataType = QP_COMP_VID_DATA_TYPE_H264;

			/* set MPEG audio */
			adv7604_configure_pl330_audio(This, state);
			break;
		case V4L2_PIX_FMT_YVU420:
        case V4L2_PIX_FMT_NV12:
			pl330_set_config(component, "SystemControl-StreamType", "raw", false);
			pl330_set_config(component, "SysFunction-Video", "passthrough", false);
			port_open.type = COMPONENT_PORT_YUV_OUT;
			port_open.format.yuv.nBitCount = 12;
			port_open.format.yuv.nDataType = QP_YUV_DATA_TYPE_YV12;
			port_open.format.yuv.nFrameRate = state->out_frame_rate;
			break;
		case V4L2_PIX_FMT_YUV420:
			pl330_set_config(component, "SystemControl-StreamType", "raw", false);
			pl330_set_config(component, "SysFunction-Video", "passthrough", false);
			port_open.type = COMPONENT_PORT_YUV_OUT;
			port_open.format.yuv.nBitCount = 12;
			port_open.format.yuv.nDataType = QP_YUV_DATA_TYPE_I420;
			port_open.format.yuv.nFrameRate = state->out_frame_rate;
			break;
	}

	port_open.memory = V4L2_MEMORY_MMAP;

	pl330_port_open(component, &port_open);

	pl330_acquire(component);

	return 0;
}



static void adv7604_configure_pl330_audio(CADV7604 *This, struct adv7604_state *state)
{
    CVideoDecoder *pVideoDecoder = &This->m_VideoDecoder;
    CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_Object, pVideoDecoder->m_Object.m_pParent);
    CQLCodec *pQLCodec = codeclib->m_pMpgCodec;

    CDevice *cdevice = (CDevice *)pQLCodec->m_pDevice;
    CComponent *component = cdevice->m_pComponent;

    /* Audio Configuration */
    /* can not set passthrough cause REG_TO_ARM_MESSAGE_STATUS status timeout */
    pl330_set_config(component, "SysFunction-Audio", "audio", false);
    pl330_set_config(component, "SysLink-AudioInput", "aiu", false);
    pl330_set_config(component, "SysLink-AudioOutput", "host", false);

    pl330_set_config(component, "AudioControlParam-AudioType", "aac", false);
    pl330_set_config(component, "AudioControlParam-ChNum", "2", false);
    pl330_set_config(component, "AudioControlEx-AacVer", "mpeg2", false);
    pl330_set_config(component, "AudioControlEx-HType", "adts", false);
    pl330_set_config(component, "AudioControlEx-CutoffFreq", "18000", false);
    pl330_set_config(component, "AudioControlEx-TNS", "1", false);
    pl330_set_config(component, "AudioControlEx-IS", "1", false);
    pl330_set_config(component, "AudioControlEx-PNS", "1", false);
    pl330_set_config(component, "AudioControlEx-MS", "1", false);

    /* set sample rate */
    switch(state->rates)
    {
        case AUDIO_RATE_32000:
            pl330_set_config(component, "AudioControlParam-SampleRate", "32k", false);
            break;
        case AUDIO_RATE_44100:
            pl330_set_config(component, "AudioControlParam-SampleRate", "44.1k", false);
            break;
        case AUDIO_RATE_48000:
            pl330_set_config(component, "AudioControlParam-SampleRate", "48k", false);
            break;
        case AUDIO_RATE_88200:
            pl330_set_config(component, "AudioControlParam-SampleRate", "88.2k", false);
            break;
        case AUDIO_RATE_96000:
            pl330_set_config(component, "AudioControlParam-SampleRate", "96k", false);
            break;
        case AUDIO_RATE_176400:
            pl330_set_config(component, "AudioControlParam-SampleRate", "176.4k", false);
            break;
        case AUDIO_RATE_192000:
            pl330_set_config(component, "AudioControlParam-SampleRate", "192k", false);
            break;
        default:
            break;
    }
}

static int adv7604_configure_pl330(CADV7604 *This, struct v4l2_dv_timings *timings)
{
    CVideoDecoder *pVideoDecoder = &This->m_VideoDecoder;
    CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_Object, pVideoDecoder->m_Object.m_pParent);
    CQLCodec *pQLCodec = codeclib->m_pMpgCodec;

    CDevice *cdevice = (CDevice *)pQLCodec->m_pDevice;
    CComponent *component = cdevice->m_pComponent;
    struct adv7604_state *state = &This->state;
    struct v4l2_bt_timings *bt = &timings->bt;

    u32 frame_rate;
    u8 sync;
    COMPONENT_PORT_OPEN port_open;

    if(!component)
        return -EIO;

    memset(&port_open, 0, sizeof(COMPONENT_PORT_OPEN));

    frame_rate = state->frame_rate;
	
    adv7604_info(This, "%s frame_rate = %d\n", __func__, frame_rate);
    adv7604_info(This, "%s bit_rate = %d\n", __func__, state->bit_rate);
    adv7604_info(This, "%s interlaced = %d\n", __func__, bt->interlaced);

    pl330_set_config(component, "SystemControl-StreamData", "video+audio", false);
    pl330_set_config(component, "SystemControl-XferMode", "frame", false);
    pl330_set_config(component, "SystemControl-SpsrFreq", "1", false);
    pl330_set_config(component, "SystemControl-FFMode", "frame", false);
    pl330_set_config(component, "SystemControl-VMode", "cavlc", false);

    if(bt->interlaced)
    {
        pl330_set_config(component, "InputControl-ScanFormat", "interlace", false);
        pl330_set_config(component, "InputControl-Fmt", "interlace", false);
        pl330_set_config(component, "DeInterlace-Mode", "topfield", false);
        frame_rate /= 2;
    }else{
        pl330_set_config(component, "InputControl-ScanFormat", "progressive", false);
        pl330_set_config(component, "InputControl-Fmt", "progressive", false);
        pl330_set_config(component, "DeInterlace-Mode", "none", false);
    }

    sync = 0;
    if(bt->polarities & V4L2_DV_VSYNC_POS_POL) {
        adv7604_info(This, "V +\n");
    }
    else {
        adv7604_info(This, "V -\n");
        sync |= 1;
    }

    if(bt->polarities & V4L2_DV_HSYNC_POS_POL) {
        adv7604_info(This, "H +\n");
    }
    else {
        adv7604_info(This, "H -\n");
        sync |= (1 << 1);
    }
    adv7604_info(This, "sync =%d\n", sync);
    pl330_set_config(component, "InputControl-SyncMode", &sync, true);


    pl330_set_config(component, "InputControl-SrcMode", "hdmi", false);
    pl330_set_config(component, "InputControl-DataType", "raw", false);
    pl330_set_config(component, "InputControl-InFrameRate", &frame_rate, true);

    pl330_set_config(component, "InputControl-CkEdge", "positive", false);

    pl330_set_config(component, "RateControl-Vbr", "0", false);
    pl330_set_config(component, "RateControl-Mode", "viu", false);

    pl330_set_config(component, "RateControl-AvgBitRate", &state->bit_rate, true);
    pl330_set_config(component, "VbrBitRate-MinBitRate", &state->bit_rate, true);
    pl330_set_config(component, "VbrBitRate-MaxBitRate", &state->bit_rate, true);

    pl330_set_config(component, "FilterControl-StartPixel", "1", false);
    pl330_set_config(component, "FilterControl-StartLine", "0", false);

    pl330_set_config(component, "GopLoopFilter-IntraPeriod", "30", false);
    pl330_set_config(component, "GopLoopFilter-BNum", "0", false);
    pl330_set_config(component, "GopLoopFilter-Idr", "close", false);

    pl330_set_config(component, "SysFunction-Function", "encode", false);
    pl330_set_config(component, "SysLink-VideoInput", "viu", false);
    pl330_set_config(component, "SysLink-VideoOutput", "host", false);

    return 0;
}

static void adv7604_fmtqueue_clean(CADV7604 *This)
{
    struct adv7604_fmt_change *strm_fmt, *next;
    adv7604_info(This, "%s\n", __func__);

    mutex_lock(&This->fmt_lock);
    list_for_each_entry_safe(strm_fmt, next, &This->queue_head, queue_head)
    {
        adv7604_info(This, "%s %d\n", __func__, strm_fmt->fmt_change);
        list_del(&strm_fmt->queue_head);
        kfree(strm_fmt);
    }
    mutex_unlock(&This->fmt_lock);
}

static int adv7604_polling_thread(void *__this)
{
    CADV7604 *this = (CADV7604 *) __this;
     //	struct v4l2_dv_timings timings;
    struct adv7604_state *state = &this->state;
    u8 status_70,raw_6f;
    IFrontend *frontend=&this->m_VideoDecoder.m_frontend;
	
    CVideoDecoder *pVideoDecoder = &this->m_VideoDecoder;
    CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_Object, pVideoDecoder->m_Object.m_pParent);
    CQLCodec *pQLCodec = codeclib->m_pMpgCodec;
   
    io_write(this,0x5f,0xff);
    io_write(this,0x73,0x13);
    for (;;)
    {
        if (kthread_should_stop())
            break;

        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(ADV7604_POLLING_TIMEOUT);

        TQP_UINT8 status_on = 0, hdmi_on = 0,hdmi_off = 0, vga_on = 0, vga_off = 0;
        adv7604_get_params(this, &status_on, &hdmi_on, &hdmi_off, &vga_on, &vga_off);
        if(status_on)
        {
            struct v4l2_dv_timings timings_ch1, timings_ch2;
            if(hdmi_on)
            {
                if (adv7604_query_dv_timings(this, &timings_ch1, 1)==0)
                {
                    adv7604_print_timings(this, &timings_ch1, "ch1 Detected format:", false);
                    frontend_s_input_status(frontend,C353_HDMI,FRONTEND_LOCK);
                    if(this->input == C353_HDMI) 
                    {
                        if (no_lock_cp(this) || !v4l_match_dv_timings(&state->timings,&timings_ch1,250000)) 
                        {
                            adv7604_info(this,"CH1 detected timing is not match setting timings, resetting\n");
                            adv7604_s_dv_timings(this, &timings_ch1);
                            frontend_update_event(frontend,FRONTEND_EVENT_LOCK);
                        }
                    }
                }
                /* get audio sample rate */
                adv7604_audio_sample_rate(this);
            } 

            if(hdmi_off)
            {
                frontend_s_input_status(frontend,C353_HDMI,FRONTEND_NO_SYNC);
                if(this->input == C353_HDMI) 
                {
                    memset(&state->timings, 0, sizeof(state->timings));
                    frontend_update_event(frontend,FRONTEND_EVENT_NO_SYNC);
                }
            }

            if(vga_on)
            {
                if (adv7604_query_dv_timings(this, &timings_ch2, 2)==0)
                {
                    adv7604_print_timings(this, &timings_ch2, "ch2 Detected format:", false);
                    frontend_s_input_status(frontend,C353_VGA,FRONTEND_LOCK);
                    if(this->input == C353_VGA) 
                    {
                        if (no_lock_cp(this) || !v4l_match_dv_timings(&state->timings,&timings_ch2,1000000)) 
                        {
                            adv7604_info(this,"CH2 detected timing is not match setting timings, resetting\n");
                            adv7604_s_dv_timings(this, &timings_ch2);
                            frontend_update_event(frontend,FRONTEND_EVENT_LOCK);
                        }
                    }
                }
            } 
            if(vga_off)
            {
                frontend_s_input_status(frontend,C353_VGA,FRONTEND_NO_SYNC);
                if(this->input == C353_VGA) 
                {
                    memset(&state->timings, 0, sizeof(state->timings));
                    frontend_update_event(frontend,FRONTEND_EVENT_NO_SYNC);                            
                }
            }
            
            adv7604_clean_params(this, &status_on);
        }
        //adv7604_fmt_qevent(this, ret);

        status_70=io_read(this,0x70);
        //raw_6f=io_read(this,0x6f);
        if((status_70 & 0x1e))
        {
            raw_6f=io_read(this,0x6f);
            //adv7604_info(this,"%s status %02x raw %02x\n",__func__,status_70,raw_6f);
            //printk("::%s status %02x raw %02x\n",__func__,status_70,raw_6f);
            if(raw_6f & 0x1e) //ports A to D
            {
                //printk("::>>hotplug_boardtype %d\n",this->device_set->boardtype);
                if (this->device_set->device_ops->cadv7604_hotplug)
                {
                    //printk("::>>>>>hotplug_boardtype %d\n",this->device_set->boardtype);
                    this->device_set->device_ops->cadv7604_hotplug(&pQLCodec->m_hci, this->device_set->boardtype);
	        }
            }else{
                frontend_s_input_status(frontend,C353_HDMI,FRONTEND_NO_SIGNAL);
            }
            io_write(this,0x71,status_70);
            //io_write(This, 0x6e, 0xc0); /* Enable V_LOCKED and DE_REGEN_LCK interrupts */
            //io_write(This, 0x73, 0x12); /* Enable CABLE_DET_A/D_ST (+5v) interrupt */
        }
    }

    __set_current_state(TASK_RUNNING);
    return 0;
}

static ssize_t rep_show(struct device *dev, struct device_attribute *attr,char *buf) 
{
    int i;
    int c=0;
    CDevice *pDevice=(CDevice *)dev_get_drvdata(dev);
    CVideoDecoder  *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, CDevice_getVideoDecoder(pDevice));
    CADV7604 *this = (CADV7604 *)GET_INHERITED_OBJECT(CADV7604, m_VideoDecoder, pVideoDecoder);
  

    for(i=0;i<0x100;i++)
    {
        if( (i & 0xf) ==0)
            c+=sprintf(&buf[c],"%02x:",i);
        c+=sprintf(&buf[c],"%02x",rep_read(this,i));
        if( (i & 0xf) ==0xf)
        {
            c+=sprintf(&buf[c],"\n");
        }else
        {
            c+=sprintf(&buf[c]," ");
        }
    }
    c+=sprintf(&buf[c],"\n");

    return c;
}


static ssize_t io_show(struct device *dev, struct device_attribute *attr,char *buf)
{
  int i;
  int c=0;
  CDevice *pDevice=(CDevice *)dev_get_drvdata(dev);
  CVideoDecoder  *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, CDevice_getVideoDecoder(pDevice));
  CADV7604 *this = (CADV7604 *)GET_INHERITED_OBJECT(CADV7604, m_VideoDecoder, pVideoDecoder);
  
  for(i=0;i<0x100;i++)
   {
       if( (i & 0xf) ==0)
          c+=sprintf(&buf[c],"%02x:",i);
        c+=sprintf(&buf[c],"%02x",io_read(this,i));
        if( (i & 0xf) ==0xf)
        {
            c+=sprintf(&buf[c],"\n");
        }else
        {
            c+=sprintf(&buf[c]," ");
        }
    }
   c+=sprintf(&buf[c],"\n");

   return c;
}

static ssize_t cp_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    int i;
    int c=0;
    CDevice *pDevice=(CDevice *)dev_get_drvdata(dev);
    CVideoDecoder  *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, CDevice_getVideoDecoder(pDevice));
    CADV7604 *this = (CADV7604 *)GET_INHERITED_OBJECT(CADV7604, m_VideoDecoder, pVideoDecoder);
  
    for(i=0;i<0x100;i++)
    {
        if( (i & 0xf) ==0)
            c+=sprintf(&buf[c],"%02x:",i);
        c+=sprintf(&buf[c],"%02x",cp_read(this,i));
        if( (i & 0xf) ==0xf)
        {
            c+=sprintf(&buf[c],"\n");
        }else
        {
            c+=sprintf(&buf[c]," ");
        }
    }
    c+=sprintf(&buf[c],"\n");

    return c;
}

static ssize_t hdmi_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    int i;
    int c=0;
    CDevice *pDevice=(CDevice *)dev_get_drvdata(dev);
    CVideoDecoder  *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, CDevice_getVideoDecoder(pDevice));
    CADV7604 *this = (CADV7604 *)GET_INHERITED_OBJECT(CADV7604, m_VideoDecoder, pVideoDecoder);
  
    for(i=0;i<0x100;i++)
    {
        if( (i & 0xf) ==0)
            c+=sprintf(&buf[c],"%02x:",i);
        c+=sprintf(&buf[c],"%02x",hdmi_read(this,i));
        if( (i & 0xf) ==0xf)
        {
            c+=sprintf(&buf[c],"\n");
        }else
        {
            c+=sprintf(&buf[c]," ");
        }
    }
    c+=sprintf(&buf[c],"\n");

    return c;
}

static DEVICE_ATTR(rep, 0644, rep_show, NULL);
static DEVICE_ATTR(io, 0644, io_show, NULL);
static DEVICE_ATTR(cp, 0644, cp_show, NULL);
static DEVICE_ATTR(hdmi, 0644, hdmi_show, NULL);
static struct attribute *debug_attrs[]={
    &dev_attr_rep.attr,
    &dev_attr_io.attr,
    &dev_attr_cp.attr,
    &dev_attr_hdmi.attr,
    NULL,
};

// IGenericDevice
static struct attribute_group debug_group=
{
  .name="adv7604",  
  .attrs=debug_attrs,  
};


EQP_Errors CADV7604_InitDevice(IVideoDecoder *This_p, 
							   QP_DEVICE_CALLBACK pFuncCallback,
							   TQP_PVOID context,
							   TQP_PVOID pInitData
							   )
{
    CQLCodecLib		*pQLCodecLib;
    CQLCodec 		*pQLCodec;

    CVideoDecoder *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
    CADV7604      *This = GET_INHERITED_OBJECT(CADV7604, m_VideoDecoder, pVideoDecoder);
    CDevice                 *pdevice;
    IFrontend *frontend=&This->m_VideoDecoder.m_frontend;

    adv7604_info(This, "%s %x\n", __func__);
    printk("CADV7604_InitDevice\n");
    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
                 "CADV7604_InitDevice()\n"
                ));

    pQLCodecLib = GET_INHERITED_OBJECT(CQLCodecLib, m_Object, pVideoDecoder->m_Object.m_pParent);
    pQLCodec = pQLCodecLib->m_pMpgCodec;
    pdevice=(CDevice *)pQLCodecLib->m_pDevice;
    mutex_init(&This->fmt_lock);
    INIT_LIST_HEAD(&This->queue_head);
    This->queue_size = 0;

    if (CObject_IsInitialized(&pVideoDecoder->m_Object))
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
                     "CADV7604_InitDevice() Already Initialized. Skipping...\n" 
                    ));
      return (QPERR_SUCCESS);
    }
    if (This->device_set->device_ops->cadv7604_reset)
    {
        This->device_set->device_ops->cadv7604_reset(&pQLCodec->m_hci, This->device_set->boardtype);
        //printk("::video_boardtype %d\n",This->device_set->boardtype);
    }

    sysfs_create_group(&pdevice->m_pGenericDevice->kobj,&debug_group);
#if 0
    if (This->device_set->device_ops->cadv7604_TI3101_reset)
    {
        This->device_set->device_ops->cadv7604_TI3101_reset(&pQLCodec->m_hci, This->device_set->boardtype);
        printk("::audio_boardtype %d\n",This->device_set->boardtype);
    }
#endif

    if(io_read(This, 0xfb) != 0x68)
    {
        adv7604_printk(KERN_ERR, This, "not an adv7604 on address 0x%x\n", ADV7604_DEVICE_ADDR_IO);
        return QPERR_FAIL;
    }

    // Initialize adv7604
    adv7604_Init_Device(This);
    adv7604_cable_select(This);
	
    /* select adv7604 hotplug GPIO_5 or GPIO_6 depending on boardtype*/
    if (This->device_set->device_ops->cadv7604_hotplug)
    {
        This->device_set->device_ops->cadv7604_hotplug(&pQLCodec->m_hci,
	                               This->device_set->boardtype);
        //printk("::hotplug_boardtype %d\n",This->device_set->boardtype);
    }
	
    /* video CLK */
    afe_write(This, 0x13, 0x80);
    io_write(This, 0x33, 0x40);

    /* Enable Video Adjustment */
    cp_write_and_or(This, 0x3e, 0x7f, 0x80);
    This->input=-1;

    // save callback address and context
    //
    pVideoDecoder->m_pDeviceCallback = pFuncCallback;
    pVideoDecoder->m_callbackContext = context;

    // program the video decoder with settings read from the registry
    //
    QPVideoDecoder_SetVideoSystem(This_p, pVideoDecoder->m_VidStandard);        
    QPVideoDecoder_SelectVideoSource(This_p, pVideoDecoder->m_VidSource);

    /* Create and run  polling thread */
    This->polling_thread = kthread_create(adv7604_polling_thread, This, "adv7604_polling");
    if(IS_ERR(This->polling_thread))
    {
        adv7604_info(This, "%s create polling thread failed\n", __func__);
    }

    wake_up_process(This->polling_thread);

    // CObject Init
    
    CObject_Init(&pVideoDecoder->m_Object);
    cadv7604_v4l2_s_input(&This->m_VideoDecoder.m_frontend, C353_HDMI);
    
    frontend_init(frontend,(frontend_setup_t*)This->device_set->inputs_info);
    //printk("::input count %d\n",This->device_set->inputs_info[0].input_count);
   
    return (QPERR_SUCCESS);
}

static void adv7604_switch_audio(CADV7604 *This, enum c353_input_type input)
{
    CVideoDecoder *pVideoDecoder = &This->m_VideoDecoder;
    CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib, m_Object, pVideoDecoder->m_Object.m_pParent);
    CQLCodec *pQLCodec = codeclib->m_pMpgCodec;

    switch(input)
    {
        case C353_HDMI:
        {
            /* high impedance 3101 audio */
            write_gpio(&pQLCodec->m_hci, 10, 1);
            udelay_loop(5);
            /* pass HDMI audio */
            write_gpio(&pQLCodec->m_hci, 11, 0);
            udelay_loop(5);
            break;
        }
        case C353_VGA:
        {
            /* pass 3101 audio */
            write_gpio(&pQLCodec->m_hci, 10, 0);
            udelay_loop(5);
            /* high impedance HDMI audio */
            write_gpio(&pQLCodec->m_hci, 11, 1);
            udelay_loop(5);
            break;
        }
        default:
            adv7604_printk(KERN_ERR,This, "input type not supported\n");
    }
}

EQP_Errors CADV7604_SetVideoRes(IVideoDecoder *This_p,
								TQP_UINT32 dwWid,
				  				TQP_UINT32 dwHgt,
				  				TQP_UINT32 dwScanFmt,
				  				TQP_UINT32 dwFps
				  				)
{
    CVideoDecoder *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
    CADV7604      *This = GET_INHERITED_OBJECT(CADV7604, m_VideoDecoder, pVideoDecoder);
   
	adv7604_info(This, "%s input=%d\n",
			__func__, This->input);

	adv7604_switch_audio(This, This->input);

	return (QPERR_SUCCESS);
}


static int cadv7604_v4l2_s_input(void *frontend, int input)
{
    CADV7604 *This = (CADV7604 *) container_of(frontend,CVideoDecoder,m_frontend);
    int prim_mode;
    struct adv7604_state *state = &This->state;
    struct v4l2_dv_timings timings;

    
    prim_mode = io_read(This, 0x01) & 0xf;


    if (input == C353_HDMI)
    {
        
        state->mode = ADV7604_MODE_HDMI;
        state->rgb_quantization_range = V4L2_DV_RGB_RANGE_AUTO;
        This->input = C353_HDMI;
        
        if (!((prim_mode == 0x5) || (prim_mode == 0x6)))
            io_write_and_or(This, 0x01, 0xf0, 0x05); /* v_freq and prim mode */
    }
    else if (input == C353_VGA)
    {
        
        state->mode = ADV7604_MODE_GR;
        state->rgb_quantization_range = V4L2_DV_RGB_RANGE_FULL;
        This->input = C353_VGA;
        state->rates = 2;
        if (!((prim_mode == 0x1) || (prim_mode == 0x2)))
            io_write_and_or(This, 0x01, 0xf0, 0x02); /* v_freq and prim mode */

    }
    adv7604_switch_audio(This, input);
    adv7604_disable_input(This);
    adv7604_select_input(This);
    adv7604_enable_input(This);
#if 1
    /* enable AGC_MODE_MAN disable GAIN_MAIN */
    //	cp_write_and_or(This, 0x73, 0x3f, 0x50);

    /* CH1_SSPD_PP_EN */
    cp_write(This, 0x84, 0x0e);

    /* CH2_SSPD_PP_EN */
    cp_write_and_or(This, 0x41, 0xfe, 0x01);
#endif
    if (adv7604_query_dv_timings(This, &timings, input+1)==0)
    {
        adv7604_print_timings(This, &timings, " s_input :", false);
        adv7604_s_dv_timings(This, &timings);
        frontend_update_event(frontend,FRONTEND_EVENT_LOCK);
    }

    adv7604_adjust_de_pos(This);
   
	return 0;
}

static void adv7604_get_frameinfo(void *frontend,frontend_frameinfo_t *frameinfo)
{
    CADV7604 *This = (CADV7604 *) container_of(frontend,CVideoDecoder,m_frontend);
    struct adv7604_state *state=&This->state;
    
    if(frameinfo)
    {
        frameinfo->width=state->timings.bt.width;
        frameinfo->height=state->timings.bt.height;
        frameinfo->is_interlace=(state->timings.bt.interlaced!=0);
        frameinfo->framerate=state->frame_rate;
        
        if(state->timings.bt.polarities & V4L2_DV_VSYNC_POS_POL)
        {
            if(state->timings.bt.polarities & V4L2_DV_HSYNC_POS_POL) 
            {
                frameinfo->polarities=H_POS_V_POS;
            }else
            {
                frameinfo->polarities=H_NEG_V_POS;
            }
        }else
        {
            if(state->timings.bt.polarities & V4L2_DV_HSYNC_POS_POL) 
            {
                frameinfo->polarities=H_POS_V_NEG;
            }else
            {
                frameinfo->polarities=H_NEG_V_NEG;
            }
        }
        frameinfo->frontend_select=adv7604_f;
    }
    
    
}
static void adv7604_get_audioinfo(void *frontend,frontend_audioinfo_t *audioinfo)
{
    CADV7604 *This = (CADV7604 *) container_of(frontend,CVideoDecoder,m_frontend);
    struct adv7604_state *state=&This->state;
    
    if(audioinfo)
    {
        switch(state->rates)
        {
        case AUDIO_RATE_32000:
            audioinfo->sample_rate=32000;
            break;
	case AUDIO_RATE_44100:
            audioinfo->sample_rate=44100;
            break;
	case AUDIO_RATE_48000:
            audioinfo->sample_rate=48000;
            break;
	case AUDIO_RATE_88200:
            audioinfo->sample_rate=88200;
            break;
       	case AUDIO_RATE_96000:
            audioinfo->sample_rate=96000;
            break;
	case AUDIO_RATE_176400:
            audioinfo->sample_rate=1764000;
            break;
	case AUDIO_RATE_192000:
            audioinfo->sample_rate=192000;
            break;
        default:
            audioinfo->sample_rate=48000;
            break;
        }
    }
}


static frontend_ops_t adv7604_ops={
    .s_input = &cadv7604_v4l2_s_input,
    //.g_input = &cadv7604_v4l2_g_input,
    .s_ctrl = &adv7604_s_ctrl,
    .g_ctrl = &adv7604_g_ctrl,
    .try_frame = &adv7604_try_frame,
    .get_frameinfo=adv7604_get_frameinfo,
    .get_audioinfo=adv7604_get_audioinfo,
};

static const struct v4l2_queryctrl adv7604_ctrls[] = {
    /* --- video --- */
    {
        .id = V4L2_CID_BRIGHTNESS,
        .name = "Brightness",
        .minimum = -128,
        .maximum = 127,
        .step = 1,
        .default_value = 0,
        .type = V4L2_CTRL_TYPE_INTEGER,
    },
    {
        .id = V4L2_CID_CONTRAST,
        .name = "Contrast",
        .minimum = 0,
        .maximum = 255,
        .step = 1,
        .default_value = 128,
        .type = V4L2_CTRL_TYPE_INTEGER,
    },
    {
        .id = V4L2_CID_SATURATION,
        .name = "Saturation",
        .minimum = 0,
        .maximum = 255,
        .step = 1,
        .default_value = 128,
        .type = V4L2_CTRL_TYPE_INTEGER,
    },
    {
        .id = V4L2_CID_HUE,
        .name = "Hue",
        .minimum = 0,
        .maximum = 255,
        .step = 1,
        .default_value = 0,
        .type = V4L2_CTRL_TYPE_INTEGER,
    },
/*
    {
        .id = V4L2_CID_AUDIO_MUTE,
        .name = "Mute",
        .minimum = 0,
        .maximum = 1,
        .step = 1,
        .default_value = 0,
        .type = V4L2_CTRL_TYPE_BOOLEAN,
    },
*/
};

// constructor
//
CADV7604 * CADV7604_Constructor(CADV7604 *pADV7604, 
							CObject *pParent,
							TQP_UINT32 dwAttributes,
							II2C *pI2C,
							TQP_UINT8 cI2CDevAddr
							)
{
    //int err;
    struct adv7604_state *state;
    //struct v4l2_ctrl_handler *hdl;
    struct pl330_frame *frame;
    IFrontend *frontend=&pADV7604->m_VideoDecoder.m_frontend;
        
    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CADV7604_Constructor()\n"
				 ));

    if (pADV7604)
    {
        CVideoDecoder	*pVideoDecoder;
        pVideoDecoder = CVideoDecoder_Constructor(&pADV7604->m_VideoDecoder, pParent, dwAttributes);

        state = &pADV7604->state;
        frame = &state->cur_out_frame;
        //memcpy(&state->pdata, &g_adv7604_platform_data, sizeof(state->pdata));
        state->restart_stdi_once = true;
        state->out_frame_rate = 0;

        frame->width = 0;
        frame->height = 0;
        /* set default bit rate */
        state->bit_rate = 16000;

        /* Set RGB Full-Range */
        state->rgb_quantization_range = V4L2_DV_RGB_RANGE_AUTO;

        /* V4L2 Input Handling */        
        frontend->ops=&adv7604_ops;
        frontend->v4l2_queryctrls = adv7604_ctrls;
        frontend->v4l2_queryctrls_num = ARRAY_SIZE(adv7604_ctrls);
        pADV7604->g_input_status = &adv7604_g_input_status;

        if (!pVideoDecoder)
        {
            pADV7604 = QP_NULL;
        }
        else
        {
            // fill in the function table

            // IGenericDevice
            pADV7604->m_VideoDecoder.m_iVideoDecoder.InitDevice = CADV7604_InitDevice;

            // IVideoDecoder
            pADV7604->m_VideoDecoder.m_iVideoDecoder.SetVideoRes = CADV7604_SetVideoRes;

            // initialize members
            pADV7604->m_pI2C = pI2C;
            pADV7604->m_cI2CDevAddr = cI2CDevAddr;
        }
    }
	return (pADV7604);
}

// destructor
//
void CADV7604_Destructor(CADV7604 *This)
{
//    TQP_UINT8     buffer[2];
//    TQP_UINT8     ucRegAddr;
//    TQP_UINT8     ucValue;
      CQLCodecLib             *pQLCodecLib;
      CDevice                 *pdevice;

//	struct adv7604_state *state = &This->state;

    pQLCodecLib = GET_INHERITED_OBJECT(CQLCodecLib, m_Object, This->m_VideoDecoder.m_Object.m_pParent);
    pdevice=(CDevice *)pQLCodecLib->m_pDevice;

    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CADV7604_Destructor()\n"
				 ));
    if(This->polling_thread)
        kthread_stop(This->polling_thread); 

    adv7604_info(This, "%s\n", __func__);
    adv7604_fmtqueue_clean(This);
    sysfs_remove_group(&pdevice->m_pGenericDevice->kobj,&debug_group);

    CVideoDecoder_Destructor(&This->m_VideoDecoder);
}

EQP_Errors CADV7604_Create(CQLCodecLib* This ,Chip_Cfg * chip_cfg)
{
    CVideoDecoder	*pVidDecoder;
    CADV7604 * pADV7604 = NULL;

    pVidDecoder = (CVideoDecoder *)QPMMMalloc2Ex(sizeof(CADV7604));
    if (!pVidDecoder)
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
                    "CQLCodecLib_CreatePeripherals() pVidDecoder == NULL!!!\n"
                    ));
        return (QPERR_FAIL);
    }
    This->m_pVidDecoder = (CVideoDecoder *)CADV7604_Constructor((CADV7604 *)pVidDecoder, 
            &This->m_Object,
            OBJECT_DEFAULT,
            I_I2C(This),
            chip_cfg->I2C_add_cfg 
            );
    if (QP_NULL == This->m_pVidDecoder)
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
                    "CQLCodecLib_CreatePeripherals() CADV7604_Constructor() failed\n"
                    ));
        QPMMFree2Ex(pVidDecoder);
        return (QPERR_FAIL);
    }
    pADV7604 = (CADV7604 *)This->m_pVidDecoder;
    pADV7604->device_set = chip_cfg->device_set;
    return QPERR_SUCCESS;
}

void CADV7604_Delete(CQLCodecLib* This)
{
    CADV7604_Destructor((CADV7604 *)This->m_pVidDecoder);
}
