///////////////////////////////////////////////////////////////////////////////
//
// cadv7604.c
//
// Description: 
//   ADV7604 implementation
// 
// History:
//	 Aug 15, 2013 created
//
// Authors: Shihpo Wang
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __CADV7604_H__
#define __CADV7604_H__

#include <linux/videodev2.h>
#include <linux/version.h>
#include <media/v4l2-common.h>
#include <media/v4l2-device.h>
#include <media/v4l2-ctrls.h>

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0))
#include "v4l2-dv-timings.h"
#else
#include <media/v4l2-common.h>
#include <linux/v4l2-dv-timings.h>
#endif

#include "cviddecoder.h"
#include "qpi2c.h"
#include "adv7604.h"
#include "boardinfo-probe.h"

#include "c353.h"
#include "frontend.h"
#include "pl330.h"

struct pl330_frame;
#ifdef __cplusplus
extern "C" {
#endif


/////////////////////////////////////////////////////////////////////////////
//
//

// default ADV7604 I2C address
//
#define ADV7604_DEVICE_ADDR_IO			(0x40 >> 1)
#define ADV7604_DEVICE_ADDR_CP			(0x44 >> 1)
#define ADV7604_DEVICE_ADDR_VDP			(0x48 >> 1)
#define ADV7604_DEVICE_ADDR_AFE_DPLL	(0x4C >> 1)
#define ADV7604_DEVICE_ADDR_TEST_MAP1	(0x60 >> 1)
#define ADV7604_DEVICE_ADDR_REPEATER	(0x64 >> 1)
#define ADV7604_DEVICE_ADDR_HDMI		(0x68 >> 1)
#define ADV7604_DEVICE_ADDR_EDID_RX		(0x6C >> 1)
#define ADV7604_DEVICE_ADDR_ESDP		(0x70 >> 1)
#define ADV7604_DEVICE_ADDR_DPP			(0x78 >> 1)
#define ADV7604_DEVICE_ADDR_INFOFRAME	(0x7C >> 1)
#define ADV7604_DEVICE_ADDR_CEC			(0x80 >> 1)
#define ADV7604_DEVICE_ADDR_AVLINK		(0x84 >> 1)



#define AUDIO_32K		320
#define AUDIO_44K		441
#define AUDIO_48K		480
#define AUDIO_88K		882
#define AUDIO_96K		960
#define AUDIO_176K		1764
#define AUDIO_192K		1920
#define AUDIO_TOLERANCE	50

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/* C353 input type */
enum adv7604_input
{
	ADV7604_HDMI,
	ADV7604_VGA
};

/////////////////////////////////////////////////////////////////////////////

/* kernel-3.7 linux/videodev2.h */

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0))
enum v4l2_dv_rgb_range {
	V4L2_DV_RGB_RANGE_AUTO = 0,
	V4L2_DV_RGB_RANGE_LIMITED = 1,
	V4L2_DV_RGB_RANGE_FULL = 2,
};

/* kernel-3.7 media/v4l2-common.h */
bool v4l_match_dv_timings(const struct v4l2_dv_timings *t1,
							const struct v4l2_dv_timings *t2,
							unsigned pclock_delta);
#endif
/////////////////////////////////////////////////////////////////////////////
struct adv7604_state {
	struct adv7604_platform_data pdata;

	struct v4l2_ctrl_handler hdl;
	enum adv7604_mode mode;
	enum adv7604_audio_sample rates;

	struct v4l2_dv_timings timings;
	u8 edid[256];
	unsigned edid_blocks;
	struct v4l2_fract aspect_ratio;
	u32 rgb_quantization_range;
//	struct workqueue_struct *work_queues;
//	struct delayed_work delayed_work_enable_hotplug;
	bool connector_hdmi;
	bool restart_stdi_once;

	__u32 standards;
	__u32 flags;

	struct pl330_frame cur_out_frame;
	struct pl330_format format;
	u32 frame_rate;
    u32 out_frame_rate;
	u32 bit_rate;

	/* controls */
	struct v4l2_ctrl *detect_tx_5v_ctrl;
	struct v4l2_ctrl *analog_sampling_phase_ctrl;
	struct v4l2_ctrl *free_run_color_manual_ctrl;
	struct v4l2_ctrl *free_run_color_ctrl;
	struct v4l2_ctrl *rgb_quantization_range_ctrl;
};

struct adv7604_fmt_change
{
	unsigned int		fmt_change;
	struct list_head	queue_head;

    struct v4l2_event   event;
};

typedef struct
{
	/* for board initial operation */
    void (*cadv7604_reset)(IHCIAPI *probe_hci,Board_Types boardtype);
    void (*cadv7604_hotplug)(IHCIAPI *probe_hci,Board_Types boardtype);
    void (*cadv7604_TI3101_reset)(IHCIAPI *probe_hci,Board_Types boardtype);
}cadv7604_ops_t;

typedef struct 
{
	//Board_Types board_type;
	bool internal_edid;
	Board_Types boardtype;
	Board_input_info *inputs_info;
	cadv7604_ops_t       *device_ops;
}cadv7604_device_cfg;


typedef struct CADV7604 
{
	// public CVideoDecoder
	//
	CVideoDecoder	m_VideoDecoder;

	// CADV7604
	II2C			*m_pI2C;				// I2C interface
	TQP_UINT8		m_cI2CDevAddr;			// i2c device address

	struct adv7604_state state;
	enum c353_input_type input;

	struct mutex				fmt_lock;
	struct task_struct			*polling_thread;
	struct list_head			queue_head;
	unsigned int				queue_size;


    int (*g_input_status)(void *This, int input, u32 *status);
        
    cadv7604_device_cfg    *device_set;     

} CADV7604;


CADV7604 * CADV7604_Constructor(CADV7604 *pADV7604, 
							    CObject *pParent,
							    TQP_UINT32 dwAttributes,
							    II2C *pI2C,
							    TQP_UINT8 cI2CDevAddr
							    );
void CADV7604_Destructor(CADV7604 *This);
EQP_Errors CADV7604_Create(CQLCodecLib* This ,Chip_Cfg * chip_cfg);
void CADV7604_Delete(CQLCodecLib* This);

struct adv7604_format
{
	char *name;
	u32 pixelformat;
	u8 depth;
	u8 flags;
};

#ifdef __cplusplus
}
#endif

#endif //__CADV7604_H__


