///////////////////////////////////////////////////////////////////////////////
//
// cadv7604.c
//
// Description: 
//   ADV7604 implementation
// 
// History:
//	 Aug 15, 2013 created
//
// Authors: Shihpo Wang
//
///////////////////////////////////////////////////////////////////////////////

#include "qptimemgr.h"
#include "qpdbgmgr.h"
#include "util.h"
#include "plgpio.h"

#include "cqlcodeclib.h"
#include "cqlcodec.h"
#include "pl330.h"
#include "pl330-dbg.h"
#include "cgpio_gspi.h"
#include "cgv7601.h"

#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/kthread.h>
#include <linux/time.h>
#include <linux/list.h>

#if (LINUX_VERSION_CODE > KERNEL_VERSION(3,12,0))
#include <media/v4l2-dv-timings.h>
#else
#include "v4l2-dv-timings.h"
#endif

#define is_h_lock(x) (x & 1)
#define is_v_lock(x) (x & 2)
#define is_lock(x) (x & 3)
#define is_std_lock(x) (x & 0x1000)
#define is_interlaced(x) (x & 0x0800)
#define data_rate(x) (x >> 14)
#define actline(x) (x & 0x07ff)

#define debug adv7604_trace_level
extern unsigned int adv7604_trace_level;

#define gv7601_printk(level, This, fmt, arg...)  \
	printk(level "Device: %02x " fmt,  This->m_cI2CDevAddr, ## arg)

#define gv7601_dbg(level, debug, This, fmt, arg...)						\
	do {																\
		if(adv7604_trace_level >= (level))											\
			gv7601_printk(KERN_DEBUG, This, fmt, ## arg);					\
	} while(0)

#define gv7601_info(This, fmt, arg...)  \
	do {							\
		if(1) \
			printk(KERN_INFO "Device: %02x " fmt,  This->m_cI2CDevAddr, ## arg); \
	} while(0)



static __inline void udelay_loop(TQPTM_Time ms)
{
	mdelay(ms);
}

static const char *video_data_rate[] = {
	"HD",
	"SD",
	"Full HD",
	"SD",
};

static inline unsigned hblanking(const struct v4l2_bt_timings *t)
{
	return t->hfrontporch + t->hsync + t->hbackporch;
}

static inline unsigned _htotal(const struct v4l2_bt_timings *t)
{
	return t->width + t->hfrontporch + t->hsync + t->hbackporch;
}

static inline unsigned vblanking(const struct v4l2_bt_timings *t)
{
	return t->vfrontporch + t->vsync + t->vbackporch;
}

static inline unsigned _vtotal(const struct v4l2_bt_timings *t)
{
	return t->height + t->vfrontporch + t->vsync + t->vbackporch;
}


/* Supported CEA timings (GV7601 SPEC Table 4-20) */
static const struct v4l2_dv_timings gv7601_timings[] = {
	 /* CEA-861 format     VD_STD */
	V4L2_DV_BT_CEA_720X480I59_94,    /* 6 & 7 */
	V4L2_DV_BT_CEA_720X576I50,       /**/
	V4L2_DV_BT_CEA_1280X720P50,      /**/
	V4L2_DV_BT_CEA_1280X720P60,      /**/

	V4L2_DV_BT_CEA_1920X1080P24,     /**/
	V4L2_DV_BT_CEA_1920X1080P25,     /**/
	V4L2_DV_BT_CEA_1920X1080P30,     /**/

	V4L2_DV_BT_CEA_1920X1080I50,     /**/
	V4L2_DV_BT_CEA_1920X1080P50,     /**/

	V4L2_DV_BT_CEA_1920X1080I60,     /**/
	V4L2_DV_BT_CEA_1920X1080P60,     /**/
	{},
};

void gv7601_write_reg(CGV7601 *This, u16 reg, u16 val)
{
	u16 command = 0x0 | reg;

	gpio_gspi_write(&This->gspi_mgr, command, val);
}

void gv7601_read_reg(CGV7601 *This, u16 reg, u16 *val)
{
	u16 command = 0x8000 | reg;

	gpio_gspi_read(&This->gspi_mgr, command, val);
}


//////////////////////////////////////////////////
//

#if 1
static void gv7601_configure_pl330_audio(CGV7601 *This,
		struct gv7601_state *state)
{
    CVideoDecoder *pVideoDecoder = &This->m_VideoDecoder;
	CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib,
			m_Object, pVideoDecoder->m_Object.m_pParent);
	CQLCodec *pQLCodec = codeclib->m_pMpgCodec;

	CDevice *cdevice = (CDevice *)pQLCodec->m_pDevice;
	CComponent *component = cdevice->m_pComponent;

	/* Audio Configuration */
	/* can not set passthrough cause REG_TO_ARM_MESSAGE_STATUS status timeout */
	pl330_set_config(component, "SysFunction-Audio", "audio", false);
	pl330_set_config(component, "SysLink-AudioInput", "aiu", false);
	pl330_set_config(component, "SysLink-AudioOutput", "host", false);

	pl330_set_config(component, "AudioControlParam-AudioType", "aac", false);
	pl330_set_config(component, "AudioControlParam-ChNum", "2", false);
	pl330_set_config(component, "AudioControlEx-AacVer", "mpeg2", false);
	pl330_set_config(component, "AudioControlEx-HType", "adts", false);
	pl330_set_config(component, "AudioControlEx-CutoffFreq", "18000", false);
	pl330_set_config(component, "AudioControlEx-TNS", "1", false);
	pl330_set_config(component, "AudioControlEx-IS", "1", false);
	pl330_set_config(component, "AudioControlEx-PNS", "1", false);
	pl330_set_config(component, "AudioControlEx-MS", "1", false);

	pl330_set_config(component, "AudioControlParam-SampleRate",
			"48k", false);
}

static int cgv7601_configure_pl330(CGV7601 *This,
		struct v4l2_dv_timings *timings)
{
    CVideoDecoder *pVideoDecoder = &This->m_VideoDecoder;
	CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib,
			m_Object, pVideoDecoder->m_Object.m_pParent);
	CQLCodec *pQLCodec = codeclib->m_pMpgCodec;

	CDevice *cdevice = (CDevice *)pQLCodec->m_pDevice;
	CComponent *component = cdevice->m_pComponent;
	struct gv7601_state *state = &This->state;

	struct v4l2_bt_timings *bt = &timings->bt;

	u32 htot, vtot;
	u32 frame_rate = state->frame_rate;
    u8 sync;
	COMPONENT_PORT_OPEN port_open;

	if(!component)
		return -EIO;

	memset(&port_open, 0, sizeof(COMPONENT_PORT_OPEN));

	htot = _htotal(bt);
	vtot = _vtotal(bt);

	gv7601_info(This, "%s frame_rate = %d\n", __func__,
		frame_rate);
	gv7601_info(This, "%s bit_rate = %d\n", __func__,
		state->bit_rate);
	gv7601_info(This, "%s interlaced = %d\n", __func__,
		bt->interlaced);

	pl330_set_config(component, "SystemControl-StreamData", "video+audio", false);
	pl330_set_config(component, "SystemControl-XferMode", "frame", false);
	pl330_set_config(component, "SystemControl-SpsrFreq", "1", false);
	pl330_set_config(component, "SystemControl-FFMode", "frame", false);
	pl330_set_config(component, "SystemControl-VMode", "cavlc", false);

	if(bt->interlaced)
	{
		pl330_set_config(component, "InputControl-ScanFormat", "interlace", false);
		pl330_set_config(component, "InputControl-Fmt", "interlace", false);
		pl330_set_config(component, "DeInterlace-Mode", "topfield", false);
        frame_rate /= 2;
	}
	else
	{
		pl330_set_config(component, "InputControl-ScanFormat", "progressive", false);
		pl330_set_config(component, "InputControl-Fmt", "progressive", false);
		pl330_set_config(component, "DeInterlace-Mode", "none", false);
	}

    sync = 0;
	if(bt->polarities & V4L2_DV_VSYNC_POS_POL) {
        gv7601_info(This, "V +\n");
    }
    else {
        gv7601_info(This, "V -\n");
        sync |= 1;
    }

	if(bt->polarities & V4L2_DV_HSYNC_POS_POL) {
        gv7601_info(This, "H +\n");
    }
    else {
        gv7601_info(This, "H -\n");
        sync |= (1 << 1);
    }
    gv7601_info(This, "sync =%d\n", sync);

	switch(sync) {
		case 0:
			pl330_set_config(component, "InputControl-SyncMode", "0", false);
			break;
		case 1:
			pl330_set_config(component, "InputControl-SyncMode", "1", false);
			break;
		case 2:
			pl330_set_config(component, "InputControl-SyncMode", "2", false);
			break;
		case 3:
			pl330_set_config(component, "InputControl-SyncMode", "3", false);
			break;
	}

    state->out_frame_rate = frame_rate;
	pl330_set_config(component, "InputControl-SrcMode", "hdmi", false);
	pl330_set_config(component, "InputControl-DataType", "raw", false);
	pl330_set_config(component, "InputControl-InFrameRate", &frame_rate, true);

	pl330_set_config(component, "InputControl-CkEdge", "positive", false);
	pl330_set_config(component, "InputControl-Force60pIn", "1", false);

	pl330_set_config(component, "RateControl-Vbr", "0", false);
	pl330_set_config(component, "RateControl-Mode", "viu", false);

#if 10
	pl330_set_config(component, "RateControl-AvgBitRate", &state->bit_rate, true);
	pl330_set_config(component, "VbrBitRate-MinBitRate", &state->bit_rate, true);
	pl330_set_config(component, "VbrBitRate-MaxBitRate", &state->bit_rate, true);
#endif
	pl330_set_config(component, "FilterControl-StartPixel", "1", false);
	pl330_set_config(component, "FilterControl-StartLine", "0", false);

	pl330_set_config(component, "GopLoopFilter-IntraPeriod", "30", false);
	pl330_set_config(component, "GopLoopFilter-BNum", "0", false);
	pl330_set_config(component, "GopLoopFilter-Idr", "close", false);

	pl330_set_config(component, "SysFunction-Function", "encode", false);
	pl330_set_config(component, "SysLink-VideoInput", "viu", false);
	pl330_set_config(component, "SysLink-VideoOutput", "host", false);

	gv7601_configure_pl330_audio(This, state);

	return 0;
}
#endif

static int cgv7601_v4l2_s_input(void *frontend, int input)
{
	CGV7601 *This = (CGV7601 *)container_of(frontend,CVideoDecoder,m_frontend);

	cgv7601_configure_pl330(This, &This->state.timings);

	return 0;
}

static void gv7601_configure_resolution(CGV7601 *This, CComponent *component,
		struct pl330_frame *frame, struct pl330_format *format,
		COMPONENT_PORT_OPEN *port_open)
{
	struct gv7601_state *state = &This->state;
	struct v4l2_dv_timings *timings = &state->timings;
	struct v4l2_bt_timings *bt = &timings->bt;
	
	int w = frame->width;
	int h = frame->height;

	gv7601_info(This, "%s %d:%d\n", __func__,
		w, h);

	pl330_set_config(component, "PictureResolution-InPicWidth",
			&bt->width, true);
	pl330_set_config(component, "PictureResolution-InPicHeight",
			&bt->height, true);

	if(w == 0 || h == 0 || w > bt->width || h > bt->height ||
	   (format->pixelformat == V4L2_PIX_FMT_YVU420) || 
       (format->pixelformat == V4L2_PIX_FMT_YUV420) ||
       (format->pixelformat == V4L2_PIX_FMT_NV12))
	{
		pl330_set_config(component, "OutPictureResolution-OutPicWidth",
				&bt->width, true);
		pl330_set_config(component, "OutPictureResolution-OutPicHeight",
				&bt->height, true);

		w = bt->width;
		h = bt->height;
	}
	else
	{
		pl330_set_config(component, "OutPictureResolution-OutPicWidth",
				&w, true);
		pl330_set_config(component, "OutPictureResolution-OutPicHeight",
				&h, true);
	}

	state->cur_out_frame.width = w;
	state->cur_out_frame.height = h;
	gv7601_info(This, "%s current out frame %ux%u\n", __func__,
		w, h);

	if((format->pixelformat == V4L2_PIX_FMT_YVU420) || 
    (format->pixelformat == V4L2_PIX_FMT_YUV420) ||
    (format->pixelformat == V4L2_PIX_FMT_NV12))
	{
		port_open->format.yuv.nWidth = w;
		port_open->format.yuv.nHeight = h;
	}
}


static int gv7601_g_input_status(void *This, u32 *status)
{
	struct gv7601_state *state = &((CGV7601 *)This)->state;

	*status = 0;
	*status |= (state->is_lock && state->ceaid != 0) ? 0 :
		V4L2_IN_ST_NO_SIGNAL | V4L2_IN_ST_NO_SYNC | V4L2_IN_ST_NO_H_LOCK;
    return 0;
}

static int gv7601_try_frame(void *frontend, struct pl330_frame *frame,
			struct pl330_format *format)
{
	CGV7601 *This = (CGV7601 *)container_of(frontend,CVideoDecoder,m_frontend); 
    CVideoDecoder *pVideoDecoder = &This->m_VideoDecoder;
	CQLCodecLib	*codeclib = GET_INHERITED_OBJECT(CQLCodecLib,
			m_Object, pVideoDecoder->m_Object.m_pParent);
	CQLCodec *pQLCodec = codeclib->m_pMpgCodec;
	CDevice *cdevice = (CDevice *)pQLCodec->m_pDevice;
	CComponent *component = cdevice->m_pComponent;
	struct gv7601_state *state = &This->state;

	COMPONENT_PORT_OPEN port_open;
	memset(&port_open, 0, sizeof(COMPONENT_PORT_OPEN));

	gv7601_info(This, "%s %ux%u input=%ux%u@%u%c.\n",
			__func__,
			frame->width, frame->height,
			state->timings.bt.width, state->timings.bt.height,
			state->frame_rate,
			state->timings.bt.interlaced ? 'i' : 'p');

	if(state->timings.bt.width == 0 && state->timings.bt.height == 0)
	{
		gv7601_info(This, "input no signal\n");
		return -EIO;
	}

	if(!state->frame_rate)
		return -EINVAL;
    
	if((format->pixelformat == V4L2_PIX_FMT_YVU420) || 
    (format->pixelformat == V4L2_PIX_FMT_YUV420) ||
    (format->pixelformat == V4L2_PIX_FMT_NV12))
	{
		if(frame->width != state->timings.bt.width ||
		   frame->height != state->timings.bt.height)
		{
			gv7601_info(This, "%s pixelformat is YV12 frame resolution %ux%u "
					"must be equal to source.\n",
					__func__, frame->width, frame->height);
		}
	}

	gv7601_configure_resolution(This, component, frame, format, &port_open);

	gv7601_info(This, "frame_rate=%u\n", state->out_frame_rate);

    if(state->out_frame_rate == 0)
        state->out_frame_rate = state->frame_rate;

	/* if resolution > 1920x1080p set frame rate to 30 */
    if(state->out_frame_rate >= 30) {
    	if(state->cur_out_frame.width == 1920)
		    if(state->timings.bt.interlaced == V4L2_DV_PROGRESSIVE)
			    state->out_frame_rate = 30;
	    else if(state->cur_out_frame.width == 1680)
	    	if(state->timings.bt.interlaced == V4L2_DV_PROGRESSIVE)
		    	state->out_frame_rate = 30;
    }

//    state->out_frame_rate = 60;
    gv7601_info(This, "%s frame_rate set to %d\n", __func__, state->out_frame_rate);
    pl330_set_config(component, "InputControl-OutFrameRate",
            &state->out_frame_rate, true);

	gv7601_info(This, "frame_rate=%u\n", state->out_frame_rate);
    //LOG_NUM(format->pixelformat);

	switch(format->pixelformat)
	{
		case V4L2_PIX_FMT_MPEG:
			pl330_set_config(component, "SystemControl-StreamType", "ts", false);
			pl330_set_config(component, "SysFunction-Video", "h264", false);
			port_open.type = COMPONENT_PORT_COMP_OUT;
			port_open.format.vid.nDataType = QP_COMP_VID_DATA_TYPE_H264;
			break;
		case V4L2_PIX_FMT_YVU420:
        case V4L2_PIX_FMT_NV12:
			pl330_set_config(component, "SystemControl-StreamType", "raw", false);
			pl330_set_config(component, "SysFunction-Video", "passthrough", false);
			port_open.type = COMPONENT_PORT_YUV_OUT;
			port_open.format.yuv.nBitCount = 12;
			port_open.format.yuv.nDataType = QP_YUV_DATA_TYPE_YV12;
			port_open.format.yuv.nFrameRate = state->out_frame_rate;
			break;
		case V4L2_PIX_FMT_YUV420:
			pl330_set_config(component, "SystemControl-StreamType", "raw", false);
			pl330_set_config(component, "SysFunction-Video", "passthrough", false);
			port_open.type = COMPONENT_PORT_YUV_OUT;
			port_open.format.yuv.nBitCount = 12;
			port_open.format.yuv.nDataType = QP_YUV_DATA_TYPE_I420;
			port_open.format.yuv.nFrameRate = state->out_frame_rate;
			break;
	}

	port_open.memory = V4L2_MEMORY_MMAP;

	pl330_port_open(component, &port_open);

	pl330_acquire(component);

	return 0;
}

static bool check_format_id(CGV7601 *this)
{
	struct gv7601_state *state = &this->state;
	//struct v4l2_dv_timings timings;
	//struct v4l2_bt_timings *bt = &state->timings.bt;
	u16 _0x25h;
	u16 _0x06h;
	int ceaid;
	bool need_reconf = false;

	gv7601_read_reg(this, 0x25, &_0x25h);
	gv7601_read_reg(this, 0x06, &_0x06h);

	if(_0x25h & 0x0040) {
		gv7601_dbg(2, debug, this, "%s not recognized format\n", __func__);
		state->ceaid = 0;
		state->out_frame_rate = 0;
		state->frame_rate = 0;
		memset(&state->timings, 0, sizeof(state->timings));
		return false;
	}

	ceaid = _0x25h & 0x003f;
	if(ceaid != state->ceaid) {
		gv7601_dbg(2, debug, this, "%s new stanrard detected\n", __func__);
		state->ceaid = ceaid;
		need_reconf = true;
	}

	state->vd_std = (_0x06h & 0x3f00) >> 8;
	gv7601_dbg(2, debug, this, "%s CEA-861 format ID=%u\n",
			__func__, state->ceaid);
	gv7601_dbg(2, debug, this, "%s Detect Video Standard=%02xh\n",
			__func__, state->vd_std);

	switch(ceaid)
	{
		case 4:
			state->timings = gv7601_timings[3];
			state->frame_rate = 60;
			break;
		case 5:
			state->timings = gv7601_timings[9];
			state->frame_rate = 60;
			break;
		case 6:
		case 7:
			state->timings = gv7601_timings[0];
			state->frame_rate = 60;
			break;
		case 16:
			state->timings = gv7601_timings[10];
			state->frame_rate = 60;
			break;
		case 19:
			state->timings = gv7601_timings[2];
			state->frame_rate = 50;
			break;
		case 20:
			state->timings = gv7601_timings[7];
			state->frame_rate = 50;
			break;
		case 21:
		case 22:
			state->timings = gv7601_timings[1];
			state->frame_rate = 50;
			break;
		case 31:
			state->timings = gv7601_timings[8];
			state->frame_rate = 50;
			break;
		case 32:
			state->timings = gv7601_timings[4];
			state->frame_rate = 24;
			break;
		case 33:
			state->timings = gv7601_timings[5];
			state->frame_rate = 25;
			break;
		case 34:
			state->timings = gv7601_timings[6];
			state->frame_rate = 30;
			break;
		default:
			return false;
	}
/*
	printk("%s %ux%u%c%u\n", __func__,
			state->timings.bt.width,
			state->timings.bt.height,
			state->timings.bt.interlaced ? 'i' : 'p',
			state->frame_rate);
*/
	return true;
}
static int cgv7601_g_ctrl(void *frontend, struct v4l2_control *ctrl)
{
	CGV7601 *This = (CGV7601 *)container_of(frontend,CVideoDecoder,m_frontend); 

	u16 reg = 0;

	switch (ctrl->id) {
		case V4L2_CID_AUDIO_MUTE:
			gv7601_read_reg(This,0x200,&reg);
			//LOG_HEX(reg);
			ctrl->value = (reg>>12)&1;
		        gv7601_read_reg(This,0x400,&reg);
		        //LOG_HEX(reg);
		break;
	}

	return 0;
}
static int cgv7601_s_ctrl(void *frontend, struct v4l2_control *ctrl)
{
	CGV7601 *This = (CGV7601 *)container_of(frontend,CVideoDecoder,m_frontend);

	u16 reg = 0;

	switch (ctrl->id) {
		case V4L2_CID_AUDIO_MUTE:
			if(ctrl->value) {
				gv7601_write_reg(This,0x200,reg|(1<<12));
				udelay(50);
				gv7601_write_reg(This,0x400,reg|(1<<11));
				udelay(50);
				gv7601_write_reg(This,0x205,0xff);
				udelay(50);
				gv7601_write_reg(This,0x405,0xff);
				udelay(50);
			} else {
				gv7601_write_reg(This,0x200,reg&(~(1<<12)));
				udelay(50);
				gv7601_write_reg(This,0x400,reg&(~(1<<11)));
				udelay(50);
				gv7601_write_reg(This,0x205,0x0);
				udelay(50);
				gv7601_write_reg(This,0x405,0x0);
				udelay(50);
			}
		break;
	}
	return 0;
}

static void check_timing(CGV7601 *this)
{
	struct gv7601_state *state = &this->state;
	//struct v4l2_bt_timings *bt = &state->timings.bt;

	u16 _0x22h;
	u16 width;
	u16 height;
	u16 htotal;
	u16 vtotal;
	//int i, data_rate;
	int data_rate;

	gv7601_read_reg(this, 0x1f, &width);
	gv7601_read_reg(this, 0x20, &htotal);
	gv7601_read_reg(this, 0x21, &vtotal);
	gv7601_read_reg(this, 0x22, &_0x22h);

	state->is_std_lock = is_std_lock(_0x22h);
	state->interlaced = is_interlaced(_0x22h);
	height = actline(_0x22h);
	data_rate = data_rate(_0x22h);
	
	state->now_width = width;
	state->now_height = height;

	gv7601_dbg(2, debug, this, "%s lock=%d, std_lock=%x, interlaced=%d datarate=%d\n",
			__func__,
			state->is_lock,
			state->is_std_lock,
			state->interlaced, data_rate);

	gv7601_dbg(2, debug, this, "%s 0x22h=0x%x\n", __func__, _0x22h);
	gv7601_dbg(2, debug, this, "%s %ux%u%c %s, htotal=%u, vtotal=%u\n",
			__func__,
			width, height, state->interlaced ? 'i' : 'p',
			video_data_rate[data_rate],
			htotal, vtotal);
}

static void check_signal(CGV7601 *this)//
{
	struct gv7601_state *state = &this->state;
	u16 lock;
    IFrontend *frontend=&this->m_VideoDecoder.m_frontend;

	gv7601_read_reg(this, 0x23, &lock);

	gv7601_dbg(2, debug, this, "%s lock=0x%x\n", __func__, lock);

	state->is_lock = is_lock(lock);
	
	state->last_interlaced = state->interlaced;
	state->last_width = state->now_width;
	state->last_height = state->now_height;
	

    #if 1
    if(!state->is_lock) {
        if(frontend_g_input_status(frontend) != FRONTEND_NO_SIGNAL)
        {
            //LOG("FRONTEND_NO_SIGNAL");
            frontend_s_input_status(frontend,CM313B_SDI,FRONTEND_NO_SIGNAL);
            frontend_update_event(frontend,FRONTEND_NO_SIGNAL);

        }
        gv7601_dbg(2, debug, this, "%s no signal lock\n", __func__);
        return;
    }
	check_format_id(this);
	check_timing(this);
	
	if(frontend_g_input_status(frontend) != FRONTEND_NO_SIGNAL)
	{
	    if ((state->last_width != state->now_width) || (state->last_height != state->now_height))
	    {
		    //LOG("FRONTEND_NO_SIGNAL...");
            frontend_s_input_status(frontend,CM313B_SDI,FRONTEND_NO_SIGNAL);
            frontend_update_event(frontend,FRONTEND_NO_SIGNAL);
	    }
	    return;
	}
	
    if(frontend_g_input_status(frontend) != FRONTEND_LOCK)
    {
        //LOG("FRONTEND_LOCK");
        frontend_s_input_status(frontend,CM313B_SDI,FRONTEND_LOCK);
        frontend_update_event(frontend,FRONTEND_EVENT_LOCK);
    }
    #else
    if(!state->is_lock) {
            frontend_s_input_status(frontend,CM313B_SDI,FRONTEND_NO_SIGNAL);

        gv7601_dbg(2, debug, this, "%s no signal lock\n", __func__);
        return;
    }
        frontend_s_input_status(frontend,CM313B_SDI,FRONTEND_LOCK);
        frontend_update_event(frontend,FRONTEND_EVENT_LOCK);

    #endif

}

static int cgv7601_mon_thread(void *__this)
{

	CGV7601 *this = (CGV7601 *) __this;

	for(;;)
	{
		if(kthread_should_stop())
			break;
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ);

		check_signal(this);
	}

	__set_current_state(TASK_RUNNING);
	return 0;
}

// IGenericDevice
//
EQP_Errors CGV7601_InitDevice(IVideoDecoder *This_p, 
							   QP_DEVICE_CALLBACK pFuncCallback,
							   TQP_PVOID context,
							   TQP_PVOID pInitData
							   )
{	
    CQLCodecLib		*pQLCodecLib;
	CQLCodec 		*pQLCodec;

    CVideoDecoder *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
    CGV7601      *This = GET_INHERITED_OBJECT(CGV7601, m_VideoDecoder, pVideoDecoder);
	//struct gv7601_state *state = &This->state;
    u16 check_7601;

	gv7601_info(This, "%s\n", __func__);

    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
                 "CGV7601_InitDevice()\n"
                ));
	pQLCodecLib = GET_INHERITED_OBJECT(CQLCodecLib, m_Object, pVideoDecoder->m_Object.m_pParent);
	pQLCodec = pQLCodecLib->m_pMpgCodec;

	mutex_init(&This->fmt_lock);
	INIT_LIST_HEAD(&This->queue_head);
	This->queue_size = 0;


	gpio_gspi_init(&This->gspi_mgr, &pQLCodec->m_hci, 15, 14, 13);
	gpio_gspi_chipselect(&This->gspi_mgr, 12);
    gv7601_read_reg(This,0x1,&check_7601);
    //LOG_HEX(check_7601);
    if(check_7601 != 0x100)
    {
        //LOG("Init fail, reg 0x001 , %d != 0x100",check_7601);
        return QPERR_FAIL;
    }

    if (CObject_IsInitialized(&pVideoDecoder->m_Object))
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
                     "CGV7601_InitDevice() Already Initialized. Skipping...\n" 
                    ));
      return (QPERR_SUCCESS);
    }

	/* CEA-861 timing mode */
	write_gpio(&pQLCodec->m_hci, 1, 1);

    // save callback address and context
    //
    pVideoDecoder->m_pDeviceCallback = pFuncCallback;
    pVideoDecoder->m_callbackContext = context;

    
    // program the video decoder with settings read from the registry
    //
    QPVideoDecoder_SetVideoSystem(This_p, pVideoDecoder->m_VidStandard);        
    QPVideoDecoder_SelectVideoSource(This_p, pVideoDecoder->m_VidSource);

#if 1
	/* Create and run  polling thread */
	This->mon_thread = kthread_create(cgv7601_mon_thread,
			This, "gv7601_mon");
	if(IS_ERR(This->mon_thread))
	{
		gv7601_info(This, "%s create monitor thread failed\n", __func__);
	}

	wake_up_process(This->mon_thread);
#endif
    // CObject Init
    //
    CObject_Init(&pVideoDecoder->m_Object);

    return 0;
}

EQP_Errors CGV7601_SetVideoRes(IVideoDecoder *This_p,
								TQP_UINT32 dwWid,
				  				TQP_UINT32 dwHgt,
				  				TQP_UINT32 dwScanFmt,
				  				TQP_UINT32 dwFps
				  				)
{
    CVideoDecoder *pVideoDecoder = GET_INHERITED_OBJECT(CVideoDecoder, m_iVideoDecoder, This_p);
    CGV7601      *This = GET_INHERITED_OBJECT(CGV7601, m_VideoDecoder, pVideoDecoder);
   
	gv7601_info(This, "%s input=%d\n",
			__func__, This->input);

	return (QPERR_SUCCESS);
}
static void gv7601_get_frameinfo(void *frontend,frontend_frameinfo_t *frameinfo)
{
	CGV7601 *This = (CGV7601 *)container_of(frontend,CVideoDecoder,m_frontend); 
	struct gv7601_state *state = &This->state;
    
    if(frameinfo)
    {
        frameinfo->width=state->timings.bt.width;
        frameinfo->height=state->timings.bt.height;
        frameinfo->is_interlace=(state->timings.bt.interlaced!=0);
        frameinfo->framerate=state->frame_rate;
        
        if(state->timings.bt.polarities & V4L2_DV_VSYNC_POS_POL)
        {
            if(state->timings.bt.polarities & V4L2_DV_HSYNC_POS_POL) 
            {
                frameinfo->polarities=H_POS_V_POS;
            }else
            {
                frameinfo->polarities=H_NEG_V_POS;
            }
        }else
        {
            if(state->timings.bt.polarities & V4L2_DV_HSYNC_POS_POL) 
            {
                frameinfo->polarities=H_POS_V_NEG;
            }else
            {
                frameinfo->polarities=H_NEG_V_NEG;
            }
        }
        frameinfo->frontend_select=gv7601_f;
    }
    
    
}
static const struct v4l2_queryctrl gv7601_ctrls[] = {
    {
        .id = V4L2_CID_AUDIO_MUTE,
        .name = "Mute",
        .minimum = 0,
        .maximum = 1,
        .step = 1,
        .default_value = 0,
        .type = V4L2_CTRL_TYPE_BOOLEAN,
        .reserved={CHIP_ADV7604_FLAG,0},
    },
};

static frontend_ops_t gv7601_ops={
    .s_input = &cgv7601_v4l2_s_input,
    .s_ctrl = cgv7601_s_ctrl,
    .g_ctrl = cgv7601_g_ctrl,
    .try_frame = &gv7601_try_frame,
    .get_frameinfo=gv7601_get_frameinfo,
};
static frontend_setup_t cm313b_frontend=
{
   .input_count=1,
   .input_name=
    {
       "SDI",
   },
     
};


EQP_Errors CGV7601_Create(CQLCodecLib* This ,Chip_Cfg * chip_cfg)
{
    CVideoDecoder	*pVidDecoder;

    pVidDecoder = (CVideoDecoder *)QPMMMalloc2Ex(sizeof(CGV7601));
    if (!pVidDecoder)
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
                    "CQLCodecLib_CreatePeripherals() pVidDecoder == NULL!!!\n"
                    ));
        return (QPERR_FAIL);
    }

    This->m_pVidDecoder = (CVideoDecoder *)CGV7601_Constructor((CGV7601 *)pVidDecoder, 
            &This->m_Object,
            OBJECT_DEFAULT,
            I_I2C(This),
            0
            );
    if (QP_NULL == This->m_pVidDecoder)
    {
        QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_ERROR),
                    "CQLCodecLib_CreatePeripherals() CGV7601_Constructor() failed\n"
                    ));
        QPMMFree2Ex(pVidDecoder);
        return (QPERR_FAIL);
    }
    return QPERR_SUCCESS;
}

void CGV7601_Delete(CQLCodecLib* This)
{
    CGV7601_Destructor((CGV7601 *)This->m_pVidDecoder);
}
// constructor
//
CGV7601 * CGV7601_Constructor(CGV7601 *pGV7601, 
								CObject *pParent,
								TQP_UINT32 dwAttributes,
								II2C *pI2C,
								TQP_UINT8 cI2CDevAddr
								)
{
	//int err;
	struct gv7601_state *state;
	struct pl330_frame *frame;
    IFrontend *frontend=&pGV7601->m_VideoDecoder.m_frontend;

	QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CGV7601_Constructor()\n"
				 ));


	if (pGV7601)
	{
		CVideoDecoder	*pVideoDecoder;

		pVideoDecoder = CVideoDecoder_Constructor(&pGV7601->m_VideoDecoder, 
												  pParent,
												  dwAttributes
												  );
		state = &pGV7601->state;
		frame = &state->cur_out_frame;

		/* reset parameters */
		frame->width = 0;
		frame->height = 0;
		state->frame_rate = 0;
		state->ceaid = 0;
		state->vd_std = 0;
		state->bit_rate = 16000;

        /* V4L2 Input Handling */
        frontend_init(frontend,&cm313b_frontend);
        frontend->ops=&gv7601_ops;
        frontend->v4l2_queryctrls = gv7601_ctrls;
        frontend->v4l2_queryctrls_num = ARRAY_SIZE(gv7601_ctrls);

		pGV7601->g_input_status = &gv7601_g_input_status;

		if (!pVideoDecoder)
		{
			pGV7601 = QP_NULL;
		}
		else
		{
			
			// fill in the function table
			//

			// IGenericDevice
			pGV7601->m_VideoDecoder.m_iVideoDecoder.InitDevice = CGV7601_InitDevice;

			// IVideoDecoder
			pGV7601->m_VideoDecoder.m_iVideoDecoder.SetVideoRes = CGV7601_SetVideoRes;

			// initialize members
			//
			pGV7601->m_pI2C = pI2C;
			pGV7601->m_cI2CDevAddr = cI2CDevAddr;
		}
	}

	return (pGV7601);
}

// destructor
//
void CGV7601_Destructor(CGV7601 *This)
{
    QPDbgMPrintf((QPDBGMCOMPLVL(g_QPPeripheralDBGCompID, QPDBGM_LEVEL_TRACE),
				 "CGV7601_Destructor()\n"
				 ));
    if(This->mon_thread)
	    kthread_stop(This->mon_thread); 

	gv7601_info(This, "%s\n", __func__);

    CVideoDecoder_Destructor(&This->m_VideoDecoder);
}

